/*
 * Class    br.org.scadabr.luciol.Application
 * File     Application.java
 * Author   Alexis BIETTI
 *
 * This interface is the entry point to launch applications from the LUCIOL
 * main method.
 */
package br.org.scadabr.luciol;


/**
 * All applications of Luciol MUST implement this.
 *
 * <p>
 * This interface is the entry point for launching application sessions and to
 * configure these sessions.
 * </p>
 *
 * <p>
 * A good implementation of this interface should :
 *
 * <ul>
 * <li>
 * Forbid the instanciation of this class (make the constructor private)
 * </li>
 * <li>
 * Allow a public access to a singleton instance of this class, by providing a
 * method : <code>public static Application getInstance()</code>.
 * </li>
 * </ul>
 * </p>
 *
 * <p>
 * An application may allow several sessions (a protocol) or one session only
 * (luciol I/O).
 * </p>
 *
 * <p>
 * Sessions MUST be numbered sequentially in order of creation. Session
 * numbers MUST start with zero.
 * </p>
 *
 * <p>
 * The only way to destroy a session is to destroy them all by invoking the
 * method {@link #deleteAllSessions()}. Thus, users SHOULD NOT care about
 * deleting a single session and renumbering them.
 * </p>
 *
 * <p>
 * Some Applications may restrict the number of sessions to a maximum of one.
 * </p>
 *
 * @author <a href="mailto:alexis.bietti@sysaware.com">Alexis BIETTI</a>
 *
 * @see AppSession
 * @see Configurable
 * @see LUCIOL
 */
public interface Application {
  /**
   * Return a title for the configuration page. This name may be the name of a
   * protocol.
   *
   * <p>
   * The format for the String is free, except that it MUST not contain one of
   * following characters: <code>'&lt;'</code>, <code>'&gt;'</code>,
   * <code>'&amp;'</code>
   * </p>
   *
   * @return a title for the HTML page.
   */
  public String applicationName();

  /**
   * Return the maximum number of sessions this application can run
   * concurrently.
   *
   * @return a strictly positive integer.
   */
  public int numberOfSessionsAllowed();

  /**
   * Open a new session for this application and return the number identifier
   * of this session.
   *
   * <p>
   * If a new session can not be created, this method returns -1.
   * </p>
   *
   * <p>
   * New sessions MUST start with a default configuration.
   * </p>
   *
   * @return the number of the newly open session, or -1 if no session could
   *         be open.
   */
  public int openNewSession();

  /**
   * Return the current number of open sessions for this application.
   *
   * @return the number of sessions currently open for this application.
   */
  public int numberOfCurrentSessions();

  /**
   * Close all sessions of this application.
   *
   * <p>
   * This method MUST reset the counter for session identifier. Thus the next
   * session created MUST have the identifier zero.
   * </p>
   */
  public void deleteAllSessions();

  /**
   * Return the session of this application identified by the parameter.
   *
   * @param session a session identifier.
   *
   * @return an application session, or null if not found.
   */
  public AppSession getSession(int session);

  /**
   * Return the software version of this application. The format of the
   * version is a String <code>"major.minor"</code>.
   *
   * @return the software version of this application.
   */
  public String getVersion();

  /**
   * Return the release date of this application.
   *
   * @return the release date of this application.
   */
  public String getDate();
}


// End Application.java
