/*
 * Class    br.org.scadabr.luciol.Configurable
 * File     Configurable.java
 * Author   Alexis BIETTI
 *
 * <p>
 * This interface defines the requirements for configurable application
 * sessions in Luciol.
 */

//-----------------------------------------------------------------------------
// Imports
//-----------------------------------------------------------------------------
package br.org.scadabr.luciol;

import br.org.scadabr.protocol.lib.xml.XMLInterface;

//-----------------------------------------------------------------------------
// Class
//-----------------------------------------------------------------------------

/**
 * Applications that want to be configured via the HTTP server must implement
 * this interface.
 *
 * <p>
 * This mechanism avoids to write specific servlets for every application.
 * </p>
 *
 * <p>
 * This interface extends XMLInterface in order to add XML functionnality to
 * this mechanism.
 *
 * <ul>
 * <li>
 * XML should be used to describe ALL parameters, including data mappings
 * </li>
 * <li>
 * Other parameters are just tuning parameters that may be modified in the
 * runtime.
 * </li>
 * </ul>
 * </p>
 *
 * <p>
 * Modifications:
 *
 * <dl>
 * <dt>
 * 10.IV.2003
 * </dt>
 * <dd>
 * Added the method {@link #getFileId()}
 * </dd>
 * </dl>
 * </p>
 *
 * @author <a href="mailto:alexis.bietti@sysaware.com">Alexis BIETTI</a>
 * @version 1.1 - 10.IV.2003
 *
 * @see br.org.scadabr.io.XMLInterface
 */
public interface Configurable extends XMLInterface {
  //---------------------------------------------------------------------------
  // Constants
  //---------------------------------------------------------------------------

  /** All parameters that use a text box : strings, numbers... */
  public static final byte CDATA = 0;

  /** Parameters that are a list of choice. */
  public static final byte SELECT = 1;

  /** For boolean parameters. */
  public static final byte CHECK = 2;

  //---------------------------------------------------------------------------
  // Methods
  //---------------------------------------------------------------------------

  /**
   * Return a title for the configuration page. This name may be the name of
   * the protocol.
   *
   * @return a title for the HTML page.
   */
  public String applicationName();

  /**
   * Get the version of the configuration file used to configure this object.
   *
   * @return the version of the configuration file used to configure this
   *         object.
   */
  public String getVersion();

  /**
   * Get the modification date of the file used to configure this object.
   *
   * @return the modification date of the file used to configure this object.
   */
  public String getDate();

  /**
   * Get the name of the author of the configuration file.
   *
   * @return the name of the author of the configuration file.
   */
  public String getAuthor();

  /**
   * Return an identifier or the name of the configuration file.
   *
   * @return an identifier or the name of the configuration file.
   */
  public String getFileId();

  /**
   * Return the number of configurable parameters this servlet has.
   *
   * @return a strictly positive number.
   */
  public int getParameterCount();

  /**
   * Return a long name for a parameter. This name may include a short
   * description, or a description of how the value must be formatted.
   *
   * <p>
   * <b>Precondition</b> : <tt>0 &lt;= param &lt; getParameterCount()</tt>
   * </p>
   *
   * @return a name for the specified parameter.
   *
   * @throws IllegalArgumentException if <tt>param</tt> is invalid.
   */
  public String[] getParameterNames();

  /**
   * Return the current values of the parameters. The class of the returned
   * object depends of the type of the parameter :
   *
   * <ul>
   * <li>
   * For CDATA, it may be String, Integer, Long, etc...
   * </li>
   * <li>
   * For CHECK, it MUST be a Boolean
   * </li>
   * <li>
   * For SELECT, it MUST be an Integer, whose value matches the index of the
   * current value.
   * </li>
   * </ul>
   *
   *
   * @return the current values of the parameters in an array.
   */
  public Object[] getParameterValues();

  /**
   * Return the types of the parameters.
   *
   * @return the types of the parameters in an array.
   *
   * @see #CDATA
   * @see #SELECT
   * @see #CHECK
   */
  public byte[] getParameterTypes();

  /**
   * If the specified parameter has the type <tt>SELECT</tt>, this method must
   * return an array of possible values for this parameter.
   *
   * <p>
   * <b>Precondition</b>: <code>0 &lt;= param &lt; getParameterCount()</code>
   * </p>
   *
   * @param param a parameter number.
   *
   * @return an array of possible values for this parameter.
   *
   * @throws IllegalArgumentException if <tt>param</tt> is invalid, or if the
   *         parameter identified by <tt>param</tt> has not the type
   *         <tt>SELECT</tt>
   *
   * @see #SELECT
   */
  public String[] getPossibleValues(int param) throws IllegalArgumentException;

  /**
   * Test if <tt>newValue</tt> can be assigned as a new value for the
   * parameter specified by <tt>num</tt>. If all values are correct, the
   * values must be changed.
   *
   * <p>
   * <b>Precondition</b> : <tt>0 &lt;= newValues.length &lt;
   * getParameterCount()</tt>
   * </p>
   *
   * @param newValues an array containing candidate new values
   *
   * @return an array of <tt>boolean</tt>, whose size is the size of
   *         <tt>newValues</tt>, where a <tt>true</tt> means the value was
   *         correct.
   *
   * @throws IllegalArgumentException if <tt>newValues</tt> is invalid.
   */
  public boolean[] setNewValues(Object[] newValues)
    throws IllegalArgumentException;
}


// End Configurable.java
