/*
 * Database object, configurable with XML.
 *
 * Class      br.org.scadabr.luciol.database.Database
 * File       Database.java
 * Author     Alexis BIETTI <alexis.bietti@sysaware.com>
 * (c)        SysAware <http://www.sysaware.com>
 *
 */

//-----------------------------------------------------------------------------
// Imports
//-----------------------------------------------------------------------------
package br.org.scadabr.luciol.database;

import java.util.Enumeration;
import java.util.Hashtable;

import br.org.scadabr.luciol.Configurable;
import br.org.scadabr.luciol.ihm.Labels;
import br.org.scadabr.protocol.lib.xml.KXMLElement;
import br.org.scadabr.protocol.util.ArrayUtils;

//-----------------------------------------------------------------------------
// Class
//-----------------------------------------------------------------------------

/**
 * Database class. A database contains DataElements.
 * <p>
 * At any time, there are always two databases :
 * 
 * <ul>
 * <li>The {@link #ACTIVE} database is to be used by apps.
 * <li>The {@link #PASSIVE} database if to be used by configurator.
 * </ul>
 * <p>
 * A <b>commit</b> copies the passive database into the active database, except
 * for data records, that are thus lost.
 * <p>
 * A <b>roll back</b> copies the active database into the passive database.
 * 
 * @author <a href="mailto:alexis.bietti@sysaware.com">Alexis BIETTI</a>
 * @version 1.0
 * 
 * @see br.org.scadabr.luciol.database.DataElement
 */
public class Database implements Configurable, DatabaseFeatures {
	// ---------------------------------------------------------------------------
	// Constants
	// ---------------------------------------------------------------------------

	/* XML names used in configuration */
	private static final String XML_DATABASE = "Database";
	private static final String XML_SHORT_TIME = "shortTime";
	private static final String XML_LONG_TIME = "longTime";

	/** Print debug info on <tt>System.out</tt> if true. */
	public static boolean DATABASE_DEBUG = false;

	/** Option names. */
	private static final String[] OPTION_NAMES = new String[] {
			Labels.DATABASE_DO_LONG_MAINTAIN_TIME,
			Labels.DATABASE_DO_SHORT_MAINTAIN_TIME };

	// ---------------------------------------------------------------------------
	// Attributes
	// ---------------------------------------------------------------------------

	/** Contains all data elements of this database. */
	private Hashtable dataElements = new Hashtable();

	/** Time stamp : the date when this database was created or commited. */
	private long timeStamp = System.currentTimeMillis();

	/** Long maintain time of digital outputs (in seconds) */
	private int longTime = 1;

	/** Short maintain time of digital outputs (in milliseconds) */
	private int shortTime = 20;

	// Meta information about the configuration file
	private String author = Labels.UNKNOWN;
	private String date = Labels.UNKNOWN;
	private String version = Labels.UNKNOWN;
	private String fileId = Labels.UNKNOWN;

	// ---------------------------------------------------------------------------
	// Constructor
	// ---------------------------------------------------------------------------

	/**
	 * Constructor.
	 */
	public Database() {
	}

	/**
	 * Print a debug message on <code>System.out</code>
	 * 
	 * @param o
	 * 
	 * @see #DATABASE_DEBUG
	 */
	private void log(Object o) {
		if (DEBUG) {
			System.out.print("[Database] ");
			System.out.println(o);
		}
	}

	// ---------------------------------------------------------------------------
	// Methods
	// ---------------------------------------------------------------------------

	/**
	 * Get a DataElement specified by the give database ID.
	 * 
	 * @param databaseID
	 *            a database ID.
	 * 
	 * @return a DataElement, or <tt> null </tt> if not found.
	 */
	public DataElement getDataElement(int databaseID) {
		return (DataElement) dataElements.get(new Integer(databaseID));
	}

	/**
	 * Return all the {@link DataElement}s as an {@link java.util.Enumeration}.
	 * 
	 * @return all the {@link DataElement}s as an {@link java.util.Enumeration}.
	 */
	public Enumeration<?> elements() {
		return dataElements.elements();
	}

	/**
	 * Add an element in this database with the specified ID. If the specified
	 * ID is already assigned, the old element is discarded.
	 * 
	 * @param element
	 *            a DataElement.
	 * @param ID
	 *            a unique database identifier.
	 */
	public void addElement(DataElement element, int ID) {
		element.databaseID = ID;
		addElement(element);
	}

	/**
	 * Add an element in this database. The element contains it database ID.
	 * 
	 * @param element
	 *            a DataElement.
	 */
	public void addElement(DataElement element) {
		dataElements.put(new Integer(element.databaseID), element);
	}

	/**
	 * Return the digital output short time in millis.
	 * 
	 * @return the digital output short time in millis.
	 */
	public long shortTime() {
		return shortTime;
	}

	/**
	 * Return the digital output long time in millis.
	 * 
	 * @return the digital output long time in millis.
	 */
	public long longTime() {
		return longTime * 1000;
	}

	// ---------------------------------------------------------------------------
	// Interface XMLInterface
	// ---------------------------------------------------------------------------

	/**
	 * See the general contract of this method in the interface
	 * {@link br.org.scadabr.io.XMLInterface}
	 */
	public void inputFromXML(KXMLElement xml) throws IllegalArgumentException {
		try {
			// Parse attributes
			shortTime = xml.getProperty(XML_SHORT_TIME, 20);
			longTime = xml.getProperty(XML_LONG_TIME, 1);
			author = xml.getProperty("author", Labels.UNKNOWN);
			date = xml.getProperty("date", Labels.UNKNOWN);
			version = xml.getProperty("version", Labels.UNKNOWN);
			fileId = xml.getProperty("fileId", Labels.UNKNOWN);

			// Clear the current configuration
			dataElements.clear();

			for (Enumeration<?> children = xml.enumerateChildren(); children
					.hasMoreElements();) {
				KXMLElement child = (KXMLElement) children.nextElement();
				Class<?> clazz = (Class<?>) DataElement.CLASSES.get(child
						.getTagName());

				if (clazz == null) {
					if (DEBUG) {
						log("XML: Unknown data element");
					}

					// Unknown element, discard silently and pass to the next...
				} else {
					// Instanciate a new DataElement
					DataElement elem = (DataElement) clazz.newInstance();
					elem.inputFromXML(child);
					addElement(elem);
				}
			}
		} catch (Exception e) {
			// If there is any problem, let's discard the whole thing and warn
			// the
			// upper level with an exception
			e.printStackTrace();
			throw new IllegalArgumentException(e.toString());
		}
	}

	/**
	 * See the general contract of this method in the interface
	 * {@link br.org.scadabr.io.XMLInterface}
	 */
	public KXMLElement outputToXML() {
		KXMLElement root = new KXMLElement();
		int size = dataElements.size();

		root.setTagName(XML_DATABASE);
		root.addProperty(XML_SHORT_TIME, shortTime);
		root.addProperty(XML_LONG_TIME, longTime);
		root.addProperty("author", author);
		root.addProperty("date", date);
		root.addProperty("version", version);
		root.addProperty("fileId", fileId);

		for (Enumeration<?> keys = dataElements.keys(); keys.hasMoreElements();) {
			DataElement el = getDataElement(((Integer) keys.nextElement())
					.intValue());

			if (el == null) {
				continue;
			}

			root.addChild(el.outputToXML());
		}

		return root;
	}

	// ---------------------------------------------------------------------------
	// Interface Configurable
	// ---------------------------------------------------------------------------

	/**
	 * See the general contract of this method in the interface
	 * {@link br.org.scadabr.luciol.Configurable}
	 */
	public String applicationName() {
		return Labels.APP_DATABASE;
	}

	/**
	 * DOCUMENT ME!
	 * 
	 * @return DOCUMENT ME!
	 */
	public String getAuthor() {
		return author;
	}

	/**
	 * DOCUMENT ME!
	 * 
	 * @return DOCUMENT ME!
	 */
	public String getVersion() {
		return version;
	}

	/**
	 * DOCUMENT ME!
	 * 
	 * @return DOCUMENT ME!
	 */
	public String getDate() {
		return date;
	}

	/**
	 * DOCUMENT ME!
	 * 
	 * @return DOCUMENT ME!
	 */
	public String getFileId() {
		return fileId;
	}

	/**
	 * See the general contract of this method in the interface
	 * {@link br.org.scadabr.luciol.Configurable}
	 */
	public int getParameterCount() {
		return 2;
	}

	/**
	 * See the general contract of this method in the interface
	 * {@link br.org.scadabr.luciol.Configurable}
	 */
	public String[] getParameterNames() {
		return OPTION_NAMES;
	}

	/**
	 * See the general contract of this method in the interface
	 * {@link br.org.scadabr.luciol.Configurable}
	 */
	public Object[] getParameterValues() {
		return new String[] { Integer.toString(longTime),
				Integer.toString(shortTime) };
	}

	/**
	 * See the general contract of this method in the interface
	 * {@link br.org.scadabr.luciol.Configurable}
	 */
	public byte[] getParameterTypes() {
		return new byte[] { CDATA, CDATA };
	}

	/**
	 * See the general contract of this method in the interface
	 * {@link br.org.scadabr.luciol.Configurable}
	 */
	public String[] getPossibleValues(int param)
			throws IllegalArgumentException {
		throw new IllegalArgumentException();
	}

	/**
	 * See the general contract of this method in the interface
	 * {@link br.org.scadabr.luciol.Configurable}
	 */
	public boolean[] setNewValues(Object[] newValues)
			throws IllegalArgumentException {
		if (newValues.length != 2) {
			throw new IllegalArgumentException();
		}

		boolean[] res = new boolean[2]; // Result
		int newLongTime = Integer.parseInt((String) newValues[0]);
		int newShortTime = Integer.parseInt((String) newValues[1]);

		// Check the new long time
		if ((newLongTime >= 1) && (newLongTime <= 30)) {
			res[0] = true;
		}

		// Check the new short time
		if ((newShortTime >= 20) && (newShortTime <= 200)
				&& ((newShortTime % 20) == 0)) {
			res[1] = true;
		}

		// Commit the new values if they're all good
		if (ArrayUtils.and(res)) {
			longTime = newLongTime;
			shortTime = newShortTime;
		}

		return res;
	}

	// ---------------------------------------------------------------------------
	// Static helpers
	// ---------------------------------------------------------------------------

	/**
	 * Copy all DataElements of <tt>src</tt> into <tt>dest</tt>.
	 * 
	 * @param src
	 *            the source Database
	 * @param dest
	 *            the destination Database
	 */
	protected static void copyInto(Database src, Database dest)
			throws DatabaseException {
		// Empty the notification queue
		while (DatabaseApp.toNotify.dequeue() != null) {
			;
		}

		// Don't worry, be happy :)
		dest.inputFromXML(src.outputToXML());
	}

	// ---------------------------------------------------------------------------
	// Testing
	// ---------------------------------------------------------------------------
	public static void main(String[] args) {
		String FILE = "<Database shortTime='200' longTime='1'>"
				+ "<!-- Analog Output Card -->"
				+ "<AnalogOutput id='0' desc='Luciol - Analog Output 1'/>"
				+ "<AnalogOutput id='1' desc='Luciol - Analog Output 2'/>"
				+ "<AnalogOutput id='2' desc='Luciol - Analog Output 3'/>"
				+ "<AnalogOutput id='3' desc='Luciol - Analog Output 4'/>"
				+ "<AnalogInput id='4' desc='Luciol - Analog Input 1' law='None' />"
				+ "<AnalogInput id='5' desc='Luciol - Analog Input 2' law='None'/>"
				+ "<AnalogInput id='6' desc='Luciol - Analog Input 3' law='None'/>"
				+ "<AnalogInput id='7' desc='Luciol - Analog Input 4' law='None'/>"
				+ "<DigitalInput id='8' desc='Luciol - Digital Input 1' logic='1'/>"
				+ "<DigitalInput id='9' desc='Luciol - Digital Input 2' logic='1'/>"
				+ "<DigitalInput id='10' desc='Luciol - Digital Input 3' logic='1'/>"
				+ "<DigitalInput id='11' desc='Luciol - Digital Input 4' logic='1'/>"
				+ "<DigitalInput id='12' desc='Luciol - Digital Input 5' logic='1'/>"
				+ "<DigitalInput id='13' desc='Luciol - Digital Input 6' logic='1'/>"
				+ "<DigitalInput id='14' desc='Luciol - Digital Input 7' logic='1'/>"
				+ "<DigitalInput id='15' desc='Luciol - Digital Input 8' logic='1'/>"
				+ "</Database>";
		Database db = new Database();
		KXMLElement xml = new KXMLElement();
		xml.parseString(FILE);
		db.inputFromXML(xml);

		System.out.println(db.outputToXML());
	}
}

// End Database.java
