/*
 * Entry point of the database application.
 *
 * Class      br.org.scadabr.luciol.database.DatabaseApp
 * File       DatabaseApp.java
 * Author     Alexis BIETTI <alexis.bietti@sysaware.com>
 * (c)        SysAware <http://www.sysaware.com>
 */

//-----------------------------------------------------------------------------
// Imports
//-----------------------------------------------------------------------------
package br.org.scadabr.luciol.database;

import br.org.scadabr.luciol.AppSession;
import br.org.scadabr.luciol.Application;
import br.org.scadabr.luciol.Configurable;
import br.org.scadabr.luciol.ihm.Labels;

import br.org.scadabr.protocol.util.FifoQueue;



//-----------------------------------------------------------------------------
// Class
//-----------------------------------------------------------------------------

/**
 * Entry point of the database application.
 *
 * @author <a href="mailto:alexis.bietti@sysaware.com">Alexis BIETTI</a>
 */
public class DatabaseApp implements Application, AppSession, Runnable,
  DatabaseFeatures {
  //---------------------------------------------------------------------------
  // Class Attributes
  //---------------------------------------------------------------------------

  /** Active database. */
  public static final Database ACTIVE = new Database();

  /** Passive database. */
  public static final Database PASSIVE = new Database();

  /** The only instance of this class allowed. */
  private static final DatabaseApp INST = new DatabaseApp();

  /** Queue of pending notifications. */
  protected static FifoQueue toNotify = new FifoQueue(100);

  //---------------------------------------------------------------------------
  // Constructor
  //---------------------------------------------------------------------------

  /**
   * Don't let anyone instanciate this class.
   */
  private DatabaseApp() {
    // Start the notification monitor
    if (DEBUG) {
      log("Start the notifier thread");
    }

    startNotifier();
  }

  /**
   * Print a debug message on System.out
   *
   * @param o
   */
  private static void log(Object o) {
    if (DEBUG) {
      System.out.print("[DatabaseApp] ");
      System.out.println(o);
    }
  }

  /**
   * DOCUMENT ME!
   */
  private void startNotifier() {
    Thread t =
      new Thread(this) {
        public String toString() {
          return "DbNotifier";
        }
      };

    t.start();
  }

  //---------------------------------------------------------------------------
  // Static methods
  //---------------------------------------------------------------------------

  /**
   * Return an instance of the database application.
   */
  public static DatabaseApp getInstance() {
    return INST;
  }

  //---------------------------------------------------------------------------
  // Interface Runnable
  //---------------------------------------------------------------------------

  /**
   * This thread monitors the {@link #toNotify} queue. New {@link
   * DataElement}s are appended to the queue when they are updated.
   */
  public void run() {
    try {
      DataElement pending = null;

      while (true) {
        pending = (DataElement) toNotify.blockingDequeue();

        if (DEBUG) {
          log("Notify listeners of " + pending);
        }

        pending.notifyAllListeners();
      }
    } catch (Exception e) {
      System.out.print("[DatabaseApp] Exception: ");
      System.out.println(e);
    } catch (Throwable t) {
      System.out.print("[DatabaseApp] Error: ");
      System.out.println(t);

      // throw t;
    } finally {
      startNotifier();
    }
  }

  //---------------------------------------------------------------------------
  // Interface Application
  //---------------------------------------------------------------------------

  /**
   * 
   */
  public String getVersion() {
    return "1.0";
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public String getDate() {
    return "03.II.2003";
  }

  /**
   *
   */
  public String applicationName() {
    return Labels.APP_DATABASE;
  }

  /**
   * <p>
   * One session only.
   * </p>
   */
  public int numberOfSessionsAllowed() {
    return 1;
  }

  /**
   * <p>
   * This method does nothing for us because the Database application is
   * lauched at starting and never stopped.
   * </p>
   */
  public int openNewSession() {
    return -1;
  }

  /**
   * <p>
   * One session only.
   * </p>
   */
  public int numberOfCurrentSessions() {
    return 1;
  }

  /**
   * <p>
   * This method does nothing for us because the Database application must
   * never be stopped.
   * </p>
   */
  public void deleteAllSessions() {
  }

  /**
   * <p>
   * Return ourselves as we are also the AppSession.
   * </p>
   */
  public AppSession getSession(int session) {
    if (session == 0) {
      return this;
    }

    return null;
  }

  //---------------------------------------------------------------------------
  // Interface AppSession
  //---------------------------------------------------------------------------

  /**
   * <p>
   * Get the application of this session.
   * </p>
   */
  public Application getApplication() {
    // We are our own Application
    return this;
  }

  /**
   * <p>
   * Return the active configuration of this application.
   * </p>
   */
  public Configurable activeConfiguration() {
    return ACTIVE;
  }

  /**
   * <p>
   * Return the passive configuration of this application.
   * </p>
   */
  public Configurable passiveConfiguration() {
    return PASSIVE;
  }

  /**
   * <p>
   * Commit the configuration.
   * </p>
   *
   * <p>
   * The general contract of COMMIT is to copy the passive configuration into
   * the active configuration.
   * </p>
   */
  public void commit() {
    Database.copyInto(PASSIVE, ACTIVE);
  }

  /**
   * <p>
   * Roll back the configuration.
   * </p>
   *
   * <p>
   * The general contract of ROLL BACK is to copy the active configuration
   * into the passive configuration.
   * </p>
   */
  public void rollBack() {
    Database.copyInto(ACTIVE, PASSIVE);
  }

  //---------------------------------------------------------------------------
  // Testing
  //---------------------------------------------------------------------------

  /**
   * Testing.
   */
  public static void main(String[] args) {
    for (int i = 0; i < 8; ++i) {
      DigitalInput  in  = new DigitalInput();
      DigitalOutput out = new DigitalOutput();

      in.description    = "Entree TOR Luciol #" + (i + 1);
      out.description   = "Sortie TOR Luciol #" + (i + 1);

      PASSIVE.addElement(in, i);
      PASSIVE.addElement(out, i + 8);
    }

    for (int i = 0; i < 4; ++i) {
      AnalogInput  in  = new AnalogInput();
      AnalogOutput out = new AnalogOutput();

      in.description    = "Entree ANA Luciol #" + (i + 1);
      out.description   = "Sortie ANA Luciol #" + (i + 1);

      PASSIVE.addElement(in, i + 16);
      PASSIVE.addElement(out, i + 20);
    }

    DoubleInput din = new DoubleInput();
    PASSIVE.addElement(din, 25);

    System.out.println(PASSIVE.outputToXML());
  }
}


// End DatabaseApp.java
