/*
 * Database for Luciol. Object element.
 *
 * Class
 * File
 * Author    Alexis BIETTI <alexis.bietti@sysaware.com>
 * (c)       SysAware <http://www.sysaware.com/>
 */
package br.org.scadabr.luciol.database;

import java.util.Enumeration;
import java.util.Hashtable;

import br.org.scadabr.protocol.lib.xml.KXMLElement;
import br.org.scadabr.protocol.util.Properties;

/**
 * <p>
 * An {@link ObjectElement} is an aggregation of several DataElement, and some
 * properties.
 * </p>
 *
 * <p>
 * For instance, a protection could be represented as an {@link ObjectElement}
 * containing several measures, a trip command, a close command, and a state.
 * As properties, it would have a name, a vendor, a serial id and a type.
 * </p>
 *
 * <p>
 * Examples of XML representation of some {@link ObjectElement}s:
 * </p>
 *
 * <p>
 * <code>&lt;ObjectElement id="1" desc="Protection"&gt;<br>
 * &nbsp; &lt; Property name="vendor" value="Alstom" /&gt;<br>
 * &nbsp; &lt;Property name="model" value="P124" /&gt;<br>
 * &nbsp; &lt;Data name="IA Measure" ref="0" /&gt;<br>
 * &nbsp; &lt;Data name="IB Measure" ref="1" /&gt;<br>
 * &nbsp; &lt;Data name="IC Measure" ref="2" /&gt;<br>
 * &nbsp; &lt;Data name="Trip command" ref="2" /&gt; <br>
 * &nbsp; &lt;Data name="Close command" ref="2" /&gt;<br>
 * &lt; /ObjectElement&gt; </code>
 * </p>
 *
 * @author <a href="mailto:alexis.bietti@sysaware.com"> Alexis BIETTI </a>
 * @version 1.0 - 5 mars 03
 */
public class ObjectElement extends DataElement {
  //---------------------------------------------------------------------------
  // Attributes
  //---------------------------------------------------------------------------

  /** Mapping of data elements (measures, commands, points) */
  public Hashtable name2id = new Hashtable();

  /** Properties (name, vendor, serial, etc.) */
  public Properties props = new Properties();

  //---------------------------------------------------------------------------
  // Constructor
  //---------------------------------------------------------------------------

  /**
   * <p>
   * Construct a new {@link ObjectElement} object.
   * </p>
   */
  public ObjectElement() {
    super();
  }

  //---------------------------------------------------------------------------
  // Methods
  //---------------------------------------------------------------------------

  /**
   * @param name
   * @param id
   */
  public void addData(String name, int id) {
    name2id.put(name, new Integer(id));
  }

  /**
   * @param name
   *
   * @return int
   */
  public int getComponentId(String name) {
    Object o = name2id.get(name);

    if (o == null) {
      return -1;
    } else if (o instanceof Integer) {
      return o.hashCode();
    } else {
      return -1;
    }
  }

  /**
   * <p>
   * Get a component of this object, according to the current active database.
   * </p>
   *
   * @return DataElement
   */
  public DataElement getComponent(String name) {
    int id = getComponentId(name);

    return DatabaseApp.ACTIVE.getDataElement(id);
  }

  /**
   * <p>
   * Enumerate the names of this object's components
   * </p>
   *
   * @return Enumeration
   */
  public Enumeration dataNames() {
    return name2id.elements();
  }

  /**
   * <p>
   * Return the whole set of {@link br.org.scadabr.util.Properties} of this
   * element
   * </p>
   *
   * @return Properties
   */
  public Properties properties() {
    return props;
  }

  /**
   * <p>
   * Copy this element with a mapping and properties similar to this object's
   * mapping and properties.
   * </p>
   *
   * @return an empty copy of this element.
   */
  public ObjectElement templateCopy() {
    ObjectElement o = new ObjectElement();

    for (Enumeration e = name2id.elements(); e.hasMoreElements();) {
      Object name = e.nextElement();
      o.name2id.put(name, name2id.get(name));
    }

    for (Enumeration e = props.elements(); e.hasMoreElements();) {
      Object name = e.nextElement();
      o.props.put(name, props.get(name));
    }

    return o;
  }

  /**
   * @see br.org.scadabr.luciol.database.DataElement#defaultValue()
   */
  public Object defaultValue() {
    return null;
  }

  /**
   * @see br.org.scadabr.luciol.database.DataElement#getDefaultLogSize()
   */
  public int getDefaultLogSize() {
    return 0;
  }

  /**
   * @see br.org.scadabr.luciol.database.DataElement#getName()
   */
  public String getName() {
    return "ObjectElement";
  }

  /**
   * @see br.org.scadabr.luciol.database.DataElement#instanciateRecord()
   */
  public Record instanciateRecord() {
    return null;
  }

  /**
   * @see br.org.scadabr.luciol.database.DataElement#servlet()
   */
  public String servlet() {
    return null;
  }

  /**
   * @see br.org.scadabr.io.XMLInterface#inputFromXML(KXMLElement)
   */
  public void inputFromXML(KXMLElement xml) throws IllegalArgumentException {
    super.inputFromXML(xml);
  }

  /**
   * @see br.org.scadabr.io.XMLInterface#outputToXML()
   */
  public KXMLElement outputToXML() {
    KXMLElement xml   = super.outputToXML();
    KXMLElement child = null;
    String      name  = null;

    for (Enumeration e = props.elements(); e.hasMoreElements();) {
      name   = (String) e.nextElement();

      child = new KXMLElement();
      child.setTagName("Property");
      child.addProperty("name", name);
      child.addProperty("value", props.get(name));

      xml.addChild(child);
    }

    for (Enumeration e = name2id.elements(); e.hasMoreElements();) {
      name   = (String) e.nextElement();

      child = new KXMLElement();
      child.setTagName("Component");
      child.addProperty("name", name);
      child.addProperty("id", name2id.get(name));

      xml.addChild(child);
    }

    return xml;
  }
}


// End ObjectElement.java
