/*
 * Record is an abstract class that represent a unit of information in the
 * database. This may be a measure, a status, or a command state. A record has
 * a time stamp. A record also have some customizable flags that can contain
 * information about the quality of the information contained in this record.
 *
 * Class     br.org.scadabr.luciol.database.Record
 * Hidden    br.org.scadabr.luciol.database.DigitalRecord
 * Hidden    br.org.scadabr.luciol.database.IntegerRecord
 * Hidden    br.org.scadabr.luciol.database.AnalogRecord
 * Hidden    br.org.scadabr.luciol.database.FloatRecord
 * File      Record.java
 * Author    Alexis BIETTI <alexis. bietti@sysaware.com>
 * (c)       SysAware <http://www.sysaware.com>
 */

//-----------------------------------------------------------------------------
// Imports
//-----------------------------------------------------------------------------
package br.org.scadabr.luciol.database;

import java.util.Vector;

import br.org.scadabr.protocol.lib.math.Float;
import br.org.scadabr.protocol.lib.math.Math32FP;

//-----------------------------------------------------------------------------
// Classes
//-----------------------------------------------------------------------------

/**
 * Record is an abstract class that represent a unit of information in the
 * database.
 *
 * <p>
 * This may be a measure, a status, or a command state. A record has a time
 * stamp. A record also have some customizable flags that can contain
 * information about the quality of the information contained in this record.
 * </p>
 *
 * @author <a href="mailto:alexis.bietti@sysaware.com"> Alexis BIETTI </a>
 */
public abstract class Record {
  //---------------------------------------------------------------------------
  // Constants
  //---------------------------------------------------------------------------

  /*
   * Types of Records.
   */
  protected static final byte REC_DIGITAL = 0;
  protected static final byte REC_ANALOG  = 1;
  protected static final byte REC_INTEGER = 2;
  protected static final byte REC_FLOAT   = 3;

  /*
   * Quality flags defined in UCA2
   */
  public static final short Q_INVALID    = 0x0002;
  public static final short Q_COMM_FAIL  = 0x0004;
  public static final short Q_FORCED     = 0x0008;
  public static final short Q_OVER_RANGE = 0x0010;
  public static final short Q_BAD_REF    = 0x0020;

  /*
   * Quality flags defined internally for Luciol applications
   */
  public static final short Q_NOT_COMPL           = 0x0100;
  public static final short Q_NOT_REFRESHED       = 0x0200;
  public static final short Q_FUNCT_HARDW_FAILURE = 0x0400;
  public static final short Q_DENIAL_OF_SERVICE   = 0x0800;

  /*
   * Quality flags reserved for DNP outputs
   */

  // public static final short Q_DNP_COMMAND = 1 << 9;
  //---------------------------------------------------------------------------
  // Attributes
  //---------------------------------------------------------------------------

  /**
   * <p>
   * Quality flags of this Record. Users should take care before modifying
   * this value.
   * </p>
   *
   * <p></p>
   */
  public short quality = 0;

  /**
   * Time stamp of this record. Users should take care before modifying this
   * value.
   */
  public long timeStamp = 0L;

  //---------------------------------------------------------------------------
  // Methods
  //---------------------------------------------------------------------------

  /**
   * Get the time stamp of this record.
   *
   * @return the time stamp of this record.
   */
  public long getTimeStamp() {
    return timeStamp;
  }

  /**
   * Set the time stamp of this record to the current system time.
   */
  public void setTimeStamp() {
    timeStamp = System.currentTimeMillis();
  }

  /**
   * Set the time stamp of this record to the given system time.
   *
   * @param aTimeStamp a system time.
   */
  public void setTimeStamp(long aTimeStamp) {
    timeStamp = aTimeStamp;
  }

  /**
   * Get the quality flags of this record.
   *
   * @return the quality flags of this record.
   */
  public short getQuality() {
    return quality;
  }

  /**
   * Set the quality flags of this record to the given value.
   *
   * @param aQuality a new quality flags for this record.
   */
  public void setQuality(short aQuality) {
    quality = aQuality;
  }

  /**
   * Test if this record is valid. A valid record has all quality flags set to
   * zero.
   *
   * @return <code> true </code> if this record is valid.
   */
  public boolean isValid() {
    return quality == 0;
  }

  //---------------------------------------------------------------------------
  // Abstract Methods
  //---------------------------------------------------------------------------

  /**
   * Get the value of this record cast as a boolean.
   *
   * @return the value of this record cast as a boolean.
   */
  public abstract boolean booleanValue();

  /**
   * Get the value of this record cast as an integer.
   *
   * @return the value of this record cast as an integer.
   */
  public abstract int intValue();

  /**
   * Get the value of this record cast as a fixed point decimal.
   *
   * @return the value of this record cast as a fixed point decimal.
   *
   * @see br.org.scadabr.math.Math32FP
   */
  public abstract int fixedPointValue();

  /**
   * Get the value of this record cast as a {@link java.lang.String}
   *
   * @return the value of this record cast as a {@link java.lang.String}.
   */
  public abstract String stringValue();

  /**
   * Get the value of this record as an {@link java.lang.Object}
   *
   * <p>
   * The class of this object depends on the type of this record.
   * </p>
   *
   * <p>
   * Implementation note: this method will become useful when we will have
   * complex object types.
   * </p>
   *
   * @return the value of this record as an Object.
   */
  public abstract Object objectValue();

  /**
   * Get the value of this record as a float.
   *
   * @return the value of this record cast as a String.
   */
  public abstract float floatValue();

  /**
   * Change the value of this record with the given value.
   *
   * <p>
   * The <code> newValue </code> may be accepted or refused
   * </p>
   *
   * @param newValue the new value for this Record
   *
   * @throws DatabaseException if the class of newValue is not compliant.
   */
  public abstract void setValue(Object newValue);

  /**
   * Try and set the value of this Record with the given value.
   *
   * @param newValue the new value for this Record
   *
   * @throws DatabaseException if the class of newValue is not compliant.
   */
  public void setValue(boolean newValue) {
    setValue(new Boolean(newValue));
  }

  /**
   * Try and set the value of this Record with the given value.
   *
   * @param newValue the new value for this Record
   *
   * @throws DatabaseException if the class of newValue is not compliant.
   */
  public void setValue(int newValue) {
    setValue(new Integer(newValue));
  }

  /**
   * Try and set the value of this Record with the given value.
   *
   * @param newValue the new value for this Record
   *
   * @throws DatabaseException if the class of newValue is not compliant.
   */
  public void setValue(float newValue) {
    setValue(new Float(newValue));
  }

  /**
   * Allocate a new Record object and copy all our attributes in this new
   * Record
   *
   * @return a newly allocated Record
   */
  public abstract Record copy();

  /**
   * Instanciate a new Record compliant with the specified type.
   *
   * @see #REC_DIGITAL
   * @see #REC_ANALOG
   * @see #REC_INTEGER
   * @see #REC_FLOAT
   */
  protected static Record instanciateRecord(byte kind)
    throws IllegalArgumentException {
    if (kind == REC_DIGITAL) {
      return new DigitalRecord();
    }

    if (kind == REC_ANALOG) {
      return new AnalogRecord();
    }

    if (kind == REC_INTEGER) {
      return new IntegerRecord();
    }

    if (kind == REC_FLOAT) {
      return new FloatRecord();
    } else {
      throw new IllegalArgumentException();
    }
  }

  /**
   * Get a string representation of this Record.
   */
  public String toString() {
    return "{Record value='" + stringValue() + "',quality="
    + Integer.toBinaryString(quality) + ",date=" + timeStamp + "}";
  }

  /**
   * This method takes all quality information contained in the quality field
   * and compute it as a set of printable strings.
   *
   * @return a vector containing all quality messages.
   */
  public Vector qualityMessages() {
    Vector aVector = new Vector();

    // Add a message for each bit set.
    if ((quality & Q_INVALID) != 0) {
      aVector.addElement("Invalid");
    }

    if ((quality & Q_BAD_REF) != 0) {
      aVector.addElement("Bad reference");
    }

    if ((quality & Q_COMM_FAIL) != 0) {
      aVector.addElement("Communication Failed");
    }

    if ((quality & Q_FORCED) != 0) {
      aVector.addElement("Value forced");
    }

    if ((quality & Q_OVER_RANGE) != 0) {
      aVector.addElement("Over range");
    }

    if ((quality & Q_NOT_COMPL) != 0) {
      aVector.addElement("Not complementary");
    }

    aVector.trimToSize();

    return aVector;
  }
}


//-----------------------------------------------------------------------------
// End public class
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// Hidden classes
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// DigitalRecord Class
//-----------------------------------------------------------------------------

/**
 * A digital record stores boolean values.
 *
 * @author Alexis BIETTI
 */
class DigitalRecord extends Record {
  /** Inner value of this Record. */
  boolean value = true;

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public Record copy() {
    DigitalRecord copy = new DigitalRecord();
    copy.value       = value;
    copy.timeStamp   = timeStamp;
    copy.quality     = quality;

    return copy;
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public boolean booleanValue() {
    return value;
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public int intValue() {
    return value ? 1 : 0;
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public int fixedPointValue() {
    return value ? Math32FP.ONE : 0;
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public String stringValue() {
    return value ? "true" : "false";
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public Object objectValue() {
    return new Boolean(value);
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public float floatValue() {
    return value ? 1 : 0;
  }

  /**
   * DOCUMENT ME!
   *
   * @param newValue DOCUMENT ME!
   */
  public void setValue(Object newValue) {
    if (newValue instanceof Boolean) {
      value = ((Boolean) newValue).booleanValue();
    } else if (newValue instanceof Integer) {
      value = (((Integer) newValue).intValue() != 0);
    } else {
      try {
        value = Float.parseFloat(newValue.toString()) != 0;
      } catch (Exception e) {
        value = (newValue != null);
      }
    }
  }
}


//-----------------------------------------------------------------------------
// FloatRecord Class
//-----------------------------------------------------------------------------

/**
 * A float record stores floating point values.
 *
 * @author <a href="mailto:alexis.bietti@sysaware.com">Alexis BIETTI</a>
 */
class FloatRecord extends Record {
  /** Inner value of this Record. */
  float floatValue = 0;

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public Record copy() {
    FloatRecord copy = new FloatRecord();
    copy.floatValue   = floatValue;
    copy.timeStamp    = timeStamp;
    copy.quality      = quality;

    return copy;
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public boolean booleanValue() {
    return (floatValue != 0);
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public int intValue() {
    return (int) floatValue;
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public int fixedPointValue() {
    return Math32FP.toFP(floatValue);
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public String stringValue() {
    return Float.toString(floatValue);
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public Object objectValue() {
    return new Float(floatValue);
  }

  /**
       *
       */
  public float floatValue() {
    return floatValue;
  }

  /**
       *
       */
  public void setValue(Object newValue) {
    if (newValue instanceof Float) {
      floatValue = ((Float) newValue).floatValue();
    } else if (newValue instanceof Math32FP) {
      floatValue = ((Math32FP) newValue).floatValue();
    } else if (newValue instanceof Integer) {
      floatValue = (float) ((Integer) newValue).intValue();
    } else if (newValue instanceof Boolean) {
      floatValue = ((Boolean) newValue).booleanValue() ? 1 : 0;
    } else if (newValue instanceof String) {
      floatValue = Float.parseFloat((String) newValue);
    }
  }
}


//-----------------------------------------------------------------------------
// AnalogRecord Class
//-----------------------------------------------------------------------------

/**
 * An analog record stores fixed point values.
 *
 * @author Alexis BIETTI
 */
class AnalogRecord extends Record {
  /** Inner value of this record. */
  int fpValue = 0;

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public Record copy() {
    AnalogRecord copy = new AnalogRecord();
    copy.fpValue     = fpValue;
    copy.timeStamp   = timeStamp;
    copy.quality     = quality;

    return copy;
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public boolean booleanValue() {
    return fpValue != 0;
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public int intValue() {
    return Math32FP.toInt(fpValue);
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public int fixedPointValue() {
    return fpValue;
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public String stringValue() {
    return Math32FP.toString(fpValue);
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public Object objectValue() {
    return new Math32FP(fpValue);
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public float floatValue() {
    return Math32FP.toFloat(fpValue);
  }

  /**
   * DOCUMENT ME!
   *
   * @param newValue DOCUMENT ME!
   */
  public void setValue(Object newValue) {
    if (newValue instanceof Math32FP) {
      fpValue = ((Math32FP) newValue).fpValue();
    } else if (newValue instanceof Integer) {
      fpValue = Math32FP.toFP(((Integer) newValue).intValue());
    } else if (newValue instanceof Short) {
      fpValue = Math32FP.toFP(((Short) newValue).shortValue());
    } else if (newValue instanceof Byte) {
      fpValue = Math32FP.toFP(((Byte) newValue).byteValue());
    } else if (newValue instanceof String) {
      fpValue = Math32FP.toFP((String) newValue);
    } else {
      fpValue = (newValue == null) ? 0 : Math32FP.ONE;
    }
  }
}


//-----------------------------------------------------------------------------
// IntegerRecord Class
//-----------------------------------------------------------------------------

/**
 * An integer record stores integer values.
 *
 * @author Alexis BIETTI
 */
class IntegerRecord extends Record {
  /** Inner value of this record. */
  int value = 0;

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public Record copy() {
    IntegerRecord copy = new IntegerRecord();
    copy.value       = value;
    copy.timeStamp   = timeStamp;
    copy.quality     = quality;

    return copy;
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public float floatValue() {
    return (float) value;
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public boolean booleanValue() {
    return value != 0;
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public int intValue() {
    return value;
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public int fixedPointValue() {
    return Math32FP.toFP(value);
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public String stringValue() {
    return Integer.toString(value);
  }

  /**
   * DOCUMENT ME!
   *
   * @return DOCUMENT ME!
   */
  public Object objectValue() {
    return new Integer(value);
  }

  /**
   * DOCUMENT ME!
   *
   * @param newValue DOCUMENT ME!
   */
  public void setValue(Object newValue) {
    if (newValue instanceof Integer) {
      value = ((Integer) newValue).intValue();
    } else if (newValue instanceof Short) {
      value = ((Short) newValue).shortValue();
    } else if (newValue instanceof Byte) {
      value = ((Byte) newValue).byteValue();
    } else if (newValue instanceof Math32FP) {
      value = Math32FP.toInt(((Math32FP) newValue).fpValue());
    } else if (newValue instanceof String) {
      value = Integer.parseInt((String) newValue);
    } else if (newValue instanceof Float) {
      value = (int) ((Float) newValue).longValue();
    } else {
      value = (newValue == null) ? 0 : 1;
    }
  }
}


// End Record.java
