/*
 * Class    br.org.scadabr.protocol.iec101.common.ControlField
 * File     ControlField.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.common;


/**
 * <p>
 * This class manages the Control field from a standard 870 frame.
 * <p>
 * Control field is handled by link layer and gives us following informations:
 * <p>
 * - Primary Message direction (PRM). Indicates if transmission is operated
 * from a master or a slave station.
 * <p>
 * More over, on master implementation case, such control field will contain:
 * <p>
 * - frame count bit that is an indication bit to alternate in order to differenciate
 * user data exchange between two stations.
 * <p>
 * - frame count bit validity. Indicates if we must take FCB value into account. this
 * depends on link service transmission procedures. For example, a reset remote link
 * service doesn't need a significant fcb, so fcv will be in this case equal to zero.
 * <p>
 * On slave implementation case, the informations will be:
 * <p>
 * - access demand bit (ACD). Used by secondary to indicate that something happened
 * and generated an event. In other words, there's some data class 1 to transmit to 
 * a primary station in communication buffer.
 * <p>
 * - data flow control (DFC).Used by slave to warn primary that comm buffers are full. 
 *
 * @author lionnel cauvy
 * @version 1.0
 * @date 01.II.03
 */

public class ControlField
{
	/**
	 * PRM message direction<br>
	 * 0	secondary station to primary station<br>
	 * 1	primary station to secondary station
	 */
	public boolean prm;
	/**
	 * <p>
	 * This boolean has a different signification according to master
	 * or slave implementation in use:
	 * <p>
	 * For a master application, FCB Frame Count Bit (PRM==1) is an
	 * alternating bit for successive S2, S3 services per station.
	 * <p>
	 * In slave case, ACD AccessDemand (PRM==0):<br>
	 * 0	no access demand for class 1 data transmission<br>
	 * 1	access demand for class 1 data transmission
	 */		
	public boolean bool1;
	/**
	 * <p>
	 * This boolean has a different signification according to master
	 * or slave implementation in use:
	 * <p>
	 * For a master application, FCB Frame Count Valid (PRM==1):<br>
	 * 0	FCB invalid<br>
	 * 1	FCB valid
	 * <p>
	 * In slave case, DFC Data Flow Control (PRM==0):<br>
	 * 0	buffer ok<br>
	 * 1	buffer overflow
	 */
	public boolean bool2;
	/**
	 * Function code. This value allows to know wich is the link service
	 * to use: SEND CONFIRM, REQUEST RESPOND, ...
	 */		
	public byte function;
	/**
	 * Octet value. The value of control field as it is carried by a standard
	 * frame.
	 */
	public short value;
	/**
	 * <p>
	 * Allocates a new control field only initialized with values of PRM
	 * and Function Code.
	 * <p>
	 * Values of FCB or ACD, and FCV or DFC will be handled later by link
	 * managers according to status of current execution.
	 *
	 * @param prm		The value of primary message direction.
	 * @param function	The function code.
	 * @param bool1		Value of FCB for master or ACD for slave as a boolean value.
	 * @param bool2		Value of FCV for master or DFC for slave as a boolean value.
	 */	
	public ControlField(boolean prm, byte function, boolean bool1, boolean bool2)
	{
		this.prm		= prm;
		this.function	= function;
		this.bool1		= bool1;
		this.bool2		= bool2;
	}
}