/*
 * Class    br.org.scadabr.protocol.iec101.common.Controller
 * File     Controller.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.common;

import br.org.scadabr.protocol.util.FifoQueue;

/**
 * <p>
 * This class defines a controller owned by physical layer. For convenience, and
 * performance reasons, controller checks FT1.2 standard for frame format
 * compliance of read bytes.
 * 
 * @author lionnel cauvy
 * @version 1.0
 * @date
 */
public class Controller extends Thread {
	/**
	 * Definition of a type of frame with fixed length.
	 */
	private static final byte FIXED = 0x01;
	/**
	 * Definition of a type of frame with variable length.
	 */
	private static final byte VARIABLE = 0x02;
	/**
	 * Definition of a single byte frame.
	 */
	private static final byte SINGLE = 0x03;
	/**
	 * Link layer reference.
	 */
	private PhysicalLayer layer;
	/**
	 * Frame FT1.2 validitor flag.
	 */
	private boolean valid;
	/**
	 * Frame length.
	 */
	private int length;
	/**
	 * Frame type.
	 */
	private byte type;
	/**
	 * Frame checksum.
	 */
	private short checksum;
	/**
	 * Read byte incremental counter.
	 */
	private int count;
	/**
	 * A queue for valid frames.
	 */
	private FifoQueue frame;
	/**
	 * A queue of bytes from receiver reference.
	 */
	private FifoQueue octets;

	private boolean running;

	/**
	 * Internal method used to reset the whole 'read variables'.
	 */
	private void reset() {
		count = 0;
		length = -1;
		type = 0x00;
		checksum = 0;
		valid = false;
	}

	/**
	 * Internal check on well-formed frame according to FT1.2 standard used by
	 * all iec 608705 protocols. To validate, this method updates a valid flag
	 * that controller tests to know if he has to send data to link layer via
	 * physical layer.
	 * 
	 * @param value
	 *            read byte value
	 * 
	 * @return a Short object if link layer is interested in (Control field,
	 *         Address field or User Data), or null byte is part of header.
	 */
	private Short check(short value) {
		try {
			if (count == 1) {
				switch (value) {
				case Interoperability.STARTBYTEFIXEDLENGTH:
					type = FIXED;
					length = Interoperability.FIXEDFRAMELENGTH
							+ Interoperability.FIXED_FRAME_HEADER_SIZE;
					break;

				case Interoperability.STARTBYTEVARIABLELENGTH:
					type = VARIABLE;
					break;

				case Interoperability.SINGLEBYTE:
					type = SINGLE;
					length = 1;
					valid = true;
					break;
				default:
					reset();
					break;
				}

				return null;
			}

			if (type == VARIABLE && count <= 4) {
				switch (count) {
				case 2:
					length = value
							+ Interoperability.VARIABLE_FRAME_HEADER_SIZE;
					break;
				case 3:
					if (value != (length - Interoperability.VARIABLE_FRAME_HEADER_SIZE))
						reset();
					break;
				case 4:
					if (value != Interoperability.STARTBYTEVARIABLELENGTH)
						reset();
					break;
				}

				return null;
			}

			if (count == length - 1) {
				if (((byte) checksum) != ((byte) value))
					reset();
				return null;
			}

			if (count == length) {
				if (value != Interoperability.STOPBYTE)
					reset();
				else
					valid = true;
				return null;
			}

			if (count > length) {
				reset();
				return null;
			}

			checksum += value;
			return new Short(value);
		} catch (Exception e) {
			CodeProperties.log("Check error: " + e);
			return null;
		}
	}

	/**
	 * This allows to check the frame validity according to FT1.2 frame format
	 * standard. Once checked, frame is sent to physical layer.
	 * 
	 * @param octets
	 *            the bytes to pack into a valid frame!
	 */
	public synchronized void verify(FifoQueue octets) {
		this.octets = octets;
		notify();
	}

	protected synchronized void kill() {
		running = false;
		notify();
	}

	/**
	 * Inherited method from Thread super class. A controller is waiting on a
	 * physical layer notification made at write time. Once notified, the
	 * controller is in charge of check the octets read by the receiver. A valid
	 * frame is sent to physical layer.
	 */
	public void run() {
		try {
			while (running) {
				synchronized (this) {
					wait();
					if (!running)
						continue;
				}
				while (!octets.isEmpty()) {
					count++;
					Short octet = check(((Short) octets.dequeue()).shortValue());
					if (octet != null) {
						frame.enqueue(octet);
					}
					if (valid) {
						reset();
						layer.read(frame);
					}
				}
			}
		} catch (InterruptedException e) {
			CodeProperties.log("Controller error: " + e);
		}
		CodeProperties.log("Exit Controller");
	}

	/**
	 * Instanciates a new controller for checking frame format according to
	 * FT1.2 standard. Once validated, frames are sent directly to physical
	 * layer.
	 * 
	 * @param layer
	 *            Physical layer reference.
	 */
	public Controller(PhysicalLayer layer) {
		this.layer = layer;
		this.running = true;
		this.frame = new FifoQueue(CodeProperties.FIFO_SIZE);
		reset();
		start();
	}
} // end Controller.java