/*
 * Class    br.org.scadabr.protocol.iec101.common.FrameReader
 * File     FrameReader.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.common;

import br.org.scadabr.protocol.util.FifoQueue;

/**
 * This class defines the way used to build a frame with
 * bytes from physical layer. We can notice that all the
 * format checking procedures have been deported to physical
 * layer (see Receiver) for reasons of performance.
 *
 * @author lionnel cauvy
 * @version 1.0
 * @date 01.II.03
 */
 
public class FrameReader extends FrameFactory
{
	/**
	 * Sets control field into frame. Control field element contains direction
	 * information (PRM), access demand (ACD) & data flow control (DFC) for master station
	 * (ie messages from remote slave PRM=0) or frame count bit (FCB) & validity (FCV) for slave 
	 * station (ie messages from remote master PRM=1).
	 *
	 * @param octet	value of the control field.
	 */			
	private static void setControlField(short octet)
	{
		//CodeProperties.lnklog("[FrameFactory] control field detected "+octet);

		frame.control = new ControlField(
			((byte)(( octet & MASK_PRM ) >> POS_PRM )==1)?true:false,
			(byte)( octet & MASK_FCT ),
			((byte)(( octet & MASK_FCB_ACD ) >> POS_FCB_ACD )==1)?true:false,
			((byte)(( octet & MASK_FCV_DFC ) >> POS_FCV_DFC )==1)?true:false
		);
		frame.control.value = octet;
	}

	/**
	 * Sets frame address field. Address field is here one-octet sized.
	 * According to the standard it can also be two octets sized.
	 *
	 * @param octet	value of one octet sized address field.
	 */	
	private static void setAddressField(short octet)
	{
		frame.address = octet;
	}
	/**
	 * Sets frame address field. Address field is here two-octets sized.
	 * According to the standard, it can also be one octet sized.
	 *
	 * @param octet1	lsb value of two octets sized address field.
	 * @param octet2	msb value of two octets sized address field.
	 */		
	private static void setAddressField(short octet1, short octet2)
	{
		frame.address = (octet2 << 8) | octet1;
	}

	/**
	 * Sets a byte into user data field.
	 *
	 * @param octet	user data byte.
	 */		
	private static void setUserData(Short octet)
	{
		frame.user.data.addElement(octet);
	}
	


	/**
	 * Build method inherited from Factory. First step of construction is to set 
	 * the control field, second address field and last user data.<br>
	 * Notice that 'header bytes' (ie start,length,checksum & stop bytes) are checked
	 * at physical layer level. So a frame read from physical layer WILL NOT contain
	 * any information about what is defined as header bytes.
	 *
	 * @param object	the object to turn into a frame. (com.ajile.util.FifoQueue) 
	 *
	 * @return the built frame.
	 *
	 * @throw exception if no object to return the built object
	 */		
	public static Frame build(Object object) throws IEC608705Exception
	{
		FifoQueue octets = (FifoQueue)object;

		// factory reinitialization.		
		frame = new Frame();
		
		try
		{			
			// control field step.	
			setControlField(((Short)octets.dequeue()).shortValue());

			// address field step.
			switch (Interoperability.ADDRESSFIELDLENGTH)
			{
			case 1:
				Short octet = (Short)octets.dequeue();
				setAddressField(octet.shortValue());
				break;
			case 2:
				Short octet1 = (Short)octets.dequeue();
				Short octet2 = (Short)octets.dequeue();
				setAddressField(octet1.shortValue(),octet2.shortValue());
				break;
			}
			
			frame.user.address = frame.address;

			// user data step.		
			while(!octets.isEmpty())
			{
				setUserData((Short)octets.dequeue());
			}
		}
		catch(NullPointerException e)
		{
			while (!octets.isEmpty())
				octets.dequeue();
			// restart phy layer.
			throw new IEC608705Exception("Frame not well formed:"+e);			
		}
		/*catch(IEC608705Exception e)
		{
			while (!octets.isEmpty())
				octets.dequeue();
			reset();
			throw new IEC608705Exception("Frame not well formed:"+e);		
		}*/
		
		return frame;
	}
	
};	// end FrameReader.java