/*
 * Class    br.org.scadabr.protocol.iec101.common.FrameWriter
 * File     FrameWriter.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.common;

import br.org.scadabr.protocol.iec101.common.*;

/**
 * <p>
 * This class defines frame coding from data from upper application layer.
 * <p>
 * Remember application layer turn application procedures into link services,
 * so link layer is in charge to turn these services into frames to send
 * on the network. This is the function this class provides.
 * <p>
 * Once built, frames will be sent to lower physical layer.
 *
 * @author lionnel cauvy
 * @version 1.0
 * @date 01.II.03
 */ 
public class FrameWriter extends FrameFactory
{
	/**
	 * <p>
	 * Implementation of Build method inherited from frame factory.
	 * <p>
	 * First step of construction is to set the start header (start byte 
	 * and length byte according to frame type), then control field, 
	 * address field and user data, and at last stop header (checksum and
	 * stop bytes).
	 *
	 * @param object	the object to turn into a frame. (a link service) 
	 *
	 * @return the built frame.
	 *
	 * @throw exception if no object to return the built object
	 */		
	public static Frame build(Object object)
	{
		short length = 0;
		
		LnkService service = (LnkService)object;
		
		frame = new Frame();
		
		if (service.single)
		{
			frame.type = Frame.SINGLE;
			frame.start = Interoperability.SINGLEBYTE;
			return frame;
		}
		
		frame.address = service.address;
		
		switch (Interoperability.ADDRESSFIELDLENGTH)
		{
		case 1:
			frame.checksum = (short)(frame.checksum + frame.address);
			break;
		case 2:
			frame.checksum = (short)(frame.checksum + ( frame.address & 0x00FF));
			frame.checksum = (short)(frame.checksum + ((frame.address & 0xFF00) >> 8));
			break;
		default:
			CodeProperties.lnklog("[FrameWriter] address field lengh not supported");	
			break;
		}
		
		frame.control = service.control;
		frame.control.value =  (short)(((frame.control.prm)?1:0) << POS_PRM);
		frame.control.value += (((frame.control.bool1)?1:0) << POS_FCB_ACD);
		frame.control.value += (((frame.control.bool2)?1:0) << POS_FCV_DFC);
		frame.control.value += (frame.control.function);
		frame.checksum = (short)(frame.checksum + frame.control.value);
		
		if(service.data!=null)
			length = (short)service.data.length;

		for(short i=0;i<length;i++)
		{
			frame.user.data.addElement(new Short(service.data[i]));
			frame.checksum = (short)(frame.checksum + service.data[i]);
		}		
			
		frame.length = (short)(length + 1 + Interoperability.ADDRESSFIELDLENGTH);	
		
		switch (frame.length)
		{

		case Interoperability.FIXEDFRAMELENGTH:

			frame.type = Frame.FIXED;	
			frame.start = Interoperability.STARTBYTEFIXEDLENGTH;
			frame.stop = Interoperability.STOPBYTE;
			break;			
			
		default:

			frame.type = Frame.VARIABLE;		
			frame.start = Interoperability.STARTBYTEVARIABLELENGTH;		
			frame.stop = Interoperability.STOPBYTE;
			break;
		}
		
		return frame;
	}
		
};	// end FrameWriter.java