/*
 * Class    br.org.scadabr.protocol.iec101.common.LayerThread
 * File     LayerThread.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.common;

import br.org.scadabr.protocol.util.FifoQueue;

/**
 * <p>
 * This class defines a layer with thread abilities. It describes the basic
 * mechanisms used by such layers to communicate.
 * 
 * @author lionnel cauvy
 * @version 1.0
 * @date 01.II.03
 */

public abstract class LayerThread extends Layer implements Runnable {
	private String name;
	private boolean running;
	private Thread thread;
	protected FifoQueue fromUpper;
	private FifoQueue fromLower;

	public String toString() {
		return name;
	}

	protected abstract void upperEvent(Object object) throws IEC608705Exception;

	protected abstract void lowerEvent(Object object) throws IEC608705Exception;

	public synchronized void receive(Object object) {
		fromUpper.enqueue(object);
		notify();
	}

	public synchronized void read(Object object) {
		fromLower.enqueue(object);
		notify();
	}

	/**
	 * Kill order received from upper. This layer MUST send down this order
	 * before stopping itself.
	 */
	public synchronized void kill() {
		if (lower != null) {
			lower.kill();
		}
		lower = null;
		running = false;
		notify();
	}

	/**
	 * Kill order received from lower layer. This layer MUST send up this order
	 * after stopping itself.
	 */
	public synchronized void lowerKill() {
		if (upper != null)
			upper.lowerKill();
		else
			kill();
	}

	/**
	 * Wait this thread to die...
	 * 
	 * @throws Interrupted
	 *             Exception
	 */
	public void join() throws InterruptedException {
		this.thread.join();
	}

	/**
	 *
	 */
	public synchronized void run() {
		try {
			while (running) {
				while (!fromUpper.isEmpty() || !fromLower.isEmpty()) {
					if (!fromLower.isEmpty())
						lowerEvent(fromLower.dequeue());
					if (!fromUpper.isEmpty())
						upperEvent(fromUpper.dequeue());
				}

				wait();
			}
		} catch (IEC608705Exception e) {
			CodeProperties.log("CEI Exception: " + e);
		} catch (InterruptedException e) {
			CodeProperties.log("thread synchro problem: " + e);
		} catch (Exception e) {
			CodeProperties.log("Exception at " + name + ": " + e);
		}
		CodeProperties.log("Exit " + name);
	}

	protected void launch() {
		if (lower != null)
			lower.launch();
		(thread = new Thread(this)).start();

	}

	/**
	 * Builds a new LayerThread object.
	 * 
	 * @param name
	 *            thread name.
	 */
	protected LayerThread(String name) {
		this.name = name;
		this.running = true;
		this.fromLower = new FifoQueue(CodeProperties.FIFO_LAYER_SIZE);
		this.fromUpper = new FifoQueue(CodeProperties.FIFO_LAYER_SIZE);
	}

}; // end LayerThread.java