/*
 * Class    br.org.scadabr.protocol.iec101.common.PhysicalLayer
 * File     PhysicalLayer.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.common;

import gnu.io.SerialPort;

import java.io.InputStream;
import java.io.OutputStream;

import br.org.scadabr.luciol.ajile.components.CallbackTimer;
import br.org.scadabr.protocol.util.FifoQueue;

import com.ajile.events.TriggerEventListener;
import com.ajile.util.ArrayUtils;

/**
 * <p>
 * This class defines physical layer. This layer from EPA model must be able to:
 * <br>
 * - write data to network (transmitter role)<br>
 * - read data from network (receiver role)<br>
 * Physical layer is the lowest layer of this architecture. It's upper layer is
 * Link Layer. For convenience, a watchdog was added at this level to validate
 * correct transmissions and invalidate others.
 * 
 * @author lionnel cauvy
 * @version 1.0
 * @date 01.II.03
 */

public abstract class PhysicalLayer extends Layer implements
		TriggerEventListener {

	/**
	 * Well formed frame controller reference.
	 */
	protected Controller controller;
	/**
	 * Receiver object reference.
	 */
	protected Receiver receiver;
	/**
	 * Serial port reference.
	 */
	protected SerialPort sport;
	/**
	 * Output stream reference.
	 */
	protected OutputStream output;
	/**
	 * Input stream reference.
	 */
	protected InputStream input;
	/**
	 * Watchdog on physical transmission.
	 */
	protected CallbackTimer timer;
	/**
	 * Written data.
	 */
	private byte[] backup;
	/**
	 * Transmission watchdog duration.
	 */
	public int timeout = 0;
	/**
	 * Allowed number of communication retries in fail case.
	 */
	public int retries = 0;
	/**
	 * Incremental counter of attempt. When its value equals retries value,
	 * physical layer is considered as invalid. This invalidates all the
	 * application.
	 */
	protected int attempt;

	/**
	 * This event is raised by the watchdog timer. So this call means there's
	 * some trouble on physical transmission. Then physical layer have to
	 * re-write the last byte array written.
	 */
	public void triggerEvent() {
		if (attempt < retries) {
			CodeProperties.phylog("retry");
			// re-write backuped data
			// if(CodeProperties.AJILE)
			// receiver.skip(backup.length);
			write(backup);
			// watchdog again!
			setWatchdog();
		} else {

			CodeProperties.log("Exit (no more retries left attempt:" + attempt
					+ " retries:" + retries + ")");
			// application stop.
			kill();
			upper.lowerKill();
		}
	}

	/**
	 * Sets a timer to validate a transmission. An answer is required before
	 * time out.
	 */
	private void setWatchdog() {
		// attempt++;
		// timer.callback(System.currentTimeMillis() + 2000 * timeout, this);
	}

	/**
	 * Resets watchdog. Last transmission was ok.
	 */
	private void unsetWatchdog() {
		attempt--;
		timer.removeCallback(this);
	}

	/**
	 * Keeps in memory last byte array written on port. This allows re-write in
	 * transmission fail-case.
	 * 
	 * @param data
	 *            the data to backup.
	 */
	private void backup(byte[] data) {
		backup = new byte[data.length];
		if (CodeProperties.AJILE)
			ArrayUtils.copyByteArray(data, 0, backup, 0, data.length);
		else {
			for (int i = 0; i < data.length; i++) {
				backup[i] = data[i];
			}
		}
	}

	/**
	 * Writes a byte array on communication port.
	 * 
	 * @param data
	 *            the data to write.
	 */
	protected abstract void write(byte[] data);

	/**
	 * Called by upper layer to transmit data to this layer. Physical layer is
	 * in charge to write data received on communication port.
	 * 
	 * @param object
	 *            the data received from upper layer. To write on communication
	 *            port.
	 */
	public void receive(Object object) {
		// object received is a byte array.
		byte[] data = (byte[]) object;
		// backup object
		if (timeout != 0)
			backup(data);
		// write
		// if(CodeProperties.AJILE)
		// receiver.skip(data.length);
		write(data);
		// watchdog.
		if (timeout != 0)
			setWatchdog();
	}

	/**
	 * Called by lower layer to transmit data to this layer. Physical layer read
	 * method is not raised by lower layer, because there's not, it's the
	 * controller thread that calls this method. Valid data has been read on
	 * communication port.
	 * 
	 * @param object
	 *            the data readed from receiver
	 */
	public void read(Object object) {
		if (timeout != 0)
			// stops the watchdog: physical transmission is ok!
			unsetWatchdog();
		// pass information to upper layer
		System.out.println(timeout);

		FifoQueue queue = (FifoQueue) object;
		upper.read(object);
	}

	/**
	 * Kill physical layer and all lower objects and/or threads. Can be called
	 * from upper layer or by physical layer itself when there's no more
	 * remaining retries for validate a transmission.
	 * 
	 * @throws runtime
	 *             exception.
	 */
	public abstract void kill();

	/**
	 * NOT USED.
	 */
	public void lowerKill() {
		// there's no lower layer here.
	}

	/**
	 * Instanciates a physical layer.Such an object has to handle bytes
	 * transmission on communication port.<br>
	 * This constructor initializes the communication port by use of javax.comm
	 * package<br>
	 */

	public abstract void launch();
} // end PhysicalLayer.java
