/*
 * Class    br.org.scadabr.protocol.iec101.common.PrivateRange
 * File     PrivateRange.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.common;

import java.util.Hashtable;
import java.util.Enumeration;

/**
 * <p>
 * This class describes private range for a station. Private range
 * provides a way to decode constructor private implementation of
 * an iec standard.
 * <p>
 * From common package, we can describe private causes of transmission
 * & private asdus.
 * <p>
 * Causes of transmission are only a value. During decoding a received
 * asdu from remote, if no compatible value for cause of transmission
 * matches, then application will try to consult private range to eventually
 * find out the matching value. If no, frame will be rejected.
 * <p>
 * Asdus are much more complex objects because they are a set of information
 * objects, themselves being a set of information elements. OUR APPLICATION
 * ONLY ACCEPTS STANDARD INFORMATION ELEMENTS ACCORDING TO IEC IMPLEMENTATION
 * 10X IN USE. HOWEVER, BY USING SPECIAL INFORMATION ELEMENT 'NUD' (ie Not UseD)
 * AND BY APPLYING TRANSLATION (see Translator), THIS APPLICATION SHOULD BE
 * ABLE TO DECODE NON-STANDARD OCTET FROM A PRIVATE ASDU.
 * <p>
 *
 * @author lionnel cauvy
 * @version 1.0
 * @date 01.II.03
 *
 * @see PrivateAsdu
 * @see PrivateCot
 */
 
public class PrivateRange
{
	/**
	 * Private asdus table.
	 */
	public Hashtable Table_Id_Asdu;
	/**
	 * Private causes of transmission table.
	 */
	public Hashtable Table_Id_Cot;
	
	/**
	 * Add a cause of transmission to private range.
	 *
	 * @param pcot	An instance of private cause of transmission.
	 */
	public void addCot(PrivateCot pcot)
	{
		Table_Id_Cot.put(new Integer(pcot.id),pcot);
	}
	
	/**
	 * Add an asdu to private range.
	 *
	 * @param pasdu	An instance of private asdu.
	 */
	public void addAsdu(PrivateAsdu pasdu)
	{
		Table_Id_Asdu.put(new Integer(pasdu.id),pasdu);
	}

	/**
	 * Get accessor on all cause of transmissions from this private range.
	 *
	 * @return an enumeration of private causes of transmission.
	 */	
	public Enumeration enumerateCots()
	{
		return Table_Id_Cot.elements();
	}

	/**
	 * Get accessor on a single cause of transmission from its value.
	 *
	 * @param id	Value of cause of transmission.
	 *
	 * @return the private Cause of Transmission or null.
	 */	
	public PrivateCot getCot(short id)
	{
		return (PrivateCot)Table_Id_Cot.get(new Integer(id));
	}


	/**
	 * Get accessor on a single asdu from its value.
	 *
	 * @param id	Value of asdu type identification.
	 *
	 * @return the private asdu or null.
	 */	
	public PrivateAsdu getAsdu(short id)
	{
		return (PrivateAsdu)Table_Id_Asdu.get(new Integer(id));
	}

	/**
	 * Get accessor on all asdus from this private range.
	 *
	 * @return an enumeration of private asdus.
	 */			
	public Enumeration enumerateAsdus()
	{
		return Table_Id_Asdu.elements();
	}

	/**
	 * Allocates a private range.
	 */	
	public PrivateRange()
	{		
		Table_Id_Asdu = new Hashtable();
		Table_Id_Cot = new Hashtable();
	}
};	// end PrivateRange.java