/*
 * Class    br.org.scadabr.protocol.iec101.common.Receiver
 * File     Receiver.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.common;

import java.io.IOException;
import java.io.InputStream;

import br.org.scadabr.protocol.util.FifoQueue;

/**
 * <p>
 * This class defines a receiver owned by physical layer. Such an object is very
 * simple because its ontly job is to block input stream until some data is
 * received. Then he packs all these bytes in a queue and send it to a format
 * controller.
 * 
 * @author lionnel cauvy
 * @version 1.0
 * @date 01.II.03
 */
public class Receiver extends Thread {
	/**
	 * Frame controller reference.
	 */
	private Controller controller;
	/**
	 * Input stream reference.
	 */
	private InputStream input;
	/**
	 * Read bytes container.
	 */
	private FifoQueue octets;

	private boolean running;

	private int skip = 0;

	// private GpioPin ajileRts;

	/**
	 * A call to exit thread.
	 */
	protected synchronized void kill() {
		running = false;
	}

	protected synchronized void skip(int nb) {
		skip = nb;
	}

	/**
	 * Inherited Thread method implementation. Receiver reads input stream of
	 * communication port. Notice that it DOESN'T skip the written octets. So
	 * controller will be in charge of erase echoed written bytes.
	 */
	public void run() {
		while (running) {
			try {
				if (input.available() > 0) {
					short value = (short) input.read();
					if (skip-- > 0)
						continue;

					System.out.print(" <-- " + value);
					octets.enqueue(new Short(value));
					// avoids blocking call to read function.
					while (input.available() > 0) {
						value = (short) input.read();
						if (skip-- <= 0) {
							System.out.print(" " + value);
							octets.enqueue(new Short(value));
						}
					}
					if (!octets.isEmpty()) {
						System.out.println();
						if (controller != null)
							controller.verify(octets);
					}
				}
				Thread.yield();
			} catch (IOException e) {
				CodeProperties.log("IO Exception");
				running = false;
			}
		}
		CodeProperties.log("Exit Receiver");

	}

	/**
	 * Instanciates a receiver on communication port input stream.
	 * 
	 * @param controller
	 *            frame format controller reference.
	 * @param input
	 *            input stream reference.
	 */
	protected Receiver(Controller controller, InputStream input) {
		this.controller = controller;
		this.input = input;
		this.running = true;
		this.octets = new FifoQueue(CodeProperties.FIFO_SIZE);
		// this.ajileRts = new GpioPin(GpioPin.GPIOD_BIT3);
		start();
	}

}; // end Receiver.java