/*
 * Class    br.org.scadabr.protocol.iec101.common103.information.KOD
 * File     KOD.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.common103.information;

import br.org.scadabr.protocol.iec101.common.InformationElement;

/**
 * <b>Kind of description</b><br>
 * <p><i>
 * KOD := UI8<0..255><br>
 * with <0> := NO KOD SPECIFIED<br>
 * <1> := ACTUAL VALUE<br>
 * <2> := DEFAULT VALUE<br>
 * <3> := RANGE (minimum value, maximum value and step size)<br>
 * <4> := (reserved)<br>
 * <5> := PRECISION (n, m)<br>
 * <6> := FACTOR<br>
 * <7> := % REFERENCE<br>
 * <8> := ENUMERATION<br>
 * <9> := DIMENSION<br>
 * <10> := DESCRIPTION<br>
 * <11> := (reserved)<br>
 * <12> := PASSWORD ENTRY<br>
 * <13> := IS READ ONLY<br>
 * <14> := IS WRITE ONLY<br>
 * <15> := (reserved)<br>
 * <16> := (reserved)<br>
 * <17> := (reserved)<br>
 * <18> := (reserved)<br>
 * <19> := CORRESPONDING FUNCTION TYPE AND INFORMATION NUMBER<br>
 * <20> := CORRESPONDING EVENT<br>
 * <21> := ENUMERATED TEXT ARRAY<br>
 * <22> := ENUMERATED VALUE ARRAY<br>
 * <23> := RELATED ENTRIES<br>
 * <24..255> := (reserved)<br>
 * Each entry has a variable number of descriptions (attributes), only the <10> DESCRIPTION attribute is
 * necessary.<br>
 * <br>
 * <b>Explanations of kind of description</b><br>
 * NO KOD SPECIFIED<br>
 * This KOD is used in a read request when neither a value nor an attribute is necessary (for example
 * read the complete directory of a GIN). No directory entry will have an attribute with this KOD.<br>
 * <br>
 * ACTUAL VALUE <br>
 * The actual value of an entry. In the case of a heading GIN, this will refer to the number of entries
 * within that GIN group of the directory. If the DATATYPE of this KOD is INDEX with a value of N, the
 * ACTUAL VALUE of this GIN is obtained by referring to the (N+1) element of the ENUMERATION KOD
 * array for ths GIN. See also the entry for RANGE and ENUMERATION TEXT ARRAY / VALUE ARRAY. <br>
 * <br>
 * DEFAULT VALUE <br>
 * The initial value of an entry. It may be used to reset an entry value. This will have the same <br>
 * <br>
 * DATATYPE as the ACTUAL VALUE. <br>
 * <br>
 * RANGE <br>
 * The RANGE indicates the possible valid values of an entry, consisting of an array of three elements:
 * the minimum value, maximum value and step size, in that order. Valid values are the minimum
 * value of RANGE plus multiples of step size up to the maximum value. The DATATYPE of this KOD will
 * normally match that of the ACTUAL VALUE KOD. Where the ACTUAL VALUE is an index representing
 * enumerated text or values, the RANGE will also be of DATATYPE INDEX comprising: <br>
 *  Minimum value = 0; <br>
 *  Maximum value = <number of enumerations 1>; <br>
 *  step size = 1. <br>
 * For each value in this range, there shall be a corresponding element in the enumeration array. <br>
 * <br>
 * PRECISION (n, m) <br>
 * PRECISION defines how a real number should be displayed. n defines the maximum number of digits
 * of the integer part, m defines the number of digits of the fractional part. <br>
 * <br>
 * FACTOR <br>
 * Multiplication factor to get the original value. <br>
 * <br>
 * % REFERENCE <br>
 * A percentage reference of a relative value, for example the measurand with quality descriptor
 * (ASDU 3 or 9) reference is 120 % or 240 %. <br>
 * <br>
 * ENUMERATION <br>
 * This KOD is used to enumerate all valid values of the ACTUAL VALUE KOD where the DATATYPE of
 * the ACTUAL VALUE is of type INDEX. This is useful for cases where the numeric values are non-linear
 * in step-size, or where the individual values are better represented by items of text. The first
 * ENUMERATION is referenced by the index value of 0. If only some of the values require enumeration,
 * the ENUMERATED TEXT ARRAY / ENUMERATED VALUE ARRAY KODs can be used instead. <br>
 * <br>
 * DIMENSION <br>
 * Text to indicate the dimension of the value of an entry, for example A, Var, V. <br>
 * <br>
 * DESCRIPTION <br>
 * A textual description of an entry or group, or an external text number. <br
 * <br>
 * PASSWORD ENTRY <br>
 * The GIN of an entry that contains the corresponding password to allow setting of this entry. <br>
 * <br>
 * CORRESPONDING FUNCTION TYPE AND INFORMATION NUMBER <br>
 * For indications and measurands, it specifies the FUNCTION TYPE and INFORMATION NUMBER,
 * respectively of the corresponding indications of non-generic communication functions.
 * For commands, it specifies the FUNCTION TYPE and INFORMATION NUMBER of the corresponding non-generic
 * command procedure and of the corresponding message. <br>
 * <br>
 * CORRESPONDING EVENT <br>
 * Defines the event (message) which reflects the status or status changes resulting from a command
 * execution. <br>
 * <br>
 * ENUMERATED TEXT ARRAY / VALUE ARRAY <br>
 * These two KODs are used together to define text values for specific special values of the ACTUAL
 * VALUE. For each ACTUAL VALUE that is required to be enumerated, the value is stored in the
 * ENUMERATED VALUE ARRAY and the matching text is stored in the corresponding position of the
 * ENUMERATED TEXT ARRAY. If all values are to be enumerated, the ENUMERATION KOD can be used
 * instead of the ACTUAL VALUE as an index for less overhead. <br>
 * <br>
 * RELATED ENTRIES <br>
 * Comprises an array of GINs which are related to this entry. Can be used to set up relationships to
 * other entries, for example to identify the entry for activation of a set of parameters. <br>
 * <br>
 * </i>
 *
 * @author lionnel cauvy
 * @version 1.0
 * @date 01.II.03
 */

public class KOD extends InformationElement
{
	/** KOD value
	  *
	  */


	/**
	 * constructs KOD object<br>
	 * kind of description 7.2.6.34
	 * @param octet value of 'physical' data
	 */
	public KOD(short octet)
	{
		super(1,1);
		octetset[0] = octet;
		decode();
	}
		
	/**
	 * builds sub information data from 'physical' data.
	 * called when physical data is all known
	 */	
	protected void decode()
	{
		dataset[0] = new Short(octetset[0]);
	}
	
	/**
	 * builds  'physical' data from sub information data.
	 * called when sub information data is all known
	 */	
	protected void code()
	{
		octetset[0] = ((Short)dataset[0]).shortValue();
	}
	
	/**
	 * default constructor
	 */
	public KOD()
	{
		super(1,1);
	}
	
	
};	// end KOD.java