/*
 * Class    br.org.scadabr.protocol.iec101.common103.information.SOF
 * File     SOF.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.common103.information;

import br.org.scadabr.protocol.iec101.common.InformationElement;

/**
 * <p>
 * <b>Status of fault 7.2.6.24</b><br>
 * (information element used by asdu id 23) <br>
 * <p><i>
 * SOF := BS8{TP, TM, TEST, OTEV, RES} <br>
 * with <br>
 * <b>TP := BS1[1]				--> dataset[0]</b> <br>
 * <0> := recorded fault without trip <br>
 * <1> := recorded fault with trip <br>
 * <b>TM := BS1[2]				--> dataset[1]</b> <br>
 * <0> := disturbance data waiting for transmission <br>
 * <1> := disturbance data currently being transmitted <br>
 * <b>TEST := BS1[3]			--> dataset[2]</b> <br>
 * <0> := disturbance data recorded during normal operation <br>
 * <1> := disturbance data recorded during test mode <br>
 * <b>OTEV := BS1[4]			--> dataset[3]</b> <br>
 * <0> := disturbance data recording initiated by start/pick-up <br>
 * <1> := disturbance data recording initiated by other events <br>
 * <b>RES := BS4[5..8] not used	--> dataset[4]</b> <br>
 * <p>
 * NOTE  SOF indicates whether the protection equipment has tripped during the fault (bit TP), whether the disturbance data are
 * currently being transmitted (bit TM), whether the disturbance data have been recorded during normal operation or test mode
 * (bit TEST), and whether disturbance data recording has been initiated by another event than start/pick-up (bit OTEV). <br>
 * </i>
 *
 * @author lionnel cauvy
 * @version 1.0
 * @date 01.II.03
 */

public class SOF extends InformationElement
{	
	/**
	 * constructs SOF object<br>
	 * status of fault 7.2.6.24
	 * @param octet value of 'physical' data
	 */
	public SOF(short octet)
	{
		super(1,5);
		octetset[0] = octet;
		
		decode();		
	}

	/**
	 * builds sub information data from 'physical' data.
	 * called when physical data is all known
	 */
	protected void decode()
	{
		
		dataset[0] = new Boolean(((octetset[0] & 0x0001)==0)?false:true);
		dataset[1] = new Boolean((((octetset[0] & 0x0002)>>1)==0)?false:true);
		dataset[2] = new Boolean((((octetset[0] & 0x0004)>>2)==0)?false:true);
		dataset[3] = new Boolean((((octetset[0] & 0x0008)>>3)==0)?false:true);
		dataset[4] = new Byte((byte)((octetset[0] & 0x00F0)>>4));
	}
	
	/**
	 * builds  'physical' data from sub information data.
	 * called when sub information data is all known
	 */	
	protected void code()
	{
		octetset[0] = (short)((((Boolean)dataset[0]).booleanValue()?1:0) | ((((Boolean)dataset[1]).booleanValue()?1:0) << 1) | ((((Boolean)dataset[2]).booleanValue()?1:0) << 2) | ((((Boolean)dataset[3]).booleanValue()?1:0) << 3) | (((Byte)dataset[4]).byteValue() << 4) );
	}
	
	/**
	 * default constructor
	 */
	public SOF()
	{
		super(1,5);
	}
	
};	// end SOF.java