/*
 * Class    br.org.scadabr.protocol.iec101.master.MAppManager
 * File     MAppManager.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.master;

import br.org.scadabr.protocol.iec101.common.AppManager;
import br.org.scadabr.protocol.iec101.common.AppService;
import br.org.scadabr.protocol.iec101.common.Asdu;
import br.org.scadabr.protocol.iec101.common.CodeProperties;
import br.org.scadabr.protocol.iec101.common.Context;
import br.org.scadabr.protocol.iec101.master101.MMapping101;
import br.org.scadabr.protocol.iec101.session.CompositeMessage;
import br.org.scadabr.protocol.iec101.session.Response;
import br.org.scadabr.protocol.iec101.session.database.DataElement;
import br.org.scadabr.protocol.iec101.session.database.Database;
import br.org.scadabr.protocol.util.UnicPrioQueue;

/**
 * <p>
 * General contract is to define manager object for application layer. An
 * application manager interfaces Mapping and therefore database with
 * application layer.
 * <p>
 * One one hand, an aplication manager is in charge of handling application
 * procedures and services. It's the command role of such an object. Notice that
 * these mechanisms manager must drive come from database notification via
 * application mapping. Or it can also be internal procedures the manager
 * launches by its own. In order to handle notifications, this class implements
 * the MappingReader interface to enable the receive of procedures via the
 * mapEvent method. The listener is recorded during iniatial step of
 * Configuration.
 * <p>
 * One the other hand, AppManager must be able to write some protocol data to
 * the database.
 * <p>
 * Notice that for a slave application, there wiil be only one unique instance
 * of this manager. Whereas for a master application, there will be as many
 * AppManager instances as there will be remote stations.
 * 
 * @author lionnel cauvy
 * @version 1.0
 * @date 01.II.03
 */
public class MAppManager extends AppManager {
	/**
	 * Final static flag for a lazy application procedure.
	 */

	public static final boolean LAZY = false;
	/**
	 * Final static flag for a workin application procedure.
	 */
	public static final boolean WORK = true;
	/**
	 * Internal initialization flag. Set to true when initial procedure has been
	 * passed.
	 */
	private boolean initialized = false;

	private Database database;
	/**
	 * Array of initial procedures to execute on start/restart.
	 */
	public MAppService initial[];
	/**
	 * A fifo queue whith unique priority key. Priority are related to the
	 * procedures and are defined in CodeProperties interface. Each procedure
	 * from mapping or internal ones will be stored into this fifo before to be
	 * launch.
	 */
	public UnicPrioQueue procedures;
	/**
	 * The current procedure. Can be null only in LAZY state.
	 */
	public MAppService current;
	public CompositeMessage currentResponseHandler;
	/**
	 * Flag to indicate LAZY/WORK state of this manager.
	 */
	public boolean state;
	private int timeout = 800;

	/**
	 * <p>
	 * Add a new application procedure to fifo.
	 * 
	 * @param proc
	 *            The application procedure to add.
	 */
	public void push(MAppService proc) {
		procedures.enqueue(proc, proc.identifier.priority);
		manage();
	}

	/**
	 * <p>
	 * Start/restart this manager.
	 */
	public void start() {
		// first, copy initial procedures in procedures queue
		System.out.println("MAppManager start");
		for (int i = 0; i < initial.length; i++) {
			procedures.enqueue(initial[i], initial[i].identifier.priority);
			System.out.println("initial " + initial[i]);
		}
		// then, just manage....
		manage();
	}

	/**
	 * <p>
	 * This method defines the way used to manage application procedures.
	 * <p>
	 * If application manager and related link manager are in 'lazy' state,
	 * manager can here launch a new procedure.<br>
	 * Otherwise, it has no effect. Indeed, this method is called in 2 different
	 * cases:<br>
	 * - first, application procedure is finished. So, if related link layer is
	 * in stable state, manager MUST launch new procedure if available. - -
	 * second, related link layer raised this method when link service is over.
	 * At this point, application layer can restart a a procedure because link
	 * manager always ends on a data class 1 polling NACK. So application layer
	 * won't receive any supplementary asdus to decode.
	 */
	public void manage() {
		// System.out.println("manage()");

		if (state == LAZY && lnk.state == LAZY) {
			current = (MAppService) procedures.dequeue();
			if (current == null) {
				if (!initialized) {
					initialized = true;
					state = LAZY;
				}
				return;
			}

			if (current.expectedAsdu != null)
				state = WORK;

			System.out.println("Enviando frame: " + current);
			layer.receive(current);
		} else {
			String appState = "AppState: WORKING;";
			if (!state)
				appState = "AppState: LAZY;";
			String lnkState = "LnkState: WORKING;";
			if (!lnk.state)
				lnkState = "LnkState: LAZY;";

			System.out.println(appState);
			System.out.println(lnkState);
		}
	}

	public void end(int res) {
	}

	/**
	 * <p>
	 * Asdu evaluation in order to validate an application procedure. This
	 * method is also eventually used to write some data in Luciol central
	 * databased according to mapping.
	 * 
	 * @param asdu
	 *            The Application LnkService Data Unit to check.
	 */
	public void evaluate(Asdu asdu) {
	}

	public synchronized int send(CompositeMessage messages) {
		do {
			current = messages.nextMessage();
			System.out.println("Enviando " + current);
			currentResponseHandler = messages;
			responseStatus = 0;
			long sentTime = System.currentTimeMillis();
			layer.receive(current);
			while ((responseStatus == CompositeMessage.WAITING)) {
				if ((System.currentTimeMillis() - sentTime) > timeout)
					responseStatus = CompositeMessage.TIMED_OUT;
			}
		} while (responseStatus == CompositeMessage.CONTINUE);
		if (responseStatus == CompositeMessage.TIMED_OUT)
			System.out.println("[IEC101] TIMED OUT!");
		return responseStatus;
	}

	public void dataReceived(DataElement element) {
		System.out.println("Type Identification: "
				+ element.getTypeIdentification() + " / Address: "
				+ element.getAddress() + " / Value: " + element.getValue());
		database.writeRecord(element);
	}

	private int responseStatus = 0;

	public void responseReceived(Response response) {
		if (currentResponseHandler != null) {
			responseStatus = currentResponseHandler.responseReceived(response);
		} else {
			responseStatus = CompositeMessage.FAILED;
		}
	}

	/**
	 * <p>
	 * This event is raised by all running periodic tasks. In this case, it may
	 * be polling data class 2 procedure, synchronization one, or at last,
	 * general interrogation one. An identifier is returned to allow to
	 * differenciate these cyclic procedures.
	 * <p>
	 * When application layer is so triggered, we must alert an application
	 * manager to execute the procedure. One by one, according to value of
	 * count, each manager will be solicitated.
	 * 
	 * @param svc
	 *            Application service.
	 * @param count
	 *            Triggering counter.( Which remote is concerned.)
	 */
	public void triggerEvent(AppService svc, int count) {
		push((MAppService) svc);
	}

	/**
	 * <p>
	 * Allocates a new manager for application layer according to a single
	 * remote secondary/slave station.
	 * 
	 * @param layer
	 *            Application layer reference.
	 * @param context
	 *            Execution context.
	 * @param nbToPoll
	 *            Maximum number of asdus to poll during a polling cycle.
	 */
	public MAppManager(MApplicationLayer layer, Context context, int nbToPoll) {
		super(layer, context);
		this.lnk = new MLnkManager(this, nbToPoll, context.address);
		procedures = new UnicPrioQueue(CodeProperties.FIFO_PROCEDURES_SIZE);
		map = new MMapping101(this);
		database = new Database();
		// map.addWriter(new MapAsdu(new AsduKey101(id, vsq, cot, oa, caa)))
	}

	public Database getDatabase() {
		return database;
	}

	public void setDatabase(Database database) {
		this.database = database;
	}

	public void setTimeout(int timeout) {
		this.timeout = timeout;
	}

	public int getTimeout() {
		return timeout;
	}

} // end MAppManager.java
