/*
 * Class    br.org.scadabr.protocol.iec101.master.MLinkLayer
 * File     MLinkLayer.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.master;

import java.util.Hashtable;

import br.org.scadabr.protocol.iec101.common.CodeProperties;
import br.org.scadabr.protocol.iec101.common.Frame;
import br.org.scadabr.protocol.iec101.common.FrameReader;
import br.org.scadabr.protocol.iec101.common.IEC608705Exception;
import br.org.scadabr.protocol.iec101.common.Interoperability;
import br.org.scadabr.protocol.iec101.common.LinkLayer;
import br.org.scadabr.protocol.iec101.session.Response;

/**
 * <p>
 * This class defines used implementation of IEC 608705 link layer from EPA
 * model.<br>
 * Link layer from iec60870 has to:<br>
 * - check transmission frame format (IEC 60870-5-1).<br>
 * - provide transmission procedures (IEC 60870-5-2).<br>
 * <p>
 * Notice that FT1.2 format compliance verification has been moved to lower
 * physical layel for convenience and reasons of performance.
 * 
 * @author lionnel cauvy
 * @version 1.0
 * @date 01.II.03
 */

public class MLinkLayer extends LinkLayer {
	/**
	 * Final static identifier to check prm direction on received frames. PRM is
	 * a bit hol by control field of a frame.
	 */
	private static final boolean MODE = false;
	/**
	 * Table of LnkManager referenced by address.
	 */
	public Hashtable manager;

	/**
	 * This event is raised when there's some data (FT1.2 formatted) from lower
	 * physical layer. So, first, link layer have got to build related frame,
	 * next to evaluate this answer from remote, and last to send it to upper
	 * application layer.
	 * 
	 * @param object
	 *            data from lower layer.
	 */
	protected void lowerEvent(Object object) throws IEC608705Exception {
		try {
			Frame frame = (Frame) (FrameReader.build(object));

			if (frame == null) {
				CodeProperties
						.lnklog("Receive a null frame from lower physical layer.");
				return;
			}
			if (frame.control.prm != MODE) {
				CodeProperties.lnklog("Receive a reverse direction frame: PRM="
						+ frame.control.prm);
				frame = null;
				return;
			}

			MLnkManager mngr = (MLnkManager) manager.get(new Integer(
					frame.address));

			if (mngr.state == MLnkManager.LAZY) {
				CodeProperties.lnklog("Unsolicited answer from remote address:"
						+ mngr.address);
				return;
			}

			byte answer = frame.control.function;
			mngr.acd = frame.control.bool1;
			mngr.dfc = frame.control.bool2;

			if (!frame.user.data.isEmpty())
				send(frame.user);
			else {
				Response response = new Response();
				response.setAcd(mngr.acd);
				response.setControlFunction(answer);
				((MAppManager) mngr.app).responseReceived(response);
			}

			mngr.state = MLnkManager.LAZY;
			((MAppManager) mngr.app).end(((MLnkService) mngr.current)
					.check(answer));
			frame = null;
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * This event is raised when there's some user data from upper application
	 * layer. So, first, link layer have got to encapsulate this object into a
	 * frame, and last to write it to lower physical layer. evaluate this answer
	 * from remote.
	 * 
	 * @param object
	 *            user data from upper application layer.
	 */
	protected void upperEvent(Object object) throws IEC608705Exception {
		// System.out.println("LinkLayer -> iniciar envio");
		MLnkService svc = (MLnkService) object;
		MLnkManager mngr = (MLnkManager) manager.get(new Integer(svc.address));
		mngr.current = svc;
		mngr.state = MLnkManager.WORK;
		switch (svc.control.function) {
		case Interoperability.SERVICE_S1_SN_USER_DATA:
		case Interoperability.SERVICE_S2_SC_USER_DATA:
		case Interoperability.SERVICE_S3_RR_ACCESS_DEMAND:
			// case Interoperability.SERVICE_S3_RR_LINK_STATUS:
		case Interoperability.SERVICE_S3_RR_USER_DATA_CLASS_1:
			mngr.fcv = true;
			mngr.fcb = !mngr.fcb;
			break;
		case Interoperability.SERVICE_S3_RR_USER_DATA_CLASS_2:
			mngr.polling = true;
			--mngr.pollCount;
			mngr.fcv = true;
			mngr.fcb = !mngr.fcb;
			break;
		default:
			mngr.fcv = false;
			break;
		}
		mngr.current.control.bool1 = mngr.fcb;
		mngr.current.control.bool2 = mngr.fcv;

		buildAndWrite(object);
	}

	/**
	 * <p>
	 * Allocates a link layer for a master application.
	 * 
	 * @param size
	 *            Number of managers to handle.
	 */
	public MLinkLayer(int size) {
		manager = new Hashtable(size);
	}

} // end MLinkLayer.java

