/*
 * Class    br.org.scadabr.protocol.master103.MConfiguration103;
 * File     MConfiguration103.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.master103;

import java.util.Enumeration;
import java.util.Vector;

import br.org.scadabr.luciol.Configurable;
import br.org.scadabr.luciol.ihm.Labels;
import br.org.scadabr.protocol.iec101.common.AsduLocator;
import br.org.scadabr.protocol.iec101.common.BinaryTranslator;
import br.org.scadabr.protocol.iec101.common.CodeProperties;
import br.org.scadabr.protocol.iec101.common.DbLocator;
import br.org.scadabr.protocol.iec101.common.IEC608705Exception;
import br.org.scadabr.protocol.iec101.common.LinearTranslator;
import br.org.scadabr.protocol.iec101.common.MapAsdu;
import br.org.scadabr.protocol.iec101.common.MapSingleValue;
import br.org.scadabr.protocol.iec101.common.PeriodicService;
import br.org.scadabr.protocol.iec101.common.PhySerialLayer;
import br.org.scadabr.protocol.iec101.common.PhysicalLayer;
import br.org.scadabr.protocol.iec101.common.PrivateCot;
import br.org.scadabr.protocol.iec101.common.Translator;
import br.org.scadabr.protocol.iec101.common103.AsduKey103;
import br.org.scadabr.protocol.iec101.common103.FromBinaryFormatter103;
import br.org.scadabr.protocol.iec101.common103.PrivateAsdu103;
import br.org.scadabr.protocol.iec101.common103.PrivateFt103;
import br.org.scadabr.protocol.iec101.common103.PrivateIn103;
import br.org.scadabr.protocol.iec101.common103.PrivateRange103;
import br.org.scadabr.protocol.iec101.common103.ToBinaryFormatter103;
import br.org.scadabr.protocol.iec101.master.MAppManager;
import br.org.scadabr.protocol.iec101.master.MAppService;
import br.org.scadabr.protocol.iec101.master.MApplicationLayer;
import br.org.scadabr.protocol.iec101.master.MLinkLayer;
import br.org.scadabr.protocol.lib.xml.KXMLElement;

/**
 * <p>
 * This class defines common methods to instanciate mapping & private objects
 * from an xml configuration. And these to instanciate KXMLElement from
 * instanciated objects.
 * <p>
 * These mechanisms are OBLIGATORY for a Luciol application in order to be
 * compatible with HMI and central application manager.
 * 
 * @author lionnel cauvy
 * @version 1.0
 * @date 01.II.03
 */
public class MConfiguration103 implements Configurable {

	/*
	 * /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*\ < < > Inner
	 * classes. Communication & protocol settings. > < >
	 * \*\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
	 */

	/**
	 * This class defines communication settings.
	 * <p>
	 * 1O3 implementation fixes parity, data & stop bits. Configurable fields
	 * are comm port selection and baudrate.
	 */
	public class Communication {
		public String port = "unknown";
		public int baud = 9600;
		public final static int parity = 2;
		public final static int data = 8;
		public final static int stop = 1;

		/**
		 * Default constructor.
		 */
		public Communication() {
			port = "unknown";
			baud = 9600;
		}

		/**
		 *
		 */
		public Communication(String port, int baud) {
			this.port = port;
			this.baud = baud;
		}
	}

	/**
	 * This class defines protocol settings.
	 * <p>
	 * Timeout is a watchdog on remote answer.
	 * <p>
	 * Retries stand for allowed number of retries in transmission fail case.
	 * <p>
	 * Polling defines polling data class 2 cyclic period in seconds.
	 * <p>
	 * NbToPoll defines maximum number of data class 2 to poll during 1 cycle.
	 * <p>
	 * Synchro defines synchronization cyclic period in minutes.
	 * <p>
	 * Gi defines general interrogation cyclic period in minutes.
	 * <p>
	 */
	public class Protocol {
		public int timeout = 2;
		public int retries = 100;
		public int caa = 1;
		public int polling = 5;
		public int nbToPoll = 10;
		public int synchro = 60;
		public int gi = 60;

		public Protocol() {
		}

		public Protocol(int timeout, int retries, int caa, int polling,
				int nbToPoll, int synchro, int gi) {
			this.timeout = timeout;
			this.retries = retries;
			this.caa = caa;
			this.polling = polling;
			this.nbToPoll = nbToPoll;
			this.synchro = synchro;
			this.gi = gi;
		}
	}

	/*
	 * /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*\ < < >
	 * Variables > < <
	 * \*\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
	 */

	/**
	 * Communication reference.
	 */
	public Communication comm;
	/**
	 * Protocol reference.
	 */
	public Protocol protocol;
	/**
	 * Application layer reference.
	 */
	public MApplicationLayer app;
	/**
	 * Backup kXML Remote
	 */
	private KXMLElement[] backup;

	private String author = "l.cauvy";
	private String version = "1.0";
	private String date = "10.II.03";
	private String fileId = "unknown";

	public String getAuthor() {
		return author;
	}

	public String getVersion() {
		return version;
	}

	public String getDate() {
		return date;
	}

	public String getFileId() {
		return fileId;
	}

	/*
	 * /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*\ < < >
	 * Configurable Interface. HMI Luciol. > < >
	 * \*\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
	 */

	private static final String APPNAME = "Master103";

	/**
	 * Communication port parameter possible values.
	 */
	protected static final String[] POSSIBLE_PORT = new String[] { "COM1",
			"COM2", "UNKNOWN" };

	/**
	 * Baudrate parameter possible values.
	 */
	protected static final String[] POSSIBLE_SPEED = new String[] { "9600",
			"19200" };

	/**
	 * Return a unique name for this application.
	 */
	public String applicationName() {
		return APPNAME;
	}

	/**
	 * Return values of a SELECT type parameter.
	 * 
	 * @param param
	 *            a parameter number.
	 * @return values of a parameter.
	 */
	public String[] getPossibleValues(int param)
			throws IllegalArgumentException {
		switch (param) {
		case 0:
			return POSSIBLE_PORT;
		case 1:
			return POSSIBLE_SPEED;
		default:
			throw new IllegalArgumentException();
		}
	}

	/**
	 * Parameters names to display in luciol hmi
	 */
	private static final String[] OPTION_NAMES = new String[] {
			Labels.M103_COM_PORT, Labels.M103_BAUDRATE, Labels.M103_CAA,
			Labels.M103_TIMEOUT, Labels.M103_RETRIES, Labels.M103_POLLING,
			Labels.M103_NPOLL, Labels.M103_SYNCHRO, Labels.M103_GI };

	/**
	 * Return parameters names.
	 */
	public String[] getParameterNames() {
		return OPTION_NAMES;
	}

	/**
	 * Return parameters number.
	 */
	public int getParameterCount() {
		return OPTION_NAMES.length;
	}

	/**
	 * Parameters types.
	 */
	private static final byte[] OPTION_TYPES = new byte[] { SELECT, SELECT,
			CDATA, CDATA, CDATA, CDATA, CDATA, CDATA, CDATA };

	/**
	 * Return types of parameters.
	 * 
	 * @return the current type of a parameter.
	 */
	public byte[] getParameterTypes() {
		return OPTION_TYPES;
	}

	/**
	 * Return current values of parameters.
	 * 
	 * @return the current value of a parameter.
	 */
	public Object[] getParameterValues() {
		int port = 0;
		int baud = 0;
		while (!comm.port.toUpperCase().equals(POSSIBLE_PORT[port]))
			port++;
		while (!String.valueOf(comm.baud).equals(POSSIBLE_SPEED[baud]))
			baud++;

		return new Object[] { new Integer(port), new Integer(baud),
				String.valueOf(protocol.caa), String.valueOf(protocol.timeout),
				String.valueOf(protocol.retries),
				String.valueOf(protocol.polling),
				String.valueOf(protocol.nbToPoll),
				String.valueOf(protocol.synchro), String.valueOf(protocol.gi) };
	}

	/**
	 * <p>
	 * Test if <tt>newValue</tt> can be assigned as a new value for the
	 * parameter specified by <tt>num</tt>.
	 * 
	 * @param newValue
	 *            a new value for this parameter.
	 */
	public boolean[] setNewValues(Object[] values)
			throws IllegalArgumentException {
		boolean[] res = new boolean[values.length];
		int i = 0;
		// com port _ COM1 COM2
		String name = POSSIBLE_PORT[((Integer) values[i]).intValue()];
		comm.port = name.toLowerCase();
		res[i++] = true;
		// baud rate _ 9600 19200
		comm.baud = Integer.parseInt(POSSIBLE_SPEED[((Integer) values[i])
				.intValue()]);
		if (comm.baud == 9600 || comm.baud == 19200)
			res[i++] = true;
		else
			res[i++] = false;
		// caa
		protocol.caa = Integer.parseInt(((String) values[i]));
		if (protocol.caa > 0 && protocol.caa <= 254)
			res[i++] = true;
		else
			res[i++] = false;
		// timeout
		protocol.timeout = Integer.parseInt(((String) values[i]));
		if (protocol.timeout >= 1)
			res[i++] = true;
		else
			res[i++] = false;
		// retries
		protocol.retries = Integer.parseInt(((String) values[i]));
		if (protocol.retries >= 0)
			res[i++] = true;
		else
			res[i++] = false;
		// polling cycle
		protocol.polling = Integer.parseInt(((String) values[i]));
		if (protocol.polling >= 0)
			res[i++] = true;
		else
			res[i++] = false;
		// number of msg per polling cycle.
		protocol.nbToPoll = Integer.parseInt(((String) values[i]));
		if (protocol.nbToPoll >= 1)
			res[i++] = true;
		else
			res[i++] = false;
		// synchro cycle
		protocol.synchro = Integer.parseInt(((String) values[i]));
		if (protocol.synchro >= 0)
			res[i++] = true;
		else
			res[i++] = false;
		// gi cycle
		protocol.gi = Integer.parseInt(((String) values[i]));
		if (protocol.gi >= 0)
			res[i++] = true;
		else
			res[i++] = false;

		return res;
	}

	/*
	 * /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*\ < < > XML
	 * Interface. InputToXML, OutputFromXML > < <
	 * \*\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
	 */

	protected static final String ROOT = "Master103";
	protected static final String AUTHOR = "author";
	protected static final String VERSION = "version";
	protected static final String DATE = "date";
	protected static final String FILEID = "fileId";

	protected static final String COMMUNICATION = "Communication";
	protected static final String PORT = "port";
	protected static final String BAUDRATE = "baudrate";

	protected static final String PROTOCOL = "Protocol";
	protected static final String TIMEOUT = "timeOut";
	protected static final String RETRIES = "retries";
	protected static final String POLLING = "polling";
	protected static final String SYNCHRO = "synchro";
	protected static final String GI = "gi";
	protected static final String NBTOPOLL = "nbToPoll";
	protected static final String TRUE = "1";
	protected static final String FALSE = "0";

	protected static final String REMOTE = "Remote";
	protected static final String NAME = "name";
	protected static final String ADDRESS = "address";

	protected static final String PRIVATERANGE = "PrivateRange";
	protected static final String PRVTASDU = "PrvtAsdu";
	protected static final String PRVTELEMENT = "PrvtElement";
	protected static final String PRVTCOT = "PrvtCot";
	protected static final String PRVTIN = "PrvtIn";
	protected static final String PRVTFT = "PrvtFt";

	protected static final String MAPPING = "Mapping";

	protected final static String PROCEDURE = "Procedure";
	protected final static String RESETCU = "RESETCU";
	protected final static String RESETFCB = "RESETFCB";
	protected final static String SYNCHROPROC = "SYNCHRO";
	protected final static String GENERALINTERROGATION = "GI";
	protected final static String POLLINGPROC = "POLLING";
	protected final static String GENERALCOMMAND = "GENERALCOMMAND";
	protected final static String DCO = "dco";

	protected final static String PROCEDURES = "Procedures";
	protected static final String WRITERS = "Writers";
	// protected static final String READER = "Reader";
	// protected static final String READERKEY = "ReaderKey";
	// protected static final String ANSWERKEY = "AnswerKey";
	protected static final String ASDU = "Asdu";
	protected static final String KEY = "Key";
	protected static final String ID = "id";
	protected static final String VSQ = "vsq";
	protected static final String COT = "cot";
	protected static final String CAA = "caa";
	protected static final String FT = "ft";
	protected static final String IN = "in";

	protected final static String MAPVALUE = "MapValue";
	protected final static String PIO = "pio";
	protected final static String PIE = "pie";
	protected final static String PVAL = "pval";
	protected final static String DBREF = "db";
	protected final static String LABEL = "label";

	protected static final String LINEARTRANSLATOR = "LinearTranslator";
	protected static final String BINARYTRANSLATOR = "BinaryTranslator";
	protected static final String TOBINARYTRANSLATOR = "Formatter103ToBinary";
	protected static final String FROMBINARYTRANSLATOR = "FormatterBinaryTo103";
	protected static final String RATE = "rate";
	protected static final String OFFSET = "offset";
	protected static final String TYPE = "type";
	protected static final String THRESHOLD = "threshold";
	protected static final String LOGIC = "logic";
	protected static final String UPPERBOUND = "UPPERBOUND";
	protected static final String LOWERBOUND = "LOWERBOUND";
	protected static final String EQUALITY = "EQUALITY";
	protected static final String POSITIVE = "POSITIVE";
	protected static final String NEGATIVE = "NEGATIVE";

	/**
	 * Instanciate a MapSingleValue from XML configuration.
	 * 
	 * @param node
	 *            a KXMLElement MapValue.
	 * 
	 * @return an instance of a MapSingleValue.
	 */
	protected MapSingleValue inputMapValue(KXMLElement node) {
		AsduLocator locator = new AsduLocator(
				(short) node.getProperty(PIO, -1), (short) node.getProperty(
						PIE, -1), (short) node.getProperty(PVAL, -1));
		DbLocator db = new DbLocator(node.getProperty(DBREF, -1), node
				.getProperty(LABEL));
		Translator translator = null;

		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();)
			translator = inputTranslator((KXMLElement) e.nextElement());

		return new MapSingleValue(db, locator, translator);

	}

	/**
	 * Instanciate a KXMLElement from an instance of MapSingleValue.
	 * 
	 * @param node
	 *            an instance of a MapSingleValue.
	 * 
	 * @return a KXMLElement MapValue.
	 */
	// protected KXMLElement outputMapValue(MapSingleValue value)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(MAPVALUE);
	// node.addProperty(PIO,value.locator.pio);
	// node.addProperty(PIE,value.locator.pie);
	// node.addProperty(PVAL,value.locator.pval);
	// node.addProperty(DBREF,value.db.index);
	// node.addProperty(LABEL,value.db.label);
	// if (value.translator!=null)
	// node.addChild(outputTranslator(value.translator));
	// return node;
	// }

	/**
	 * Instanciate a Translator implementer from XML configuration.
	 * 
	 * @param node
	 *            a KXMLElement Translator.
	 * 
	 * @return a newly allocated Translator object.
	 */
	protected Translator inputTranslator(KXMLElement node) {
		String stype;
		byte type;
		String slogic;
		boolean logic;

		String name = node.getTagName();
		if (name.equals(LINEARTRANSLATOR)) {
			return new LinearTranslator(node.getProperty(RATE, -1), node
					.getProperty(OFFSET, -1));
		} else if (name.equals(BINARYTRANSLATOR)) {
			stype = node.getProperty(TYPE);
			if (stype.equals(UPPERBOUND))
				type = BinaryTranslator.UPPERBOUND;
			else if (stype.equals(LOWERBOUND))
				type = BinaryTranslator.LOWERBOUND;
			else
				type = BinaryTranslator.EQUALITY;
			slogic = node.getProperty(LOGIC);
			if (slogic.equals(POSITIVE))
				logic = true;
			else
				logic = false;
			float threshold = 0;
			if (CodeProperties.AJILE)
				threshold = node.getFloatProperty(THRESHOLD, (float) 0);
			else {
				String a = node.getProperty(THRESHOLD);
				threshold = java.lang.Float.parseFloat(a);
				System.out.println(new Float(threshold));
			}
			return new BinaryTranslator(type, threshold, logic);
		} else if (name.equals(FROMBINARYTRANSLATOR)) {
			slogic = node.getProperty(LOGIC);
			if (slogic.equals(POSITIVE))
				logic = true;
			else
				logic = false;
			return new FromBinaryFormatter103(logic);
		} else if (name.equals(TOBINARYTRANSLATOR)) {
			slogic = node.getProperty(LOGIC);
			if (slogic.equals(POSITIVE))
				logic = true;
			else
				logic = false;
			return new ToBinaryFormatter103(logic);
		}

		return null;
	}

	/**
	 * Instanciate a KXMLElement from an instance of Translator.
	 * 
	 * @param node
	 *            an instance of Translator.
	 * 
	 * @return a KXMLElement XXXXXTranslator.
	 */
	// protected KXMLElement outputTranslator(Translator translator)
	// {
	// KXMLElement node = new KXMLElement();
	//		
	// if (translator instanceof SMapranslator)
	// {
	// SMapranslator trans = (SMapranslator)translator;
	//			
	// node.setTagName(BINARYTRANSLATOR);
	//			
	// switch (trans.type)
	// {
	// case SMapranslator.UPPERBOUND:
	// node.addProperty(TYPE,UPPERBOUND);
	// break;
	// case SMapranslator.LOWERBOUND:
	// node.addProperty(TYPE,LOWERBOUND);
	// break;
	// case SMapranslator.EQUALITY:
	// node.addProperty(TYPE,EQUALITY);
	// break;
	// default:
	// break;
	// }
	// node.addProperty(THRESHOLD,Float.toString(trans.threshold));
	// node.addProperty(LOGIC,(trans.logic)?POSITIVE:NEGATIVE);
	// }
	// else if (translator instanceof Formatter103)
	// {
	// Formatter103 trans = (Formatter103)translator;
	//			
	// node.setTagName(FORMATTERTRANSLATOR);
	//			
	// switch (trans.type)
	// {
	// case Formatter103.UPPERBOUND:
	// node.addProperty(TYPE,UPPERBOUND);
	// break;
	// case Formatter103.LOWERBOUND:
	// node.addProperty(TYPE,LOWERBOUND);
	// break;
	// case Formatter103.EQUALITY:
	// node.addProperty(TYPE,EQUALITY);
	// break;
	// default:
	// break;
	// }
	// node.addProperty(THRESHOLD,Float.toString(trans.threshold));
	// node.addProperty(LOGIC,(trans.logic)?POSITIVE:NEGATIVE);
	// }
	// else if (translator instanceof LinearTranslator)
	// {
	// LinearTranslator trans = (LinearTranslator)translator;
	// node.setTagName(LINEARTRANSLATOR);
	// node.getProperty(RATE,trans.rate);
	// node.getProperty(OFFSET,trans.offset);
	// }
	// return node;
	// }

	/**
	 * Instanciate a Writer and its key from XML configuration.
	 * 
	 * @param node
	 *            a KXMLElement Writer.
	 * 
	 * @return an object array with Writer and related WriterKey.
	 */
	protected MapAsdu inputWriter(KXMLElement node) {
		AsduKey103 key = null;
		MapSingleValue[] value = new MapSingleValue[node.countChildren() - 1];
		int i = 0;
		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();) {
			KXMLElement child = (KXMLElement) e.nextElement();
			String name = child.getTagName();
			if (name.equals(KEY))
				key = new AsduKey103((short) child.getProperty(ID, -1),
						(short) child.getProperty(VSQ, -1), AsduKey103.NOTUSED,
						child.getProperty(CAA, -1), (short) child.getProperty(
								FT, -1), (short) child.getProperty(IN, -1));
			else
				value[i++] = inputMapValue(child);
		}
		return new MapAsdu(key, value);
	}

	/**
	 * Instanciate a KXMLElement from an instance of Writer and another of
	 * related key WriterKey.
	 * 
	 * @param node
	 *            an instance of Writer and one of WriterKey.
	 * 
	 * @return a KXMLElement Writer.
	 */
	// protected KXMLElement outputWriter(Writer writer, WriterKey103 key)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(WRITER);
	//		
	// KXMLElement knode = new KXMLElement();
	// knode.setTagName(WRITERKEY);
	// knode.addProperty(ID,key.id);
	// knode.addProperty(VSQ,key.vsq);
	// knode.addProperty(CAA,key.caa);
	// knode.addProperty(FT,key.ft);
	// knode.addProperty(IN,key.in);
	//		
	// node.addChild(knode);
	//		
	// for(int i=0;i<writer.value.length;)
	// node.addChild(outputMapValue(writer.value[i++]));
	//		
	// return node;
	// }

	/**
	 * Instanciate a Reader from XML configuration.
	 * 
	 * @param node
	 *            a KXMLElement Reader.
	 * 
	 * @return an instance of Reader.
	 */
	// protected Reader inputReader(KXMLElement node)
	// {
	// ReaderKey103 key = null;
	// ReaderKey103 exp = null;
	// MapSingleValue[] value = new MapSingleValue[node.countChildren()-2];
	// int i = 0;
	// for(Enumeration e = node.enumerateChildren();e.hasMoreElements();)
	// {
	// KXMLElement child = (KXMLElement)e.nextElement();
	// String name = child.getTagName();
	// if(name.equals(READERKEY))
	// key = new ReaderKey103(
	// (short)child.getProperty(ID,-1),
	// (short)child.getProperty(VSQ,-1),
	// (short)child.getProperty(COT,-1),
	// child.getProperty(CAA,-1),
	// (short)child.getProperty(FT,-1),
	// (short)child.getProperty(IN,-1)
	// );
	// else if (name.equals(ANSWERKEY))
	// exp = new ReaderKey103(
	// (short)child.getProperty(ID,-1),
	// (short)child.getProperty(VSQ,-1),
	// (short)child.getProperty(COT,-1),
	// child.getProperty(CAA,-1),
	// (short)child.getProperty(FT,-1),
	// (short)child.getProperty(IN,-1)
	// );
	// else
	// value[i++] = inputMapValue(child);
	// }
	// return new Reader(value,key,new MasterServiceHelper(exp));
	// }

	/**
	 * Instanciate a KXMLElement from an instance of Reader.
	 * 
	 * @param node
	 *            an instance of Reader..
	 * 
	 * @return a KXMLElement Reader.
	 */
	// protected KXMLElement outputReader(Reader reader)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(READER);
	// ReaderKey103 key = (ReaderKey103)reader.key;
	// KXMLElement knode = new KXMLElement();
	// knode.setTagName(READERKEY);
	// knode.addProperty(ID,key.id);
	// knode.addProperty(VSQ,key.vsq);
	// knode.addProperty(CAA,key.cot);
	// knode.addProperty(CAA,key.caa);
	// knode.addProperty(FT,key.ft);
	// knode.addProperty(IN,key.in);
	// node.addChild(knode);
	// ReaderKey103 answer =
	// (ReaderKey103)((MasterServiceHelper)reader.svc).expectedKey;
	// KXMLElement snode = new KXMLElement();
	// snode.setTagName(ANSWERKEY);
	// snode.addProperty(ID,answer.id);
	// snode.addProperty(VSQ,answer.vsq);
	// snode.addProperty(CAA,answer.cot);
	// snode.addProperty(CAA,answer.caa);
	// snode.addProperty(FT,answer.ft);
	// snode.addProperty(IN,answer.in);
	// node.addChild(snode);
	// for(int i=0;i<reader.value.length;)
	// node.addChild(outputMapValue(reader.value[i++]));
	// return node;
	// }

	/**
	 * <p>
	 * Instanciate a master application procedure from XML configuration.
	 * <p>
	 * Notice manager is passed as argument in order to fill 'address' &
	 * 'context' fields of such procedures.
	 * 
	 * @param node
	 *            A KXMLElement AppService.
	 * @param manager
	 *            The application layer manager.
	 * 
	 * @return an instance of MAppService.
	 */
	protected MAppService inputProcedure(KXMLElement node, MAppManager manager) {
		MAppService proc;
		String name = node.getProperty(TYPE);
		DbLocator db = new DbLocator(node.getProperty(DBREF, -1), node
				.getProperty(LABEL));
		if (name.equals(RESETCU))
			return new MResetCuAppSvc103(manager, db);
		else if (name.equals(RESETFCB))
			return new MResetFcbAppSvc103(manager, db);
		else if (name.equals(SYNCHROPROC))
			return new MSynchroAppSvc103(manager, db, protocol.caa);
		else if (name.equals(GENERALINTERROGATION))
			return new MGeneralInterrogationAppSvc103(manager, db, protocol.caa);
		else if (name.equals(POLLINGPROC))
			return new MPoll2AppSvc103(manager, db);
		else if (name.equals(GENERALCOMMAND)) {
			short ft = (short) node.getProperty(FT, -1);
			short in = (short) node.getProperty(IN, -1);
			byte dco = (byte) node.getProperty(DCO, -1);
			return new MGeneralCommandAppSvc103(manager, db, protocol.caa, ft,
					in, dco);
		}

		return null;
	}

	/**
	 * Instanciate a KXMLElement from an instance of AppService.
	 * 
	 * @param node
	 *            an instance of AppService.
	 * 
	 * @return a KXMLElement AppService.
	 */
	// protected KXMLElement outputProcedure(AppService procedure)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PROCEDURE);
	// node.addProperty(DBREF,procedure.notifier.index);
	// node.addProperty(LABEL,procedure.notifier.label);
	// if (procedure instanceof MResetCU103)
	// node.addProperty(TYPE,RESETCU);
	// else if (procedure instanceof MResetFCB103)
	// node.addProperty(TYPE,RESETFCB);
	// else if (procedure instanceof MSynchro103)
	// node.addProperty(TYPE,SYNCHROPROC);
	// else if (procedure instanceof MGeneralInterrogation103)
	// node.addProperty(TYPE,GENERALINTERROGATION);
	// else if (procedure instanceof MPolling103)
	// node.addProperty(TYPE,POLLINGPROC);
	// else if (procedure instanceof MGeneralCommand103)
	// {
	// MGeneralCommand103 cd = (MGeneralCommand103)procedure;
	// node.addProperty(TYPE,GENERALCOMMAND);
	// node.addProperty(FT,cd.ft);
	// node.addProperty(IN,cd.in);
	// node.addProperty(DCO,cd.dco);
	// }
	// return node;
	// }

	/**
	 * Instanciate a Mapping from XML configuration.
	 * 
	 * @param node
	 *            A KXMLElement Mapping.
	 * 
	 * @return an instance of Mapping.
	 */
	protected MMapping103 inputMapping(KXMLElement node, MAppManager manager) {

		MMapping103 map = new MMapping103(manager);

		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();) {
			KXMLElement child = (KXMLElement) e.nextElement();
			if (child.getTagName().equals(WRITERS)) {
				for (Enumeration e2 = child.enumerateChildren(); e2
						.hasMoreElements();) {
					map.addWriter(inputWriter((KXMLElement) e2.nextElement()));
				}
			} else if (child.getTagName().equals(PROCEDURES)) {
				for (Enumeration e2 = child.enumerateChildren(); e2
						.hasMoreElements();) {
					map.addProcedure(inputProcedure((KXMLElement) e2
							.nextElement(), manager));
				}
			}
		}

		return map;
	}

	/**
	 * Instanciate a KXMLElement from an instance of Mapping.
	 * 
	 * @param node
	 *            An instance of Mapping.
	 * 
	 * @return a KXMLElement Mapping.
	 */
	// protected KXMLElement outputMapping(Mapping103 map)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(MAPPING);
	//		
	// Enumeration wkeys = map.enumerateWriterKeys();
	// Enumeration writers = map.enumerateWriters();
	// Enumeration readers = map.enumerateReaders();
	// Enumeration procedures = map.enumerateProcedures();
	//		
	// for(;wkeys.hasMoreElements();)
	// node.addChild(outputWriter((Writer)writers.nextElement(),(WriterKey103)wkeys.nextElement()));
	// for(;readers.hasMoreElements();)
	// node.addChild(outputReader((Reader)readers.nextElement()));
	// for(;procedures.hasMoreElements();)
	// node.addChild(outputProcedure((AppService)procedures.nextElement()));
	//		
	// return node;
	// }

	/**
	 * Instanciate a private asdu from xml configuration.
	 * 
	 * @param node
	 *            a KXMLElement PrvtAsdu.
	 * 
	 * @return an instance of PrvtAsdu.
	 */
	protected PrivateAsdu103 inputPrvtAsdu(KXMLElement node) {
		int i = 0;
		String[] ie = new String[node.countChildren()];
		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();) {
			KXMLElement child = (KXMLElement) e.nextElement();
			// WARN: KXMLElement REHASH!!! WE SHOULD ADD A POSITION FLAG
			ie[i++] = child.getProperty(TYPE);
		}
		return new PrivateAsdu103(node.getProperty(LABEL), (short) node
				.getProperty(ID, -1), (short) node.getProperty(VSQ, -1),
				(short) node.getProperty(COT, -1), node.getProperty(CAA, -1),
				(short) node.getProperty(FT, -1), (short) node.getProperty(IN,
						-1), ie);
	}

	/**
	 * Instanciate a KXMLElement from an instance of private asdu.
	 * 
	 * @param asdu
	 *            an instance of PrivateAsdu.
	 * 
	 * @return a KXMLElement PrvtAsdu.
	 */
	// protected KXMLElement outputPrvtAsdu(PrivateAsdu asdu)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PRVTASDU);
	// node.addProperty(ID,asdu.id);
	// node.addProperty(LABEL,asdu.label);
	// for(int i=0;i<asdu.ie.length;)
	// {
	// KXMLElement child = new KXMLElement();
	// child.setTagName(PRVTELEMENT);
	// child.addProperty(TYPE,asdu.ie[i++]);
	// node.addChild(child);
	// }
	// return node;
	// }

	/**
	 * Instanciate a Private cause of trnasmission from xml configuration.
	 * 
	 * @param node
	 *            a KXMLElement PrvtCot
	 * 
	 * @return an instance of PrivateCot
	 */
	protected PrivateCot inputPrvtCot(KXMLElement node) {
		return new PrivateCot((short) node.getProperty(ID, -1), node
				.getProperty(LABEL));
	}

	/**
	 * Instanciate a KXMLElement from an instance of Private cause of
	 * transmission..
	 * 
	 * @param in
	 *            an instance of PrvtCot.
	 * 
	 * @return a KXMLElement PrivateCot.
	 */
	// protected KXMLElement outputPrvtCot(PrivateCot cot)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PRVTCOT);
	// node.addProperty(ID,cot.id);
	// node.addProperty(LABEL,cot.label);
	// return node;
	// }

	/**
	 * Instanciate a Private information number from xml configuration.
	 * 
	 * @param node
	 *            a KXMLElement PrvtIn
	 * 
	 * @return an instance of PrivateIn
	 */
	protected PrivateIn103 inputPrvtIn(KXMLElement node) {
		return new PrivateIn103((short) node.getProperty(ID, -1), node
				.getProperty(LABEL));
	}

	/**
	 * Instanciate a KXMLElement from an instance of Private information number.
	 * 
	 * @param in
	 *            an instance of PrivateIn.
	 * 
	 * @return a KXMLElement PrvtIn
	 */
	// protected KXMLElement outputPrvtIn(PrivateIn103 in)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PRVTIN);
	// node.addProperty(ID,in.id);
	// node.addProperty(LABEL,in.label);
	// return node;
	// }

	/**
	 * Instanciate a Private function type from xml configuration.
	 * 
	 * @param node
	 *            a KXMLElement PrvtFt
	 * 
	 * @return an instance of PrivateFt
	 */
	protected PrivateFt103 inputPrvtFt(KXMLElement node) {
		return new PrivateFt103((short) node.getProperty(ID, -1), node
				.getProperty(LABEL));
	}

	/**
	 * Instanciate a KXMLElement from an instance of Private function type.
	 * 
	 * @param ft
	 *            an instance of PrivateFt103.
	 * 
	 * @return a KXMLElement PrvtFt
	 */
	// protected KXMLElement outputPrvtFt(PrivateFt103 ft)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PRVTFT);
	// node.addProperty(ID,ft.id);
	// node.addProperty(LABEL,ft.label);
	// return node;
	// }

	/**
	 * Instanciate a private range from xml configuration.
	 * 
	 * @param node
	 *            PrivateRange KXMLElement.
	 * 
	 * @return an instance of PrivateRange.
	 */
	protected PrivateRange103 inputPrivateRange(KXMLElement node) {
		PrivateRange103 prange = new PrivateRange103();
		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();) {
			KXMLElement child = (KXMLElement) e.nextElement();
			String name = child.getTagName();
			if (name.equals(PRVTASDU))
				prange.addAsdu(inputPrvtAsdu(child));
			else if (name.equals(PRVTCOT))
				prange.addCot(inputPrvtCot(child));
			else if (name.equals(PRVTIN))
				prange.addIn(inputPrvtIn(child));
			else if (name.equals(PRVTFT))
				prange.addFt(inputPrvtFt(child));
			// else
			// throw new
			// IllegalArgumentException("PrivateRange has unknown sub element");
		}
		return prange;
	}

	/**
	 * Instanciate a KXMLElement from an instance of private range.
	 * 
	 * @param node
	 *            an instance of PrivateRange.
	 * 
	 * @return PrivateRange KXMLElement.
	 */
	// protected KXMLElement outputPrivateRange(PrivateRange103 prange)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PRIVATERANGE);
	// for(Enumeration e = prange.enumerateAsdus();e.hasMoreElements();)
	// node.addChild(outputPrvtAsdu((PrivateAsdu)e.nextElement()));
	// for(Enumeration e = prange.enumerateCots();e.hasMoreElements();)
	// node.addChild(outputPrvtCot((PrivateCot)e.nextElement()));
	// for(Enumeration e = prange.enumerateFts();e.hasMoreElements();)
	// node.addChild(outputPrvtFt((PrivateFt103)e.nextElement()));
	// for(Enumeration e = prange.enumerateIns();e.hasMoreElements();)
	// node.addChild(outputPrvtIn((PrivateIn103)e.nextElement()));
	// return node;
	// }

	/**
	 * Instanciate an application manager from kXML configuration.
	 * 
	 * @param node
	 *            a KXMLElement Remote.
	 * 
	 * @return an instance of AppManager.
	 */
	protected MAppManager inputRemote(KXMLElement node) {
		MContext103 context = new MContext103(node.getProperty(ADDRESS, -1),
				node.getTagName());
		MAppManager manager = new MAppManager(app, context, protocol.nbToPoll);
		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();) {
			KXMLElement child = (KXMLElement) e.nextElement();
			String name = child.getTagName();
			if (name.equals(PRIVATERANGE))
				manager.prange = inputPrivateRange(child);
			else if (name.equals(MAPPING)) {
				manager.map = inputMapping(child, manager);
				// manager.map.manager = manager;
			}
		}
		// periodic application procedures
		manager.periodic = new Vector();
		int duration;
		if ((duration = 60000 * protocol.gi) > 0)
			manager.periodic.addElement(new PeriodicService(duration,
					new MGeneralInterrogationAppSvc103(manager, null,
							protocol.caa), manager));
		if ((duration = 60000 * protocol.synchro) > 0)
			manager.periodic
					.addElement(new PeriodicService(duration,
							new MSynchroAppSvc103(manager, null, protocol.caa),
							manager));
		if ((duration = 60000 * protocol.polling) > 0)
			manager.periodic.addElement(new PeriodicService(duration,
					new MPoll2AppSvc103(manager, null), manager));

		// initial procedures
		MAppService proc;
		manager.initial = new MAppService[3];
		manager.initial[0] = new MResetCuAppSvc103(manager, null);
		manager.initial[1] = new MSynchroAppSvc103(manager, null, protocol.caa);
		manager.initial[2] = new MGeneralInterrogationAppSvc103(manager, null,
				protocol.caa);
		return manager;
	}

	/**
	 * Instanciate a KXMLElement from an instance of AppManager.
	 * 
	 * @param mgr
	 *            an instance of AppManager
	 * 
	 * @return a KXMLElement Remote.
	 */
	// protected KXMLElement outputRemote(MAppManager103 manager)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(REMOTE);
	// node.addProperty(ADDRESS,manager.address);
	// node.addProperty(NAME,manager.name);
	// if(manager.prange!=null)
	// node.addChild(outputPrivateRange((PrivateRange103)manager.prange));
	// if(manager.map!=null)
	// node.addChild(outputMapping((Mapping103)manager.map));
	// return node;
	// }

	/**
	 * Instanciate a Protocol for general application settings from a
	 * KXMLElement.
	 * 
	 * @param node
	 *            a KXMLElement Protocol
	 * 
	 * @return an instance of Protocol.
	 */
	protected Protocol inputProtocol(KXMLElement node) {
		return new Protocol(node.getProperty(TIMEOUT, 1), node.getProperty(
				RETRIES, 1), node.getProperty(CAA, 1), node.getProperty(
				POLLING, 5), node.getProperty(NBTOPOLL, 1), node.getProperty(
				SYNCHRO, 0), node.getProperty(GI, 0));
	}

	/**
	 * Instanciate KXMLElement from an instance of Protocol.
	 * 
	 * @param protocol
	 *            an instance of Protocol.
	 * 
	 * @return a KXMLElement Protocol.
	 */
	protected KXMLElement outputProtocol(Protocol protocol) {
		KXMLElement node = new KXMLElement();
		node.setTagName(PROTOCOL);
		node.addProperty(CAA, protocol.caa);
		node.addProperty(TIMEOUT, protocol.timeout);
		node.addProperty(RETRIES, protocol.retries);
		node.addProperty(POLLING, protocol.polling);
		node.addProperty(NBTOPOLL, protocol.nbToPoll);
		node.addProperty(SYNCHRO, protocol.synchro);
		node.addProperty(GI, protocol.gi);
		return node;
	}

	/**
	 * Instanciate a Communication for general comm settings from a KXMLElement.
	 * 
	 * @param node
	 *            a KXMLElement Communication.
	 * 
	 * @return an instance of Communication.
	 */
	protected Communication inputCommunication(KXMLElement node) {
		return new Communication(node.getProperty(PORT).toLowerCase(), node
				.getProperty(BAUDRATE, 9600));

	}

	/**
	 * Instanciate a KXMLElement from an instance of Communication. from a
	 * KXMLElement.
	 * 
	 * @param comm
	 *            an instance of Communication.
	 * 
	 * @return a KXMLElement Communication.
	 */
	protected KXMLElement outputCommunication(Communication comm) {
		KXMLElement node = new KXMLElement();
		node.setTagName(COMMUNICATION);
		node.addProperty(PORT, comm.port.toUpperCase());
		node.addProperty(BAUDRATE, comm.baud);
		return node;
	}

	/**
	 * Update a whole application configuration from a KXMLElement node.
	 * 
	 * @param node
	 *            a KXMLElement ROOT.
	 */
	public void inputFromXML(KXMLElement node) {
		if (!node.getTagName().equals(ROOT))
			return;

		author = node.getProperty(AUTHOR);
		version = node.getProperty(VERSION);
		date = node.getProperty(DATE);
		fileId = node.getProperty(FILEID);

		int size = node.countChildren() - 2;
		int i = 0;
		MLinkLayer lnk = null;
		PhysicalLayer phy = null;
		backup = new KXMLElement[size];
		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();) {
			KXMLElement child = (KXMLElement) e.nextElement();
			String name = child.getTagName();
			if (name.equals(COMMUNICATION)) {
				comm = inputCommunication(child);
			} else if (name.equals(PROTOCOL)) {
				protocol = inputProtocol(child);

				app = new MApplicationLayer(size);
				lnk = new MLinkLayer(size);
				phy = new PhySerialLayer(protocol.retries, protocol.timeout,
						comm.port, comm.baud, comm.data, comm.stop, comm.parity);
				app.reader = new MAsduReader103();
				app.writer = new MAsduWriter103();
				app.lower = lnk;
				lnk.upper = app;
				lnk.lower = phy;
				phy.upper = lnk;
			} else if (name.equals(REMOTE)) {
				backup[i++] = child;
				MAppManager manager = inputRemote(child);
				Integer address = new Integer(manager.context.address);
				manager.layer = app;
				app.manager.put(address, manager);
				lnk.manager.put(address, manager.lnk);
			}
		}
	}

	/**
	 * Instanciate a KXMLElement ROOT from an configuration.
	 */
	public KXMLElement outputToXML() {
		KXMLElement node = new KXMLElement();
		node.setTagName(ROOT);

		node.addProperty(AUTHOR, author);
		node.addProperty(VERSION, version);
		node.addProperty(DATE, date);
		node.addProperty(FILEID, fileId);

		node.addChild(outputCommunication(comm));
		node.addChild(outputProtocol(protocol));
		// for(Enumeration e = app.enumerateManagers();e.hasMoreElements();)
		// node.addChild(outputRemote((MAppManager103)e.nextElement()));
		for (int i = 0; i < backup.length;)
			node.addChild(backup[i++]);
		return node;
	}

	/**
	 * Kills all threads in current session.
	 */
	public void kill() {
		if (app != null) {
			app.kill();
			try {
				app.join();
			} catch (Exception e) {
				CodeProperties.log("M103 killed.");
			}
			CodeProperties.log("M103 really killed.");
		}
	}

	/**
	 * starts application
	 */
	public void start() {
		if (!comm.port.equals("unknown")) {
			app.start();
		}
		// if (!ACTIVE.comm.port.equals("UNKNOWN"))
		// {
		// // instanciates a network configuration
		// SlaveNetwork103 network = new
		// SlaveNetwork103(ACTIVE.getRemotes(),ACTIVE.protocol.polling,ACTIVE.protocol.synchro,ACTIVE.protocol.gi,ACTIVE.protocol.nbToPoll);
		// // instanciates an application layer for this app.
		// application = new MasterApplicationLayer(
		// ACTIVE.map,
		// new MasterAsduReader103(ACTIVE.prange),
		// new MasterAsduWriter103(),
		// network,
		// network,
		// ACTIVE.comm.port,
		// ACTIVE.comm.baudrate,
		// ACTIVE.comm.data,
		// ACTIVE.comm.stop,
		// ACTIVE.comm.parity,
		// ACTIVE.protocol.timeout,
		// ACTIVE.protocol.retries
		// );
		//			
		// try
		// {
		// work();
		// }
		// catch(IEC608705Exception e)
		// {
		// CodeProperties.log("CEI Exception: "+e);
		// }
		// }
		// else
		// CodeProperties.log("Stopping.");
	}

	public void work() throws IEC608705Exception {
		// CodeProperties.log("Application starting.");
		// application.start();
	}

	/*
	 * /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*\ < < >
	 * Constructor > < <
	 * \*\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
	 */
	/**
	 * Instanciate a new configurator for master 103 application.
	 */
	public MConfiguration103() {
		comm = new Communication();
		protocol = new Protocol();
		app = new MApplicationLayer(0);
		backup = new KXMLElement[0];
	}
} // end MConfiguration103.java