/*
 * Class    br.org.scadabr.protocol.iec101.slave.SMapping
 * File     SMapping.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.slave;

import br.org.scadabr.luciol.database.DataElement;
import br.org.scadabr.protocol.iec101.common.AppService;
import br.org.scadabr.protocol.iec101.common.MapSingleValue;
import br.org.scadabr.protocol.iec101.common.Mapping;

import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * <p>
 * This class defines a mapping for slave application.
 * <p>
 * This inheriting class specializes Mapping object adding:
 * <p>
 * - measure asdus to poll periodically.
 * <p>
 * - general interrogation asdus to return on related master
 * request.
 * <p>
 * - event asdus to enable triggering of spontaneous events from
 * database notification.
 *
 * @author lionnel cauvy
 * @version 1.0
 * @date 01.II.03
 */
 
public abstract class SMapping extends Mapping
{
	/**
	 * Measure vector.
	 */
	private Vector measure;
	/**
	 * General interrogation vector.
	 */
	private Vector gi;
	/**
	 * Table to map database reference to spontaneous events.
	 */
	private Hashtable Map_DbRef_Event;
	/**
	 * This event is raised on database notification.
	 *
	 * @param source	The element from database that has changed.
	 */	
	public void dataEventPerformed(DataElement source)
	{
		// reject if invalid.
		if(!source.lastRecord().isValid())
			return;	
		// spontaneous event application service
		Vector vector = (Vector)Map_DbRef_Event.get(new Integer(source.databaseID));
		for(Enumeration e =vector.elements();e.hasMoreElements();)
		{
			AppService svc = (AppService)e.nextElement();
			resolveMapping(svc);
			((SAppManager)manager).event(svc);
		}		
	}
	/**
	 * Allows to get data in luciol central database for a special
	 * application procedure from its defined mapping.
	 *
	 * @param proc	The procedure to complete with data elements from luciol database.
	 */
	public void resolveMapping(AppService proc)
	{
		MapSingleValue[] value = proc.asdu.value;
//		InformationElement[] ie = svc.asdu.ie;
		for(int i=0;i<value.length;i++)
		{
			value[i].mapped = read(value[i].db.index);
			if (value[i].translator!=null)
				value[i].mapped = value[i].translator.apply(value[i].mapped);
//			ie[value[i].locator.pie].setData(value[i].locator.pval,val);
		}
	}
	/**
	 * Read data value at a database index and return it as an Object.
	 *
	 * @param index	The database area to read
	 *
	 * @return read value as a java.lang.Object
	 */
	public Object read(int index)
	{
		DataElement element = getDataElement(index);
		if(element==null)
			return null;
		return element.lastRecord().objectValue();
	}

	/**
	 * Add an application procedure, and by the way, an asdu, to
	 * answer on master general interrogation request.
	 *
	 * @param proc	The procedure to store.
	 */	
	public void addGi(AppService proc)
	{
		gi.addElement(proc);
	}
	/**
	 * Add an application procedure, and by the way, an asdu, to
	 * return on master data class 2 polling procedure request.
	 *
	 * @param proc	The procedure to record.
	 */	
	public void addMeasure(AppService asdu)
	{
		measure.addElement(asdu);
	}
	/**
	 * Get accessor on number of different asdus that are 'measurand' ones.
	 *
	 * @return number of different measures performed by this application.
	 */
	public int measureSize()
	{
		return measure.size();
	}
	/**
	 * Add a 'spontaneous event' asdu to map. Same db index reference
	 * can trigger several different events. So events are 
	 *
	 * @param event	The event to map to database.
	 */
	public void addEvent(AppService event)
	{
		int index = event.asdu.value[0].db.index;
		Integer id = new Integer(index);
		Vector events = null;
		if ((events = (Vector)Map_DbRef_Event.get(id))==null)
		{
			events = new Vector();
			recordEventListener(index);
		}
		events.addElement(event);
		Map_DbRef_Event.put(id,events);
	}	
	/**
	 * Get accessor on all general interrogation asdus recorded
	 * in this mapping.
	 *
	 * @return an enumeration of asdus.
	 */	
	public Enumeration enumerateGis()
	{
		return gi.elements();
	}
	/**
	 * Get accessor on all measure asdus recorded
	 * in this mapping.
	 *
	 * @return an enumeration of asdus.
	 */
	public Enumeration enumerateMeasures()
	{
		return measure.elements();
	}
	/**
	 * Allocates a mapping object for slave application.
	 *
	 * @param manager	Application layer manager.
	 */	
	protected SMapping(SAppManager manager)
	{
		super(manager);
		Map_DbRef_Event = new Hashtable();
		gi = new Vector();
		measure = new Vector();
	}	
}	// end SMapping.java