/*
 * Class    br.org.scadabr.protocol.slave101.SConfiguration101
 * File     SConfiguration101.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.slave101;

import java.util.Enumeration;

import br.org.scadabr.luciol.Configurable;
import br.org.scadabr.luciol.ihm.Labels;
import br.org.scadabr.protocol.iec101.common.AsduLocator;
import br.org.scadabr.protocol.iec101.common.BinaryTranslator;
import br.org.scadabr.protocol.iec101.common.CodeProperties;
import br.org.scadabr.protocol.iec101.common.DbLocator;
import br.org.scadabr.protocol.iec101.common.IEC608705Exception;
import br.org.scadabr.protocol.iec101.common.LinearTranslator;
import br.org.scadabr.protocol.iec101.common.MapSingleValue;
import br.org.scadabr.protocol.iec101.common.PhysicalLayer;
import br.org.scadabr.protocol.iec101.common.Translator;
import br.org.scadabr.protocol.iec101.common101.AsduKey101;
import br.org.scadabr.protocol.iec101.common101.FromBinaryFormatter101;
import br.org.scadabr.protocol.iec101.common101.ToBinaryFormatter101;
import br.org.scadabr.protocol.iec101.slave.SApplicationLayer;
import br.org.scadabr.protocol.iec101.slave.SDataClass1AppSvc;
import br.org.scadabr.protocol.iec101.slave.SDataClass2AppSvc;
import br.org.scadabr.protocol.iec101.slave.SLinkLayer;
import br.org.scadabr.protocol.lib.xml.KXMLElement;

/**
 * <p>
 * This class defines common methods to instanciate mapping & private objects
 * from an xml configuration. And these to instanciate KXMLElement from
 * instanciated objects.
 * <p>
 * These mechanisms are OBLIGATORY for a Luciol application in order to be
 * compatible with HMI and central application manager.
 * 
 * @author lionnel cauvy
 * @version 1.0
 * @date 01.II.03
 */
public class SConfiguration101 implements Configurable {

	/*
	 * /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*\ < < > Inner
	 * classes. Communication & protocol settings. > < >
	 * \*\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
	 */

	/**
	 * This class defines communication settings.
	 * <p>
	 * 1O3 implementation fixes parity, data & stop bits. Configurable fields
	 * are comm port selection and baudrate.
	 */
	public class Communication {
		public String port = "unknown";
		public int baud = 9600;
		public final static int parity = 2;
		public final static int data = 8;
		public final static int stop = 1;

		/**
		 * Default constructor.
		 */
		public Communication() {
			port = "unknown";
			baud = 9600;
		}

		/**
		 *
		 */
		public Communication(String port, int baud) {
			this.port = port;
			this.baud = baud;
		}
	}

	/**
	 * This class defines protocol settings.
	 * <p>
	 * <Protocol clock='1' extendedTimeTag='1' counterResetTimeOut caa='1'
	 * address='1'>
	 */
	public class Protocol {
		public int polling = 5;
		public boolean clock = true;
		public boolean extendedTime = true;
		public int counterReset = 1000;
		public int caa = 1;
		public int address = 1;
		public byte lnkSz = 0x01;
		public byte cotSz = 0x01;
		public byte caaSz = 0x01;
		public byte ioaSz = 0x01;
	}

	/*
	 * /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*\ < < >
	 * Variables > < <
	 * \*\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
	 */

	/**
	 * Communication reference.
	 */
	public Communication comm;
	/**
	 * Protocol reference.
	 */
	public Protocol protocol;
	/**
	 * Application layer reference.
	 */
	public SApplicationLayer app;
	/**
	 * Backup kXML mapping
	 */
	private KXMLElement backupMap;
	/**
	 *
	 */
	private String author = "l.cauvy";
	/**
	 *
	 */
	private String version = "1.0";
	/**
	 *
	 */
	private String date = "10.IV.03";

	private String fileId = "unknown";

	public String getAuthor() {
		return author;
	}

	public String getVersion() {
		return version;
	}

	public String getFileId() {
		return fileId;
	}

	public String getDate() {
		return date;
	}

	/*
	 * /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*\ < < >
	 * Configurable Interface. HMI Luciol. > < >
	 * \*\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
	 */

	private static final String APPNAME = "Slave101";

	/**
	 * Communication port parameter possible values.
	 */
	protected static final String[] POSSIBLE_PORT = new String[] { "COM1",
			"COM2", "UNKNOWN" };

	/**
	 * Baudrate parameter possible values.
	 */
	protected static final String[] POSSIBLE_SPEED = new String[] {
	/*
	 * "100","110","200","300","600","1200","2400",
	 * "4800","9600","14400","19200","38400","56000",
	 * "57600","64000","115200","128000","256000"
	 */
	"9600", "19200" };

	/**
	 * 1 or 2 possible values.
	 */
	protected static final String[] SZ_12 = new String[] { "1", "2" };

	/**
	 * 1 or 2 or 3 possible values.
	 */
	protected static final String[] SZ_123 = new String[] { "1", "2", "3" };

	/**
	 * Return a unique name for this application.
	 */
	public String applicationName() {
		return APPNAME;
	}

	/**
	 * Return values of a SELECT type parameter.
	 * 
	 * @param param
	 *            a parameter number.
	 * @return values of a parameter.
	 */
	public String[] getPossibleValues(int param)
			throws IllegalArgumentException {
		switch (param) {
		case 0:
			return POSSIBLE_PORT;
		case 1:
			return POSSIBLE_SPEED;
		case 6:
		case 7:
		case 8:
			return SZ_12;
		case 9:
			return SZ_123;
		default:
			throw new IllegalArgumentException();
		}
	}

	/**
	 * Parameters names to display in luciol hmi
	 */
	private static final String[] OPTION_NAMES = new String[] {
			Labels.S101_COM_PORT, Labels.S101_BAUDRATE, Labels.S101_ADDRESS,
			Labels.S101_POLLING, Labels.S101_CLOCK,/*
													 * Labels.S101_EXTENDEDTIME,
													 * Labels.S101_COUNTERTIME,
													 */Labels.S101_CAA,
			Labels.S101_LNKSZ, Labels.S101_COTSZ, Labels.S101_CAASZ,
			Labels.S101_IOASZ };

	/**
	 * Return parameters names.
	 */
	public String[] getParameterNames() {
		return OPTION_NAMES;
	}

	/**
	 * Return parameters number.
	 */
	public int getParameterCount() {
		return OPTION_NAMES.length;
	}

	/**
	 * Parameters types.
	 */
	private static final byte[] OPTION_TYPES = new byte[] { SELECT, SELECT,
			CDATA, CDATA, CHECK,/* CHECK,CDATA, */CDATA, SELECT, SELECT,
			SELECT, SELECT };

	/**
	 * Return types of parameters.
	 * 
	 * @return the current type of a parameter.
	 */
	public byte[] getParameterTypes() {
		return OPTION_TYPES;
	}

	/**
	 * Return current values of parameters.
	 * 
	 * @return the current value of a parameter.
	 */
	public Object[] getParameterValues() {
		int port = 0;
		int baud = 0;
		int lnk = 0;
		int cot = 0;
		int caa = 0;
		int ioa = 0;
		while (!comm.port.toUpperCase().equals(POSSIBLE_PORT[port]))
			port++;
		while (!String.valueOf(comm.baud).equals(POSSIBLE_SPEED[baud]))
			baud++;
		while (!String.valueOf(protocol.lnkSz).equals(SZ_12[lnk]))
			lnk++;
		while (!String.valueOf(protocol.cotSz).equals(SZ_12[cot]))
			cot++;
		while (!String.valueOf(protocol.caaSz).equals(SZ_12[caa]))
			caa++;
		while (!String.valueOf(protocol.caaSz).equals(SZ_123[ioa]))
			ioa++;

		return new Object[] { new Integer(port), new Integer(baud),
				String.valueOf(protocol.address),
				String.valueOf(protocol.polling),
				new Boolean(protocol.clock),
				// new Boolean(protocol.extendedTime),
				// String.valueOf(protocol.counterReset),
				String.valueOf(protocol.caa), new Integer(lnk),
				new Integer(cot), new Integer(caa), new Integer(ioa) };
	}

	/**
	 * <p>
	 * Test if <tt>newValue</tt> can be assigned as a new value for the
	 * parameter specified by <tt>num</tt>.
	 * 
	 * @param newValue
	 *            a new value for this parameter.
	 */
	public boolean[] setNewValues(Object[] values)
			throws IllegalArgumentException {
		boolean[] res = new boolean[values.length];
		int i = 0;
		// com port _ COM1 COM2
		String name = POSSIBLE_PORT[((Integer) values[i]).intValue()];
		comm.port = name.toLowerCase();
		res[i++] = true;
		// baud rate _ 9600 19200
		comm.baud = Integer.parseInt(POSSIBLE_SPEED[((Integer) values[i])
				.intValue()]);
		if (comm.baud == 9600 || comm.baud == 19200)
			res[i++] = true;
		else
			res[i++] = false;
		// address
		protocol.address = Integer.parseInt(((String) values[i]));
		if (protocol.address > 0 && protocol.address < 255)
			res[i++] = true;
		else
			res[i++] = false;
		// polling cycle
		protocol.polling = Integer.parseInt(((String) values[i]));
		if (protocol.polling >= 0)
			res[i++] = true;
		else
			res[i++] = false;
		// change system clock permission
		protocol.clock = ((Boolean) values[i]).booleanValue();
		res[i++] = true;
		// CP56Time2a
		// protocol.extendedTime = ((Boolean)values[i]).booleanValue();
		// res[i++] = true;
		// // counter reset
		// protocol.counterReset = Integer.parseInt(((String)values[i]));
		// if(protocol.counterReset >= 0)
		// res[i++] = true;
		// else
		// res[i++] = false;
		// common address of asdus
		protocol.caa = Integer.parseInt(((String) values[i]));
		if (protocol.caa > 0 && protocol.caa < 255)
			res[i++] = true;
		else
			res[i++] = false;
		// Link address field size
		protocol.lnkSz = Byte
				.parseByte(SZ_12[((Integer) values[i]).intValue()]);
		res[i++] = true;
		// Cause of transmission field size
		protocol.cotSz = Byte
				.parseByte(SZ_12[((Integer) values[i]).intValue()]);
		res[i++] = true;
		// Common address of asdus field size
		protocol.caaSz = Byte
				.parseByte(SZ_12[((Integer) values[i]).intValue()]);
		res[i++] = true;
		// Information object address field size
		protocol.ioaSz = Byte
				.parseByte(SZ_123[((Integer) values[i]).intValue()]);
		res[i++] = true;

		return res;
	}

	/*
	 * /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*\ < < > XML
	 * Interface. InputToXML, OutputFromXML > < <
	 * \*\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
	 */

	protected static final String ROOT = "Slave101";
	protected static final String AUTHOR = "author";
	protected static final String VERSION = "version";
	protected static final String DATE = "date";

	protected static final String COMMUNICATION = "Communication";
	protected static final String PORT = "port";
	protected static final String BAUDRATE = "baudrate";

	protected static final String PROTOCOL = "Protocol";
	protected static final String ADDRESS = "address";
	protected static final String COUNTERTIME = "counterTimeReset";
	protected static final String POLLING = "polling";
	protected static final String CLOCK = "clock";
	protected static final String EXTENDEDTIME = "extendedTimeTag";
	protected static final String OA = "oa";
	protected static final String IOA = "ioa";
	protected static final String FIELDSIZE = "FieldSize";
	protected static final String CAASZ = "caaSz";
	protected static final String IOASZ = "ioaSz";
	protected static final String LNKSZ = "lnkSz";
	protected static final String COTSZ = "cotSz";
	protected static final String TRUE = "1";
	protected static final String FALSE = "0";

	protected static final String PRIVATERANGE = "PrivateRange";
	protected static final String PRVTASDU = "PrvtAsdu";
	protected static final String PRVTELEMENT = "PrvtElement";
	protected static final String PRVTCOT = "PrvtCot";
	protected static final String PRVTIN = "PrvtIn";
	protected static final String PRVTFT = "PrvtFt";

	protected static final String MAPPING = "Mapping";

	protected final static String PROCEDURE = "AppService";
	protected final static String RESETCU = "RESETCU";
	protected final static String RESETFCB = "RESETFCB";
	protected final static String SYNCHROPROC = "SYNCHRO";
	protected final static String GENERALINTERROGATION = "GI";
	protected final static String POLLINGPROC = "POLLING";
	protected final static String GENERALCOMMAND = "GENERALCOMMAND";
	protected final static String DCO = "dco";

	protected final static String PROCEDURES = "Procedures";
	protected static final String WRITERS = "Writers";
	// protected static final String READER = "Reader";
	// protected static final String READERKEY = "ReaderKey";
	// protected static final String ANSWERKEY = "AnswerKey";
	protected static final String ASDU = "Asdu";
	protected static final String CLASS1 = "class1";
	protected static final String CLASS2 = "class2";
	protected static final String WRITE = "write";
	protected static final String GI = "gi";
	protected static final String CI = "ci";
	protected static final String KEY = "Key";
	protected static final String ID = "id";
	protected static final String VSQ = "vsq";
	protected static final String COT = "cot";
	protected static final String CAA = "caa";

	protected final static String MAPVALUE = "MapValue";
	protected final static String PIO = "pio";
	protected final static String PIE = "pie";
	protected final static String PVAL = "pval";
	protected final static String DBREF = "db";
	protected final static String LABEL = "label";

	protected static final String LINEARTRANSLATOR = "LinearTranslator";
	protected static final String BINARYTRANSLATOR = "BinaryTranslator";
	protected static final String TOBINARYTRANSLATOR = "Formatter101ToBinary";
	protected static final String FROMBINARYTRANSLATOR = "FormatterBinaryTo101";
	protected static final String RATE = "rate";
	protected static final String OFFSET = "offset";
	protected static final String TYPE = "type";
	protected static final String THRESHOLD = "threshold";
	protected static final String LOGIC = "logic";
	protected static final String UPPERBOUND = "UPPERBOUND";
	protected static final String LOWERBOUND = "LOWERBOUND";
	protected static final String EQUALITY = "EQUALITY";
	protected static final String POSITIVE = "POSITIVE";
	protected static final String NEGATIVE = "NEGATIVE";

	/**
	 * Instanciate a MapSingleValue from XML configuration.
	 * 
	 * @param node
	 *            a KXMLElement MapValue.
	 * 
	 * @return an instance of a MapSingleValue.
	 */
	protected MapSingleValue inputMapValue(KXMLElement node) {
		AsduLocator locator = new AsduLocator(
				(short) node.getProperty(IOA, -1), (short) node.getProperty(
						PIE, -1), (short) node.getProperty(PVAL, -1));
		DbLocator db = new DbLocator(node.getProperty(DBREF, -1), node
				.getProperty(LABEL));
		Translator translator = null;

		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();)
			translator = inputTranslator((KXMLElement) e.nextElement());

		return new MapSingleValue(db, locator, translator);

	}

	/**
	 * Instanciate a KXMLElement from an instance of MapSingleValue.
	 * 
	 * @param node
	 *            an instance of a MapSingleValue.
	 * 
	 * @return a KXMLElement MapValue.
	 */
	// protected KXMLElement outputMapValue(MapSingleValue value)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(MAPVALUE);
	// node.addProperty(PIO,value.locator.pio);
	// node.addProperty(PIE,value.locator.pie);
	// node.addProperty(PVAL,value.locator.pval);
	// node.addProperty(DBREF,value.db.index);
	// node.addProperty(LABEL,value.db.label);
	// if (value.translator!=null)
	// node.addChild(outputTranslator(value.translator));
	// return node;
	// }

	/**
	 * Instanciate a Translator implementer from XML configuration.
	 * 
	 * @param node
	 *            a KXMLElement Translator.
	 * 
	 * @return a newly allocated Translator object.
	 */
	protected Translator inputTranslator(KXMLElement node) {
		String stype;
		byte type;
		String slogic;
		boolean logic;

		String name = node.getTagName();
		if (name.equals(LINEARTRANSLATOR)) {
			return new LinearTranslator(node.getProperty(RATE, -1), node
					.getProperty(OFFSET, -1));
		} else if (name.equals(BINARYTRANSLATOR)) {
			stype = node.getProperty(TYPE);
			if (stype.equals(UPPERBOUND))
				type = BinaryTranslator.UPPERBOUND;
			else if (stype.equals(LOWERBOUND))
				type = BinaryTranslator.LOWERBOUND;
			else
				type = BinaryTranslator.EQUALITY;
			slogic = node.getProperty(LOGIC);
			if (slogic.equals(POSITIVE))
				logic = true;
			else
				logic = false;
			float threshold = 0;
			if (CodeProperties.AJILE)
				threshold = node.getFloatProperty(THRESHOLD, (float) 0);
			else {
				String a = node.getProperty(THRESHOLD);
				threshold = java.lang.Float.parseFloat(a);
				System.out.println(new Float(threshold));
			}
			return new BinaryTranslator(type, threshold, logic);
		} else if (name.equals(FROMBINARYTRANSLATOR)) {
			slogic = node.getProperty(LOGIC);
			if (slogic.equals(POSITIVE))
				logic = true;
			else
				logic = false;
			return new FromBinaryFormatter101(logic);
		} else if (name.equals(TOBINARYTRANSLATOR)) {
			slogic = node.getProperty(LOGIC);
			if (slogic.equals(POSITIVE))
				logic = true;
			else
				logic = false;
			return new ToBinaryFormatter101(logic);
		}

		return null;
	}

	/**
	 * Instanciate a KXMLElement from an instance of Translator.
	 * 
	 * @param node
	 *            an instance of Translator.
	 * 
	 * @return a KXMLElement XXXXXTranslator.
	 */
	// protected KXMLElement outputTranslator(Translator translator)
	// {
	// KXMLElement node = new KXMLElement();
	//		
	// if (translator instanceof SMapranslator)
	// {
	// SMapranslator trans = (SMapranslator)translator;
	//			
	// node.setTagName(BINARYTRANSLATOR);
	//			
	// switch (trans.type)
	// {
	// case SMapranslator.UPPERBOUND:
	// node.addProperty(TYPE,UPPERBOUND);
	// break;
	// case SMapranslator.LOWERBOUND:
	// node.addProperty(TYPE,LOWERBOUND);
	// break;
	// case SMapranslator.EQUALITY:
	// node.addProperty(TYPE,EQUALITY);
	// break;
	// default:
	// break;
	// }
	// node.addProperty(THRESHOLD,Float.toString(trans.threshold));
	// node.addProperty(LOGIC,(trans.logic)?POSITIVE:NEGATIVE);
	// }
	// else if (translator instanceof Formatter101)
	// {
	// Formatter101 trans = (Formatter101)translator;
	//			
	// node.setTagName(FORMATTERTRANSLATOR);
	//			
	// switch (trans.type)
	// {
	// case Formatter101.UPPERBOUND:
	// node.addProperty(TYPE,UPPERBOUND);
	// break;
	// case Formatter101.LOWERBOUND:
	// node.addProperty(TYPE,LOWERBOUND);
	// break;
	// case Formatter101.EQUALITY:
	// node.addProperty(TYPE,EQUALITY);
	// break;
	// default:
	// break;
	// }
	// node.addProperty(THRESHOLD,Float.toString(trans.threshold));
	// node.addProperty(LOGIC,(trans.logic)?POSITIVE:NEGATIVE);
	// }
	// else if (translator instanceof LinearTranslator)
	// {
	// LinearTranslator trans = (LinearTranslator)translator;
	// node.setTagName(LINEARTRANSLATOR);
	// node.getProperty(RATE,trans.rate);
	// node.getProperty(OFFSET,trans.offset);
	// }
	// return node;
	// }

	/**
	 * Instanciate an asdu and its key from XML configuration.
	 * 
	 * @param node
	 *            a KXMLElement Writer.
	 * 
	 * @return an object array with Writer and related WriterKey.
	 */
	protected SMapAsdu101 inputAsdu(KXMLElement node) {
		AsduKey101 key = null;
		MapSingleValue[] value = new MapSingleValue[node.countChildren() - 1];
		int i = 0;

		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();) {
			KXMLElement child = (KXMLElement) e.nextElement();
			String name = child.getTagName();
			if (name.equals(KEY)) {
				key = new AsduKey101((short) child.getProperty(ID, -1),
						(short) child.getProperty(VSQ, -1), AsduKey101.NOTUSED,
						AsduKey101.NOTUSED, child.getProperty(CAA, -1));
			} else
				value[i++] = inputMapValue(child);
		}
		return new SMapAsdu101(key, value, node.getProperty(CLASS1, TRUE,
				FALSE, false), node.getProperty(CLASS2, TRUE, FALSE, false),
				node.getProperty(GI, TRUE, FALSE, false), node.getProperty(
						WRITE, TRUE, FALSE, false), node.getProperty(CI, TRUE,
						FALSE, false));
	}

	/**
	 * Instanciate a KXMLElement from an instance of Writer and another of
	 * related key WriterKey.
	 * 
	 * @param node
	 *            an instance of Writer and one of WriterKey.
	 * 
	 * @return a KXMLElement Writer.
	 */
	// protected KXMLElement outputWriter(Writer writer, WriterKey101 key)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(WRITER);
	//		
	// KXMLElement knode = new KXMLElement();
	// knode.setTagName(WRITERKEY);
	// knode.addProperty(ID,key.id);
	// knode.addProperty(VSQ,key.vsq);
	// knode.addProperty(CAA,key.caa);
	// knode.addProperty(FT,key.ft);
	// knode.addProperty(IN,key.in);
	//		
	// node.addChild(knode);
	//		
	// for(int i=0;i<writer.value.length;)
	// node.addChild(outputMapValue(writer.value[i++]));
	//		
	// return node;
	// }

	/**
	 * Instanciate a Reader from XML configuration.
	 * 
	 * @param node
	 *            a KXMLElement Reader.
	 * 
	 * @return an instance of Reader.
	 */
	// protected Reader inputReader(KXMLElement node)
	// {
	// ReaderKey101 key = null;
	// ReaderKey101 exp = null;
	// MapSingleValue[] value = new MapSingleValue[node.countChildren()-2];
	// int i = 0;
	// for(Enumeration e = node.enumerateChildren();e.hasMoreElements();)
	// {
	// KXMLElement child = (KXMLElement)e.nextElement();
	// String name = child.getTagName();
	// if(name.equals(READERKEY))
	// key = new ReaderKey101(
	// (short)child.getProperty(ID,-1),
	// (short)child.getProperty(VSQ,-1),
	// (short)child.getProperty(COT,-1),
	// child.getProperty(CAA,-1),
	// (short)child.getProperty(FT,-1),
	// (short)child.getProperty(IN,-1)
	// );
	// else if (name.equals(ANSWERKEY))
	// exp = new ReaderKey101(
	// (short)child.getProperty(ID,-1),
	// (short)child.getProperty(VSQ,-1),
	// (short)child.getProperty(COT,-1),
	// child.getProperty(CAA,-1),
	// (short)child.getProperty(FT,-1),
	// (short)child.getProperty(IN,-1)
	// );
	// else
	// value[i++] = inputMapValue(child);
	// }
	// return new Reader(value,key,new MasterServiceHelper(exp));
	// }

	/**
	 * Instanciate a KXMLElement from an instance of Reader.
	 * 
	 * @param node
	 *            an instance of Reader..
	 * 
	 * @return a KXMLElement Reader.
	 */
	// protected KXMLElement outputReader(Reader reader)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(READER);
	// ReaderKey101 key = (ReaderKey101)reader.key;
	// KXMLElement knode = new KXMLElement();
	// knode.setTagName(READERKEY);
	// knode.addProperty(ID,key.id);
	// knode.addProperty(VSQ,key.vsq);
	// knode.addProperty(CAA,key.cot);
	// knode.addProperty(CAA,key.caa);
	// knode.addProperty(FT,key.ft);
	// knode.addProperty(IN,key.in);
	// node.addChild(knode);
	// ReaderKey101 answer =
	// (ReaderKey101)((MasterServiceHelper)reader.svc).expectedKey;
	// KXMLElement snode = new KXMLElement();
	// snode.setTagName(ANSWERKEY);
	// snode.addProperty(ID,answer.id);
	// snode.addProperty(VSQ,answer.vsq);
	// snode.addProperty(CAA,answer.cot);
	// snode.addProperty(CAA,answer.caa);
	// snode.addProperty(FT,answer.ft);
	// snode.addProperty(IN,answer.in);
	// node.addChild(snode);
	// for(int i=0;i<reader.value.length;)
	// node.addChild(outputMapValue(reader.value[i++]));
	// return node;
	// }

	/**
	 * <p>
	 * Instanciate a master application procedure from XML configuration.
	 * <p>
	 * Notice manager is passed as argument in order to fill 'address' &
	 * 'context' fields of such procedures.
	 * 
	 * @param node
	 *            A KXMLElement AppService.
	 * @param manager
	 *            The application layer manager.
	 * 
	 * @return an instance of MAppService.
	 */
	// protected MAppService inputProcedure(KXMLElement node, MAppManager
	// manager)
	// {
	// MAppService proc;
	// String name = node.getProperty(TYPE);
	// DbLocator db = new DbLocator(
	// node.getProperty(DBREF,-1),
	// node.getProperty(LABEL)
	// );
	// if (name.equals(RESETCU))
	// return new MResetCuAppSvc101(manager,db);
	// else if (name.equals(REQSTATUS))
	// return new MReqStatusAppSvc101(manager,db);
	// else if (name.equals(SYNCHROPROC))
	// return new MSynchroAppSvc101(manager,db);
	// else if (name.equals(GENERALINTERROGATION))
	// return new MGeneralInterrogationAppSvc101(manager,db);
	// else if (name.equals(POLLINGPROC))
	// return new MPoll2Proc101(manager,db);
	// else if (name.equals(GENERALCOMMAND))
	// {
	// short ft = (short)node.getProperty(FT,-1);
	// short in = (short)node.getProperty(IN,-1);
	// byte dco = (byte)node.getProperty(DCO,-1);
	// return new MGeneralCommandProc101(manager,db,ft,in,dco);
	// }
	//		
	// return null;
	// }

	/**
	 * Instanciate a KXMLElement from an instance of AppService.
	 * 
	 * @param node
	 *            an instance of AppService.
	 * 
	 * @return a KXMLElement AppService.
	 */
	// protected KXMLElement outputProcedure(AppService procedure)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PROCEDURE);
	// node.addProperty(DBREF,procedure.notifier.index);
	// node.addProperty(LABEL,procedure.notifier.label);
	// if (procedure instanceof MResetCU101)
	// node.addProperty(TYPE,RESETCU);
	// else if (procedure instanceof MResetFCB101)
	// node.addProperty(TYPE,RESETFCB);
	// else if (procedure instanceof MSynchro101)
	// node.addProperty(TYPE,SYNCHROPROC);
	// else if (procedure instanceof MGeneralInterrogation101)
	// node.addProperty(TYPE,GENERALINTERROGATION);
	// else if (procedure instanceof MPolling101)
	// node.addProperty(TYPE,POLLINGPROC);
	// else if (procedure instanceof MGeneralCommand101)
	// {
	// MGeneralCommand101 cd = (MGeneralCommand101)procedure;
	// node.addProperty(TYPE,GENERALCOMMAND);
	// node.addProperty(FT,cd.ft);
	// node.addProperty(IN,cd.in);
	// node.addProperty(DCO,cd.dco);
	// }
	// return node;
	// }

	/**
	 * Instanciate a Mapping from XML configuration.
	 * 
	 * @param node
	 *            A KXMLElement Mapping.
	 * 
	 * @return an instance of Mapping.
	 */
	protected SMapping101 inputMapping(KXMLElement node, SAppManager101 manager) {
		SMapping101 map = new SMapping101(manager);

		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();) {
			KXMLElement child = (KXMLElement) e.nextElement();
			SMapAsdu101 asdu = inputAsdu(child);
			// general interrogation asdus.
			if (asdu.class1)
				map.addEvent(new SDataClass1AppSvc(manager, asdu));
			if (asdu.class2)
				map.addMeasure(new SDataClass2AppSvc(manager, asdu));
			if (asdu.gi)
				// check no time tag asdus
				map.addGi(new SDataClass1AppSvc(manager, asdu));
			if (asdu.ci)
				map.addCounter(new SDataClass1AppSvc(manager, asdu));
		}
		// for(Enumeration e = procs.enumerateChildren();e.hasMoreElements();)
		// {
		// map.addProcedure(inputProcedure((KXMLElement)e.nextElement(),manager));
		// }

		return map;
	}

	/**
	 * Instanciate a KXMLElement from an instance of Mapping.
	 * 
	 * @param node
	 *            An instance of Mapping.
	 * 
	 * @return a KXMLElement Mapping.
	 */
	// protected KXMLElement outputMapping(Mapping101 map)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(MAPPING);
	//		
	// Enumeration wkeys = map.enumerateWriterKeys();
	// Enumeration writers = map.enumerateWriters();
	// Enumeration readers = map.enumerateReaders();
	// Enumeration procedures = map.enumerateProcedures();
	//		
	// for(;wkeys.hasMoreElements();)
	// node.addChild(outputWriter((Writer)writers.nextElement(),(WriterKey101)wkeys.nextElement()));
	// for(;readers.hasMoreElements();)
	// node.addChild(outputReader((Reader)readers.nextElement()));
	// for(;procedures.hasMoreElements();)
	// node.addChild(outputProcedure((AppService)procedures.nextElement()));
	//		
	// return node;
	// }

	/**
	 * Instanciate a private asdu from xml configuration.
	 * 
	 * @param node
	 *            a KXMLElement PrvtAsdu.
	 * 
	 * @return an instance of PrvtAsdu.
	 */
	// protected PrivateAsdu101 inputPrvtAsdu(KXMLElement node)
	// {
	// int i = 0;
	// String[] ie = new String[node.countChildren()];
	// for(Enumeration e = node.enumerateChildren();e.hasMoreElements();)
	// {
	// KXMLElement child = (KXMLElement)e.nextElement();
	// // WARN: KXMLElement REHASH!!! WE SHOULD ADD A POSITION FLAG
	// ie[i++] = child.getProperty(TYPE);
	// }
	// return new PrivateAsdu101(
	// node.getProperty(LABEL),
	// (short)node.getProperty(ID,-1),
	// (short)node.getProperty(VSQ,-1),
	// (short)node.getProperty(COT,-1),
	// node.getProperty(CAA,-1),
	// (short)node.getProperty(FT,-1),
	// (short)node.getProperty(IN,-1),
	// ie
	// );
	// }

	/**
	 * Instanciate a KXMLElement from an instance of private asdu.
	 * 
	 * @param asdu
	 *            an instance of PrivateAsdu.
	 * 
	 * @return a KXMLElement PrvtAsdu.
	 */
	// protected KXMLElement outputPrvtAsdu(PrivateAsdu asdu)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PRVTASDU);
	// node.addProperty(ID,asdu.id);
	// node.addProperty(LABEL,asdu.label);
	// for(int i=0;i<asdu.ie.length;)
	// {
	// KXMLElement child = new KXMLElement();
	// child.setTagName(PRVTELEMENT);
	// child.addProperty(TYPE,asdu.ie[i++]);
	// node.addChild(child);
	// }
	// return node;
	// }

	/**
	 * Instanciate a Private cause of trnasmission from xml configuration.
	 * 
	 * @param node
	 *            a KXMLElement PrvtCot
	 * 
	 * @return an instance of PrivateCot
	 */
	// protected PrivateCot inputPrvtCot(KXMLElement node)
	// {
	// return new PrivateCot(
	// (short)node.getProperty(ID,-1),
	// node.getProperty(LABEL)
	// );
	// }

	/**
	 * Instanciate a KXMLElement from an instance of Private cause of
	 * transmission..
	 * 
	 * @param in
	 *            an instance of PrvtCot.
	 * 
	 * @return a KXMLElement PrivateCot.
	 */
	// protected KXMLElement outputPrvtCot(PrivateCot cot)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PRVTCOT);
	// node.addProperty(ID,cot.id);
	// node.addProperty(LABEL,cot.label);
	// return node;
	// }

	/**
	 * Instanciate a Private information number from xml configuration.
	 * 
	 * @param node
	 *            a KXMLElement PrvtIn
	 * 
	 * @return an instance of PrivateIn
	 */
	// protected PrivateIn101 inputPrvtIn(KXMLElement node)
	// {
	// return new PrivateIn101(
	// (short)node.getProperty(ID,-1),
	// node.getProperty(LABEL)
	// );
	// }

	/**
	 * Instanciate a KXMLElement from an instance of Private information number.
	 * 
	 * @param in
	 *            an instance of PrivateIn.
	 * 
	 * @return a KXMLElement PrvtIn
	 */
	// protected KXMLElement outputPrvtIn(PrivateIn101 in)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PRVTIN);
	// node.addProperty(ID,in.id);
	// node.addProperty(LABEL,in.label);
	// return node;
	// }

	/**
	 * Instanciate a Private function type from xml configuration.
	 * 
	 * @param node
	 *            a KXMLElement PrvtFt
	 * 
	 * @return an instance of PrivateFt
	 */
	// protected PrivateFt101 inputPrvtFt(KXMLElement node)
	// {
	// return new PrivateFt101(
	// (short)node.getProperty(ID,-1),
	// node.getProperty(LABEL)
	// );
	// }

	/**
	 * Instanciate a KXMLElement from an instance of Private function type.
	 * 
	 * @param ft
	 *            an instance of PrivateFt101.
	 * 
	 * @return a KXMLElement PrvtFt
	 */
	// protected KXMLElement outputPrvtFt(PrivateFt101 ft)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PRVTFT);
	// node.addProperty(ID,ft.id);
	// node.addProperty(LABEL,ft.label);
	// return node;
	// }

	/**
	 * Instanciate a private range from xml configuration.
	 * 
	 * @param node
	 *            PrivateRange KXMLElement.
	 * 
	 * @return an instance of PrivateRange.
	 */
	// protected PrivateRange101 inputPrivateRange(KXMLElement node)
	// {
	// PrivateRange101 prange = new PrivateRange101();
	// for(Enumeration e = node.enumerateChildren();e.hasMoreElements();)
	// {
	// KXMLElement child = (KXMLElement)e.nextElement();
	// String name = child.getTagName();
	// if (name.equals(PRVTASDU))
	// prange.addAsdu(inputPrvtAsdu(child));
	// else if (name.equals(PRVTCOT))
	// prange.addCot(inputPrvtCot(child));
	// else if (name.equals(PRVTIN))
	// prange.addIn(inputPrvtIn(child));
	// else if (name.equals(PRVTFT))
	// prange.addFt(inputPrvtFt(child));
	// else
	// throw new
	// IllegalArgumentException("PrivateRange has unknown sub element");
	// }
	// return prange;
	// }

	/**
	 * Instanciate a KXMLElement from an instance of private range.
	 * 
	 * @param node
	 *            an instance of PrivateRange.
	 * 
	 * @return PrivateRange KXMLElement.
	 */
	// protected KXMLElement outputPrivateRange(PrivateRange101 prange)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PRIVATERANGE);
	// for(Enumeration e = prange.enumerateAsdus();e.hasMoreElements();)
	// node.addChild(outputPrvtAsdu((PrivateAsdu)e.nextElement()));
	// for(Enumeration e = prange.enumerateCots();e.hasMoreElements();)
	// node.addChild(outputPrvtCot((PrivateCot)e.nextElement()));
	// for(Enumeration e = prange.enumerateFts();e.hasMoreElements();)
	// node.addChild(outputPrvtFt((PrivateFt101)e.nextElement()));
	// for(Enumeration e = prange.enumerateIns();e.hasMoreElements();)
	// node.addChild(outputPrvtIn((PrivateIn101)e.nextElement()));
	// return node;
	// }

	/**
	 * Instanciate a Protocol for general application settings from a
	 * KXMLElement.
	 * 
	 * @param node
	 *            a KXMLElement Protocol
	 * 
	 * @return an instance of Protocol.
	 */
	protected Protocol inputProtocol(KXMLElement node) {
		Protocol protocol = new Protocol();
		protocol.polling = node.getProperty(POLLING, 0);
		protocol.clock = node.getProperty(CLOCK, TRUE, FALSE, false);
		protocol.extendedTime = node.getProperty(EXTENDEDTIME, TRUE, FALSE,
				false);
		protocol.counterReset = node.getProperty(COUNTERTIME, 0);
		protocol.caa = node.getProperty(CAA, 0);
		protocol.address = node.getProperty(ADDRESS, 0);
		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();) {
			KXMLElement child = (KXMLElement) e.nextElement();
			String name = child.getTagName();
			if (name.equals(FIELDSIZE)) {
				protocol.caaSz = (byte) child.getProperty(CAASZ, 1);
				protocol.ioaSz = (byte) child.getProperty(IOASZ, 1);
				protocol.cotSz = (byte) child.getProperty(COTSZ, 1);
				protocol.lnkSz = (byte) child.getProperty(LNKSZ, 1);
			}
		}
		return protocol;
	}

	/**
	 * Instanciate KXMLElement from an instance of Protocol.
	 * 
	 * @param protocol
	 *            an instance of Protocol.
	 * 
	 * @return a KXMLElement Protocol.
	 */
	protected KXMLElement outputProtocol(Protocol protocol) {
		KXMLElement node = new KXMLElement();
		KXMLElement fsnode = new KXMLElement();
		node.setTagName(PROTOCOL);
		fsnode.setTagName(FIELDSIZE);
		node.addProperty(POLLING, protocol.polling);
		node.addProperty(CLOCK, (protocol.clock) ? TRUE : FALSE);
		// node.addProperty(EXTENDEDTIME,(protocol.extendedTime)?TRUE:FALSE);
		// node.addProperty(COUNTERTIME,protocol.counterReset);
		node.addProperty(CAA, protocol.caa);
		node.addProperty(ADDRESS, protocol.address);
		fsnode.addProperty(CAASZ, protocol.caaSz);
		fsnode.addProperty(IOASZ, protocol.ioaSz);
		fsnode.addProperty(COTSZ, protocol.cotSz);
		fsnode.addProperty(LNKSZ, protocol.lnkSz);
		node.addChild(fsnode);
		return node;
	}

	/**
	 * Instanciate a Communication for general comm settings from a KXMLElement.
	 * 
	 * @param node
	 *            a KXMLElement Communication.
	 * 
	 * @return an instance of Communication.
	 */
	protected Communication inputCommunication(KXMLElement node) {
		return new Communication(node.getProperty(PORT).toLowerCase(), node
				.getProperty(BAUDRATE, 9600));

	}

	/**
	 * Instanciate a KXMLElement from an instance of Communication. from a
	 * KXMLElement.
	 * 
	 * @param comm
	 *            an instance of Communication.
	 * 
	 * @return a KXMLElement Communication.
	 */
	protected KXMLElement outputCommunication(Communication comm) {
		KXMLElement node = new KXMLElement();
		node.setTagName(COMMUNICATION);
		node.addProperty(PORT, comm.port.toUpperCase());
		node.addProperty(BAUDRATE, comm.baud);
		return node;
	}

	/**
	 * Update a whole application configuration from a KXMLElement node.
	 * 
	 * @param node
	 *            a KXMLElement ROOT.
	 */
	public void inputFromXML(KXMLElement node) {
		if (!node.getTagName().equals(ROOT))
			return;

		author = node.getProperty(AUTHOR);
		version = node.getProperty(VERSION);
		date = node.getProperty(DATE);
		fileId = node.getProperty("fileId", "unknown");
		SLinkLayer lnk = null;
		PhysicalLayer phy = null;
		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();) {
			KXMLElement child = (KXMLElement) e.nextElement();
			String name = child.getTagName();
			if (name.equals(COMMUNICATION)) {
				comm = inputCommunication(child);
			} else if (name.equals(PROTOCOL)) {
				protocol = inputProtocol(child);

				app = new SApplicationLayer();
				lnk = new SLinkLayer();
				// phy = new PhySerialLayer(comm.port, comm.baud,
				// Communication.data, Communication.stop,
				// Communication.parity);
				SContext101 context = new SContext101(protocol.address,
						protocol.caa, protocol.clock, protocol.extendedTime);
				app.reader = new SAsduReader101(protocol.cotSz, protocol.caaSz,
						protocol.ioaSz, protocol.lnkSz, context);
				app.writer = new SAsduWriter101(protocol.cotSz, protocol.caaSz,
						protocol.ioaSz, protocol.lnkSz);
				app.lower = lnk;
				lnk.upper = app;
				lnk.lower = phy;
				phy.upper = lnk;
				app.manager = new SAppManager101(app, context, protocol.polling);
			}
			// else if (name.equals(PRIVATERANGE))
			// {
			// backupPrvt = child;
			// app.manager.prange = inputPrivateRange(child);
			// ((SAsduWriter103)app.writer).prange =
			// (PrivateRange103)app.manager.prange;
			// }
			else if (name.equals(MAPPING)) {
				backupMap = child;
				app.manager.map = inputMapping(child,
						(SAppManager101) app.manager);
				app.manager.map.manager = app.manager;
				int i = 0;
				for (Enumeration e2 = ((SMapping101) app.manager.map)
						.enumerateMeasures(); e.hasMoreElements();) {
					Object tmp = e2.nextElement();
					i++;
				}
				lnk.manager.setNbMeasure(i);
			}
		}
	}

	/**
	 * Instanciate a KXMLElement ROOT from an configuration.
	 */
	public KXMLElement outputToXML() {
		KXMLElement node = new KXMLElement();
		node.setTagName(ROOT);

		node.addProperty(AUTHOR, author);
		node.addProperty(VERSION, version);
		node.addProperty(DATE, date);
		node.addProperty("fileId", fileId);

		node.addChild(outputCommunication(comm));
		node.addChild(outputProtocol(protocol));
		node.addChild(backupMap);
		return node;
	}

	/**
	 * Kills all threads in current session.
	 */
	public void kill() {

		if (app != null) {
			app.kill();
			try {
				app.join();
			} catch (Exception e) {
			}
			CodeProperties.log("S101 really killed.");
		}
	}

	/**
	 * starts application.
	 */
	public void start() {

		if (!comm.port.equals("unknown")) {
			app.start();
		}
		// if (!ACTIVE.comm.port.equals("UNKNOWN"))
		// {
		// // instanciates a network configuration
		// SlaveNetwork101 network = new
		// SlaveNetwork101(ACTIVE.getRemotes(),ACTIVE.protocol.polling,ACTIVE.protocol.synchro,ACTIVE.protocol.gi,ACTIVE.protocol.nbToPoll);
		// // instanciates an application layer for this app.
		// application = new MasterApplicationLayer(
		// ACTIVE.map,
		// new MasterAsduReader101(ACTIVE.prange),
		// new MasterAsduWriter101(),
		// network,
		// network,
		// ACTIVE.comm.port,
		// ACTIVE.comm.baudrate,
		// ACTIVE.comm.data,
		// ACTIVE.comm.stop,
		// ACTIVE.comm.parity,
		// ACTIVE.protocol.timeout,
		// ACTIVE.protocol.retries
		// );
		//			
		// try
		// {
		// work();
		// }
		// catch(IEC608705Exception e)
		// {
		// CodeProperties.log("CEI Exception: "+e);
		// }
		// }
		// else
		// CodeProperties.log("Stopping.");
	}

	public void work() throws IEC608705Exception {
		// CodeProperties.log("Application starting.");
		// application.start();
	}

	/*
	 * /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*\ < < >
	 * Constructor > < <
	 * \*\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
	 */

	/**
	 * Instanciate a new configurator for master 101 application.
	 */
	public SConfiguration101() {
		comm = new Communication();
		protocol = new Protocol();
		app = new SApplicationLayer();
		backupMap = new KXMLElement();
	}

} // end SConfiguration101.java