/*
 * Class    br.org.scadabr.protocol.iec101.master103.SAsduReader103
 * File     SAsduReader103.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.slave103;

import br.org.scadabr.protocol.iec101.common.Asdu;
import br.org.scadabr.protocol.iec101.common.AsduReader;
import br.org.scadabr.protocol.iec101.common.CodeProperties;
import br.org.scadabr.protocol.iec101.common.IEC608705Exception;
import br.org.scadabr.protocol.iec101.common.InformationElement;
import br.org.scadabr.protocol.iec101.common.InformationObject;
import br.org.scadabr.protocol.iec101.common.PrivateAsdu;
import br.org.scadabr.protocol.iec101.common.PrivateCot;
import br.org.scadabr.protocol.iec101.common.PrivateRange;
import br.org.scadabr.protocol.iec101.common.UserData;
import br.org.scadabr.protocol.iec101.common103.Asdu103;
import br.org.scadabr.protocol.iec101.common103.Interoperability103;
import br.org.scadabr.protocol.iec101.common103.PrivateRange103;
import br.org.scadabr.protocol.iec101.common103.information.*;

import java.util.Enumeration;

/**
 * This class describes mechanisms to read 103 asdus as a master station.
 *
 * @author lionnel cauvy
 * @version 1.0
 * @date 01.II.03
 */
public class SAsduReader103 implements AsduReader
{	
	/**
	 * Private range reference.
	 */
	private PrivateRange103 prange;
	/**
	 * Current private asdu (if so) reference.
	 */
	private PrivateAsdu pasdu;
	/**
	 * Private reading flag
	 */
	private boolean privateRead;
	/**
	 * Asdu reference.
	 */
	private Asdu103 asdu;	
	/**
	 * Fills variable structure qualifier field in asdu.
	 *
	 * @param octet		vsq value.
	 *
	 * @return true or false according to this information standard compliance.
	 */
	protected boolean setVariableStructureQualifier(short octet)
	{
		boolean valid = true;
		
		asdu.vsq = octet;
		asdu.nio = (byte)(octet & MASK_NB);
		asdu.sq = ( ((short)(octet & MASK_SQ) >> 7)==1 )?true:false;
		//CodeProperties.applog("[AsduFactory] variable structure qualifier _ SQ="+asdu.sq+" N="+asdu.nio);
		
		return valid;
	}
	/**
	 * Fills cause of transmission field in 103 asdu.
	 *
	 * @param octet		cause of transmission value.
	 *
	 * @return true or false according to this information standard compliance.
	 */
	protected boolean setCauseTransmission(short octet)
	{
		boolean valid = true;
	
		switch (octet)	
		{					
		case Interoperability103.COT_SPONTANEOUS:
			//CodeProperties.applog("[AsduReader] cause of transmission _ spontaneous");
			break;
		case Interoperability103.COT_CYCLIC:
			//CodeProperties.applog("[AsduReader] cause of transmission _ cyclic");
			break;
		case Interoperability103.COT_RESET_FCB:
			//CodeProperties.applog("[AsduReader] cause of transmission _ resetFCB");
			break;
		case Interoperability103.COT_RESET_CU:
			//CodeProperties.applog("[AsduReader] cause of transmission _ resetCU");
			break;
		case Interoperability103.COT_START_RESTART:
			//CodeProperties.applog("[AsduReader] cause of transmission _ start_restart");
			break;
		case Interoperability103.COT_POWER:
			//CodeProperties.applog("[AsduReader] cause of transmission _ power_on");
			break;
		case Interoperability103.test_mode:
			//CodeProperties.applog("[AsduReader] cause of transmission _ test_mode");
			break;
		case Interoperability103.COT_SYNCHRO:
			//CodeProperties.applog("[AsduReader] cause of transmission _ time_synchro");
			break;
		case Interoperability103.COT_GI:
			//CodeProperties.applog("[AsduReader] cause of transmission _ general_interrogation");
			break;
		case Interoperability103.COT_END_GI:
			//CodeProperties.applog("[AsduReader] cause of transmission _ terminationGI");
			break;
		case Interoperability103.local_op:
			//CodeProperties.applog("[AsduReader] cause of transmission _ local_op");
			break;
		case Interoperability103.COT_REMOTE:
			//CodeProperties.applog("[AsduReader] cause of transmission _ remote_op");
			break;
		case Interoperability103.COT_ACK_COM:
			//CodeProperties.applog("[AsduReader] cause of transmission _ ack_com");
			break;
		case Interoperability103.COT_NACK_COM:
			//CodeProperties.applog("[AsduReader] cause of transmission _ nack_com");
			break;
		case Interoperability103.disturbance_data:
			//CodeProperties.applog("[AsduReader] cause of transmission _ disturbance_data");
			break;
		case Interoperability103.ack_write_com:
			//CodeProperties.applog("[AsduReader] cause of transmission _ ack_write_com");
			break;
		case Interoperability103.nack_write_com:
			//CodeProperties.applog("[AsduReader] cause of transmission _ nack_write_com");
			break;
		case Interoperability103.valid_read_com:
			//CodeProperties.applog("[AsduReader] cause of transmission _ valid_read_com");
			break;
		case Interoperability103.invalid_read_com:
			//CodeProperties.applog("[AsduReader] cause of transmission _ invalid_read_com");
			break;
		case Interoperability103.generic_write_conf:
			//CodeProperties.applog("[AsduReader] cause of transmission _ generic_write_conf");
			break;
		default:
			// is this a private cause of transmission?
			PrivateCot pcot = prange.getCot(octet);
			if (pcot!=null)
			{
				CodeProperties.applog("[AsduReader] cause of transmission _ "+pcot.label);
			}
			else
			{
				CodeProperties.applog("[AsduReader cause of transmission _ unknown COT");
				valid = false;
			}
			break;
		}		
		asdu.cot = octet;		
		return valid;
	}	
	/**
	 * Fills common address of asdu field in 103 asdu.
	 *
	 * @param octet		caa value.
	 *
	 * @return true or false according to this information standard compliance.
	 */
	protected boolean setCommonAddress(short octet)
	{
		asdu.caa = octet;		
		return true;
	}	
	/**
	 * Fills Function Type and Information Number fields in 103 asdu.
	 *
	 * @param octet1	function type value.
	 * @param octet2	information number value.
	 *
	 * @return true or false according to this information standard compliance.
	 */
	protected boolean setInformationObjectIdentifier(short octet1, short octet2)
	{
		asdu.ft = octet1;
		asdu.in = octet2;		
		return true;
	}	
	/**
	 * Fills type identification field in 103 asdu. Private asdus
	 * can be detected here.
	 *
	 * @param octet		type identification value.
	 *
	 * @return true or false according to this information standard compliance.
	 */
	protected boolean setTypeIdentification(short octet)
	{
		boolean valid = true;
		
		switch (octet)
		{
		case Interoperability103.ID_SYNCHRO:
			//CodeProperties.applog("[MasterAsduReader] type identification _ TimeSynchronization");
			break;
		case Interoperability103.ID_INIT_GI:
			//CodeProperties.applog("[MasterAsduReader] type identification _ Init General Interrogation");
			break;
		case Interoperability103.ID_GENERAL_CD:
			//CodeProperties.applog("[MasterAsduReader] type identification _ TerminationGeneralInterrogation");
			break;
		default:
			// is this a private asdu?
			pasdu = prange.getAsdu(octet);
			if (pasdu!=null)
			{
				privateRead = true;
				CodeProperties.applog("[SAsduReader103] Asdu type:"+octet+" from private range.");
			}
			// no sorry.....
			else
			{
				CodeProperties.applog("[SAsduReader103] type identification _ unknow type! "+octet);
				valid = false;
			}
			break;
		}		
		asdu.id = octet;		
		return valid;
	}
	/**
	 * Build main method inherited from asdu factory. Here, this
	 * completes a 103 asdu.
	 *
	 * @param object	short array to turn into an asdu.
	 * @param prange	The private range of related remote that sent this asdu.
	 *
	 * @return the built asdu.
	 */
	public Asdu build(Object object, PrivateRange prange) throws IEC608705Exception
	{
		reset();
		
		this.prange = (PrivateRange103)prange;
					
		if(object==null)
			return null;
		
		UserData user = (UserData)object;
		asdu = new Asdu103();
		
		asdu.address = user.address;

		if(user.data==null)
			return asdu;

		if(user.data.isEmpty())
			return asdu;	
					
		short num = 0;
		Enumeration data = user.data.elements();

		// DATA UNIT IDENTIFIER
		//		type identification
		if (!setTypeIdentification(((Short)data.nextElement()).shortValue()))
		{
			reset();
			return null;
		}
		//		variable structure qualifier
		if(!setVariableStructureQualifier(((Short)data.nextElement()).shortValue()))
		{
			reset();
			return null;
		}
		//		cause of transmission
		// 		(T103 COT size is 1 octet)
		//if(Interoperability.ApplicationLayer.causeTransmissionSize2)
		//	setCauseTransmission( (((Short)data.nextElement()).shortValue() & 0xFF) | ((((Short)data.nextElement()).shortValue() & 0xFF)<<8) );
		// else....
		if(!setCauseTransmission(((Short)data.nextElement()).shortValue()))
		{
			reset();
			return null;
		}
		// 		common address
		// 		(T103 1 octet)
		// 		(T103 COT size is 1 octet)
		//if(Interoperability.ApplicationLayer.commonAddressSize2)
		//	setCommonAddress( (((Short)data.nextElement()).shortValue() & 0xFF) | ((((Short)data.nextElement()).shortValue() & 0xFF)<<8) );
		// else...
		if(!setCommonAddress(((Short)data.nextElement()).shortValue()))
		{
			reset();
			return null;
		}			
		// INFORMATION OBJECT IDENTIFIER
		if(!setInformationObjectIdentifier(((Short)data.nextElement()).shortValue(),((Short)data.nextElement()).shortValue()))
			return null;
		
		// INFORMATION OBJECT
		// sq from variable structure qualifier defines structure of info object set
		if (!asdu.sq)
		{
			// 1 information object, n information elements
			// number from variable structure qualifier defines number n of information elements
			//setInformationObjectStructure((byte)1,number);
			num = 1;
		}
		else
		{
			// n information objects, 1 information element per object
			// number from variable structure qualifier defines number n of information objects
			//setInformationObjectStructure(number,(byte)1);
			num = asdu.nio;
		}
		if (!asdu.sq & asdu.nio==0)
		{
			num = 0x00;
		}
		try
		{
			// id from type identification defines which elements to use
			for(short i=0;i<num;i++)
			{
				InformationObject info = new InformationObject();
				
				switch (asdu.id)
				{					
				case Interoperability103.ID_SYNCHRO:
					info.setElement(new CP56Time2a(((Short)data.nextElement()).shortValue(),((Short)data.nextElement()).shortValue(),((Short)data.nextElement()).shortValue(),((Short)data.nextElement()).shortValue(),((Short)data.nextElement()).shortValue(),((Short)data.nextElement()).shortValue(),((Short)data.nextElement()).shortValue()));
					break;
					
				case Interoperability103.ID_INIT_GI:
					info.setElement(new SCN(((Short)data.nextElement()).shortValue()));
					break;
					
				case Interoperability103.ID_GENERAL_CD:
					info.setElement(new DCO(((Short)data.nextElement()).shortValue()));
					info.setElement(new RII(((Short)data.nextElement()).shortValue()));
					break;
					
				default:
					// is this private asdu?
					if (privateRead)
					{
						try
						{
							for(int z=0;z<pasdu.ie.length;z++)
							{						
								Class clazz = (Class)((PrivateRange103)prange).CLASSES.get(pasdu.ie[z]);
								InformationElement elt = (InformationElement)clazz.newInstance();								
								for(short y=0;y<elt.octetset.length;y++)
									elt.setOctet(y,((Short)data.nextElement()).shortValue());
								info.setElement(elt);	
							}
						}
						catch(InstantiationException e2)
						{
							CodeProperties.applog("[SAsduReader103] not instanciable "+e2);
						}
						catch(IllegalAccessException e3)
						{
							CodeProperties.applog("[SAsduReader103] bad access "+e3);
						}
					}
					// no so sorry...
					else
					{
						CodeProperties.applog("[SAsduReader103] Unknown asdu type identification");
					}
					break;
				}
				asdu.add(info);
				
			}
		}
		catch(ArrayIndexOutOfBoundsException e)
		{
			throw new IEC608705Exception("IEC Exception: bad asdu format");
		}
			
		return asdu;
	}	
	/**
	 * Reset factory
	 */
	protected void reset()
	{
		asdu = null;
		pasdu = null;
		privateRead = false;		
	}	
}	// end SAsduReader103.java