/*
 * Class    br.org.scadabr.protocol.slave103.SConfiguration103;
 * File     SConfiguration103.java
 * Author   lionnel cauvy
 */
package br.org.scadabr.protocol.iec101.slave103;

import java.util.Enumeration;

import br.org.scadabr.luciol.Configurable;
import br.org.scadabr.luciol.ihm.Labels;
import br.org.scadabr.protocol.iec101.common.AsduKey;
import br.org.scadabr.protocol.iec101.common.AsduLocator;
import br.org.scadabr.protocol.iec101.common.BinaryTranslator;
import br.org.scadabr.protocol.iec101.common.CodeProperties;
import br.org.scadabr.protocol.iec101.common.DbLocator;
import br.org.scadabr.protocol.iec101.common.LinearTranslator;
import br.org.scadabr.protocol.iec101.common.MapAsdu;
import br.org.scadabr.protocol.iec101.common.MapSingleValue;
import br.org.scadabr.protocol.iec101.common.PhysicalLayer;
import br.org.scadabr.protocol.iec101.common.PrivateCot;
import br.org.scadabr.protocol.iec101.common.Translator;
import br.org.scadabr.protocol.iec101.common103.AsduKey103;
import br.org.scadabr.protocol.iec101.common103.FromBinaryFormatter103;
import br.org.scadabr.protocol.iec101.common103.Interoperability103;
import br.org.scadabr.protocol.iec101.common103.PrivateAsdu103;
import br.org.scadabr.protocol.iec101.common103.PrivateFt103;
import br.org.scadabr.protocol.iec101.common103.PrivateIn103;
import br.org.scadabr.protocol.iec101.common103.PrivateRange103;
import br.org.scadabr.protocol.iec101.common103.ToBinaryFormatter103;
import br.org.scadabr.protocol.iec101.slave.SApplicationLayer;
import br.org.scadabr.protocol.iec101.slave.SDataClass1AppSvc;
import br.org.scadabr.protocol.iec101.slave.SDataClass2AppSvc;
import br.org.scadabr.protocol.iec101.slave.SLinkLayer;
import br.org.scadabr.protocol.iec101.slave.SMapAsdu;
import br.org.scadabr.protocol.lib.xml.KXMLElement;

/**
 * <p>
 * This class defines common methods to instanciate objects in use for a slave
 * 103 application from an xml file.
 * <p>
 * These mechanisms are OBLIGATORY for a Luciol application in order to be
 * compatible with HMI and central application manager.
 * 
 * @author lionnel cauvy
 * @version 1.0
 * @date 01.II.03
 */
public class SConfiguration103 implements Configurable {

	/*
	 * /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*\ < < > Inner
	 * classes. Communication & protocol settings. > < >
	 * \*\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
	 */

	/**
	 * This class defines communication settings.
	 * <p>
	 * 1O3 implementation fixes parity, data & stop bits. Configurable fields
	 * are comm port selection and baudrate.
	 */
	public class Communication {
		public String port = "unknown";
		public int baud = 9600;
		public final static int parity = 2;
		public final static int data = 8;
		public final static int stop = 1;

		public Communication() {
		}

		public Communication(String port, int baud) {
			this.port = port;
			this.baud = baud;
		}
	}

	/**
	 * This class defines protocol settings.
	 * <p>
	 * Id is a String of maximum 12 characters for local identification.
	 * <p>
	 * Address permits the slave application to be identified on a network.
	 * <p>
	 * Caa defines value of Common Address of Asdus to use.
	 * <p>
	 * Polling defines measure cyclic period in seconds.
	 * <p>
	 * Clock enables to change system clock on a synchro request from a master.
	 */
	public class Protocol {
		public String id = "LUCIOLXXXXXX";
		public int address = 1;
		public short ft = 160;
		public int caa = 1;
		public int polling = 5;
		public boolean clock = false;

		public Protocol() {
		}

		public Protocol(String id, int address, short ft, int caa, int polling,
				boolean clock) {
			this.id = id;
			this.address = address;
			this.ft = ft;
			this.caa = caa;
			this.polling = polling;
			this.clock = clock;
		}
	}

	/*
	 * /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*\ < < >
	 * Variables > < <
	 * \*\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
	 */

	/**
	 * Communication reference.
	 */
	public Communication comm;
	/**
	 * Protocol reference.
	 */
	public Protocol protocol;
	/**
	 * Application layer reference.
	 */
	public SApplicationLayer app;
	/**
	 * Backup kXML Remote. Avoid to access runtime objects since configuration
	 * is the application constructor.
	 */
	private KXMLElement backupPrvt;
	private KXMLElement backupMap;

	private String author = "l.cauvy";
	private String version = "1.0";
	private String date = "19.II.03";
	private String fileId = "unknown";

	public String getAuthor() {
		return author;
	}

	public String getVersion() {
		return version;
	}

	public String getFileId() {
		return fileId;
	}

	public String getDate() {
		return date;
	}

	/*
	 * /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*\ < < >
	 * Configurable Interface. HMI Luciol. > < >
	 * \*\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
	 */

	private static final String APPNAME = "Slave103";

	/**
	 * Communication port parameter possible values.
	 */
	protected static final String[] POSSIBLE_PORT = new String[] { "COM1",
			"COM2", "UNKNOWN" };

	/**
	 * Baudrate parameter possible values.
	 */
	protected static final String[] POSSIBLE_SPEED = new String[] { "9600",
			"19200" };

	/**
	 * Return a unique name for this application.
	 */
	public String applicationName() {
		return APPNAME;
	}

	/**
	 * Return values of a SELECT type parameter.
	 * 
	 * @param param
	 *            a parameter number.
	 * @return values of a parameter.
	 */
	public String[] getPossibleValues(int param)
			throws IllegalArgumentException {
		switch (param) {
		case 0:
			return POSSIBLE_PORT;
		case 1:
			return POSSIBLE_SPEED;
		default:
			throw new IllegalArgumentException();
		}
	}

	/**
	 * Parameters names to display in luciol hmi
	 */
	private static final String[] OPTION_NAMES = new String[] {
			Labels.S103_COM_PORT, Labels.S103_BAUDRATE, Labels.S103_ID,
			Labels.S103_ADDRESS, Labels.S103_FT, Labels.S103_CAA,
			Labels.S103_POLLING, Labels.S103_CLOCK };

	/**
	 * Return parameters names.
	 */
	public String[] getParameterNames() {
		return OPTION_NAMES;
	}

	/**
	 * Return parameters number.
	 */
	public int getParameterCount() {
		return OPTION_NAMES.length;
	}

	/**
	 * Parameters types.
	 */
	private static final byte[] OPTION_TYPES = new byte[] { SELECT, SELECT,
			CDATA, CDATA, CDATA, CDATA, CDATA, CHECK };

	/**
	 * Return types of parameters.
	 * 
	 * @return the current type of a parameter.
	 */
	public byte[] getParameterTypes() {
		return OPTION_TYPES;
	}

	/**
	 * Return current values of parameters.
	 * 
	 * @return the current value of a parameter.
	 */
	public Object[] getParameterValues() {
		int port = 0;
		int baud = 0;
		while (!comm.port.toUpperCase().equals(POSSIBLE_PORT[port]))
			port++;
		while (!String.valueOf(comm.baud).equals(POSSIBLE_SPEED[baud]))
			baud++;

		return new Object[] { new Integer(port), new Integer(baud),
				protocol.id, String.valueOf(protocol.address),
				String.valueOf(protocol.ft), String.valueOf(protocol.caa),
				String.valueOf(protocol.polling), new Boolean(protocol.clock) };
	}

	/**
	 * <p>
	 * Test if <tt>newValue</tt> can be assigned as a new value for the
	 * parameter specified by <tt>num</tt>.
	 * 
	 * @param newValue
	 *            a new value for this parameter.
	 */
	public boolean[] setNewValues(Object[] values)
			throws IllegalArgumentException {
		boolean[] res = new boolean[values.length];
		int i = 0;
		// com port _ COM1 COM2
		String name = POSSIBLE_PORT[((Integer) values[i]).intValue()];
		comm.port = name.toLowerCase();
		res[i++] = true;
		// baud rate _ 9600 19200
		comm.baud = Integer.parseInt(POSSIBLE_SPEED[((Integer) values[i])
				.intValue()]);
		if (comm.baud == 9600 || comm.baud == 19200)
			res[i++] = true;
		else
			res[i++] = false;
		// Identification
		protocol.id = (String) values[i];
		if (protocol.id.length() <= 12)
			res[i++] = true;
		else
			res[i++] = false;
		// address
		protocol.address = Integer.parseInt(((String) values[i]));
		if (protocol.address >= 1 && protocol.address <= 254)
			res[i++] = true;
		else
			res[i++] = false;
		// function type
		protocol.ft = Short.parseShort(((String) values[i]));
		if (protocol.ft >= 0 && protocol.ft <= 255)
			res[i++] = true;
		else
			res[i++] = false;
		// common address of asdus
		protocol.caa = Integer.parseInt(((String) values[i]));
		if (protocol.caa >= 1 && protocol.caa <= 254)
			res[i++] = true;
		else
			res[i++] = false;
		// polling cycle
		protocol.polling = Integer.parseInt(((String) values[i]));
		if (protocol.polling >= 0)
			res[i++] = true;
		else
			res[i++] = false;
		// change system clock permission
		protocol.clock = ((Boolean) values[i]).booleanValue();
		res[i++] = true;

		return res;
	}

	/*
	 * /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*\ < < > XML
	 * Interface. InputToXML, OutputFromXML > < <
	 * \*\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
	 */

	protected static final String ROOT = "Slave103";
	protected static final String AUTHOR = "author";
	protected static final String VERSION = "version";
	protected static final String DATE = "date";
	protected static final String FILEID = "fileId";

	protected static final String COMMUNICATION = "Communication";
	protected static final String PORT = "port";
	protected static final String BAUDRATE = "baudrate";

	protected static final String PROTOCOL = "Protocol";
	protected static final String IDENTIFICATION = "id";
	protected static final String ADDRESS = "address";
	protected static final String CAA = "caa";
	protected static final String POLLING = "polling";
	protected static final String CLOCK = "clock";
	protected static final String TRUE = "1";
	protected static final String FALSE = "0";

	protected static final String PRIVATERANGE = "PrivateRange";
	protected static final String PRVTASDU = "PrvtAsdu";
	protected static final String PRVTELEMENT = "PrvtElement";
	protected static final String PRVTCOT = "PrvtCot";
	protected static final String PRVTIN = "PrvtIn";
	protected static final String PRVTFT = "PrvtFt";

	protected static final String MAPPING = "Mapping";

	protected static final String MEASURANDS = "Measurands";
	protected static final String GENERALINTERROGATION = "GeneralInterrogation";
	protected static final String EVENTS = "Events";
	protected static final String WRITERS = "Writers";

	protected static final String ASDU = "Asdu";
	protected static final String CLASS1 = "class1";
	protected static final String CLASS2 = "class2";
	protected static final String GI = "gi";
	protected static final String WRITE = "write";
	protected static final String KEY = "Key";
	protected static final String ID = "id";
	protected static final String VSQ = "vsq";
	protected static final String COT = "cot";
	// protected static final String CAA = "caa";
	protected static final String FT = "ft";
	protected static final String IN = "in";

	protected final static String MAPVALUE = "MapValue";
	protected final static String PIO = "pio";
	protected final static String PIE = "pie";
	protected final static String PVAL = "pval";
	protected final static String DBREF = "db";
	protected final static String LABEL = "label";

	protected static final String LINEARTRANSLATOR = "LinearTranslator";
	protected static final String BINARYTRANSLATOR = "BinaryTranslator";
	protected static final String TOBINARYTRANSLATOR = "Formatter103ToBinary";
	protected static final String FROMBINARYTRANSLATOR = "FormatterBinaryTo103";
	protected static final String RATE = "rate";
	protected static final String OFFSET = "offset";
	protected static final String TYPE = "type";
	protected static final String THRESHOLD = "threshold";
	protected static final String LOGIC = "logic";
	protected static final String UPPERBOUND = "UPPERBOUND";
	protected static final String LOWERBOUND = "LOWERBOUND";
	protected static final String EQUALITY = "EQUALITY";
	protected static final String POSITIVE = "POSITIVE";
	protected static final String NEGATIVE = "NEGATIVE";

	/**
	 * Instanciate a MapSingleValue from XML configuration.
	 * 
	 * @param node
	 *            a KXMLElement MapValue.
	 * 
	 * @return an instance of a MapSingleValue.
	 */
	protected MapSingleValue inputMapValue(KXMLElement node) {
		AsduLocator locator = new AsduLocator(
				(short) node.getProperty(PIO, -1), (short) node.getProperty(
						PIE, -1), (short) node.getProperty(PVAL, -1));
		DbLocator db = new DbLocator(node.getProperty(DBREF, -1), node
				.getProperty(LABEL));
		Translator translator = null;

		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();)
			translator = inputTranslator((KXMLElement) e.nextElement());

		return new MapSingleValue(db, locator, translator);

	}

	/**
	 * Instanciate a KXMLElement from an instance of MapSingleValue.
	 * 
	 * @param node
	 *            an instance of a MapSingleValue.
	 * 
	 * @return a KXMLElement MapValue.
	 */
	// protected KXMLElement outputMapValue(MapSingleValue value)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(MAPVALUE);
	// node.addProperty(PIO,value.locator.pio);
	// node.addProperty(PIE,value.locator.pie);
	// node.addProperty(PVAL,value.locator.pval);
	// node.addProperty(DBREF,value.db.index);
	// node.addProperty(LABEL,value.db.label);
	// if (value.translator!=null)
	// node.addChild(outputTranslator(value.translator));
	// return node;
	// }

	/**
	 * Instanciate a Translator implementer from XML configuration.
	 * 
	 * @param node
	 *            a KXMLElement Translator.
	 * 
	 * @return a newly allocated Translator object.
	 */
	protected Translator inputTranslator(KXMLElement node) {
		String stype;
		byte type;
		String slogic;
		boolean logic;

		String name = node.getTagName();
		if (name.equals(LINEARTRANSLATOR)) {
			return new LinearTranslator(node.getProperty(RATE, -1), node
					.getProperty(OFFSET, -1));
		} else if (name.equals(BINARYTRANSLATOR)) {
			stype = node.getProperty(TYPE);
			if (stype.equals(UPPERBOUND))
				type = BinaryTranslator.UPPERBOUND;
			else if (stype.equals(LOWERBOUND))
				type = BinaryTranslator.LOWERBOUND;
			else
				type = BinaryTranslator.EQUALITY;
			slogic = node.getProperty(LOGIC);
			if (slogic.equals(POSITIVE))
				logic = true;
			else
				logic = false;
			return new BinaryTranslator(type, node.getFloatProperty(THRESHOLD,
					(float) 0), logic);
		} else if (name.equals(TOBINARYTRANSLATOR)) {
			slogic = node.getProperty(LOGIC);
			if (slogic.equals(POSITIVE))
				logic = true;
			else
				logic = false;
			return new ToBinaryFormatter103(logic);
		} else if (name.equals(FROMBINARYTRANSLATOR)) {
			slogic = node.getProperty(LOGIC);
			if (slogic.equals(POSITIVE))
				logic = true;
			else
				logic = false;
			return new FromBinaryFormatter103(logic);
		}

		return null;
	}

	/**
	 * Instanciate a KXMLElement from an instance of Translator.
	 * 
	 * @param node
	 *            an instance of Translator.
	 * 
	 * @return a KXMLElement XXXXXTranslator.
	 */
	// protected KXMLElement outputTranslator(Translator translator)
	// {
	// KXMLElement node = new KXMLElement();
	//		
	// if (translator instanceof BinaryTranslator)
	// {
	// BinaryTranslator trans = (BinaryTranslator)translator;
	//			
	// node.setTagName(BINARYTRANSLATOR);
	//			
	// switch (trans.type)
	// {
	// case BinaryTranslator.UPPERBOUND:
	// node.addProperty(TYPE,UPPERBOUND);
	// break;
	// case BinaryTranslator.LOWERBOUND:
	// node.addProperty(TYPE,LOWERBOUND);
	// break;
	// case BinaryTranslator.EQUALITY:
	// node.addProperty(TYPE,EQUALITY);
	// break;
	// default:
	// break;
	// }
	// node.addProperty(THRESHOLD,Float.toString(trans.threshold));
	// node.addProperty(LOGIC,(trans.logic)?POSITIVE:NEGATIVE);
	// }
	// else if (translator instanceof Formatter103)
	// {
	// Formatter103 trans = (Formatter103)translator;
	//			
	// node.setTagName(FORMATTERTRANSLATOR);
	//			
	// switch (trans.type)
	// {
	// case Formatter103.UPPERBOUND:
	// node.addProperty(TYPE,UPPERBOUND);
	// break;
	// case Formatter103.LOWERBOUND:
	// node.addProperty(TYPE,LOWERBOUND);
	// break;
	// case Formatter103.EQUALITY:
	// node.addProperty(TYPE,EQUALITY);
	// break;
	// default:
	// break;
	// }
	// node.addProperty(THRESHOLD,Float.toString(trans.threshold));
	// node.addProperty(LOGIC,(trans.logic)?POSITIVE:NEGATIVE);
	// }
	// else if (translator instanceof LinearTranslator)
	// {
	// LinearTranslator trans = (LinearTranslator)translator;
	// node.setTagName(LINEARTRANSLATOR);
	// node.getProperty(RATE,trans.rate);
	// node.getProperty(OFFSET,trans.offset);
	// }
	// return node;
	// }

	// /**
	// * Instanciate a Writer and its key from XML configuration.
	// *
	// * @param node a KXMLElement Writer.
	// *
	// * @return an object array with Writer and related WriterKey.
	// */
	// protected Object[] inputWriter(KXMLElement node)
	// {
	// Object[] result = new Object[2];
	// MapSingleValue[] value = new MapSingleValue[node.countChildren()-1];
	// int i = 0;
	// for(Enumeration e = node.enumerateChildren();e.hasMoreElements();)
	// {
	// KXMLElement child = (KXMLElement)e.nextElement();
	// String name = child.getTagName();
	// if(name.equals(WRITERKEY))
	// result[0] = new WriterKey103(
	// (short)child.getProperty(ID,-1),
	// (short)child.getProperty(VSQ,-1),
	// child.getProperty(CAA,-1),
	// (short)child.getProperty(FT,-1),
	// (short)child.getProperty(IN,-1)
	// );
	// else
	// value[i++] = inputMapValue(child);
	// }
	// result[1] = new Writer(value);
	//			
	// return result;
	// }

	/**
	 * Instanciate a KXMLElement from an instance of Writer and another of
	 * related key WriterKey.
	 * 
	 * @param node
	 *            an instance of Writer and one of WriterKey.
	 * 
	 * @return a KXMLElement Writer.
	 */
	// protected KXMLElement outputWriter(Writer writer, WriterKey103 key)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(WRITER);
	//		
	// KXMLElement knode = new KXMLElement();
	// knode.setTagName(WRITERKEY);
	// knode.addProperty(ID,key.id);
	// knode.addProperty(VSQ,key.vsq);
	// knode.addProperty(CAA,key.caa);
	// knode.addProperty(FT,key.ft);
	// knode.addProperty(IN,key.in);
	//		
	// node.addChild(knode);
	//		
	// for(int i=0;i<writer.value.length;)
	// node.addChild(outputMapValue(writer.value[i++]));
	//		
	// return node;
	// }

	/**
	 * Instanciate a Reader from XML configuration.
	 * 
	 * @param node
	 *            a KXMLElement Reader.
	 * 
	 * @return an instance of Reader.
	 */
	protected SMapAsdu inputMapAsdu(KXMLElement node) {
		AsduKey103 key = null;
		MapSingleValue[] value = new MapSingleValue[node.countChildren() - 1];
		int i = 0;
		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();) {
			KXMLElement child = (KXMLElement) e.nextElement();
			String name = child.getTagName();
			if (name.equals(KEY))
				key = new AsduKey103((short) child.getProperty(ID, -1),
						(short) child.getProperty(VSQ, -1), (short) child
								.getProperty(COT, -1), child.getProperty(CAA,
								-1), (short) child.getProperty(FT, -1),
						(short) child.getProperty(IN, -1));
			else
				value[i++] = inputMapValue(child);
		}
		return new SMapAsdu(key, value, node.getProperty(CLASS1, TRUE, FALSE,
				false), node.getProperty(CLASS2, TRUE, FALSE, false), node
				.getProperty(GI, TRUE, FALSE, false), node.getProperty(WRITE,
				TRUE, FALSE, false));
	}

	/**
	 * Instanciate a KXMLElement from an instance of Reader.
	 * 
	 * @param node
	 *            an instance of Reader..
	 * 
	 * @return a KXMLElement Reader.
	 */
	// protected KXMLElement outputReader(Reader reader)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(READER);
	// ReaderKey103 key = (ReaderKey103)reader.key;
	// KXMLElement knode = new KXMLElement();
	// knode.setTagName(READERKEY);
	// knode.addProperty(ID,key.id);
	// knode.addProperty(VSQ,key.vsq);
	// knode.addProperty(CAA,key.cot);
	// knode.addProperty(CAA,key.caa);
	// knode.addProperty(FT,key.ft);
	// knode.addProperty(IN,key.in);
	// node.addChild(knode);
	// ReaderKey103 answer =
	// (ReaderKey103)((MasterServiceHelper)reader.svc).expectedKey;
	// KXMLElement snode = new KXMLElement();
	// snode.setTagName(ANSWERKEY);
	// snode.addProperty(ID,answer.id);
	// snode.addProperty(VSQ,answer.vsq);
	// snode.addProperty(CAA,answer.cot);
	// snode.addProperty(CAA,answer.caa);
	// snode.addProperty(FT,answer.ft);
	// snode.addProperty(IN,answer.in);
	// node.addChild(snode);
	// for(int i=0;i<reader.value.length;)
	// node.addChild(outputMapValue(reader.value[i++]));
	// return node;
	// }

	/**
	 * Instanciate a Mapping from XML configuration.
	 * 
	 * @param node
	 *            A KXMLElement Mapping.
	 * @param manager
	 *            Application layer manager.
	 * 
	 * @return an instance of Mapping.
	 */
	protected SMapping103 inputMapping(KXMLElement node, SAppManager103 manager) {
		SMapping103 map = new SMapping103(manager);
		map.addIdentification(new SDataClass1AppSvc(manager, new MapAsdu(
				new AsduKey103(Interoperability103.ID_IDENTIFICATION,
						(short) 129, Interoperability103.COT_RESET_CU,
						protocol.caa, protocol.ft,
						Interoperability103.IN_RESET_CU))));
		map.addIdentification(new SDataClass1AppSvc(manager, new MapAsdu(
				new AsduKey103(Interoperability103.ID_IDENTIFICATION,
						(short) 129, Interoperability103.COT_RESET_FCB,
						protocol.caa, protocol.ft,
						Interoperability103.IN_RESET_FCB))));
		map.addIdentification(new SDataClass1AppSvc(manager, new MapAsdu(
				new AsduKey103(Interoperability103.ID_IDENTIFICATION,
						(short) 129, Interoperability103.COT_START_RESTART,
						protocol.caa, protocol.ft,
						Interoperability103.IN_START_RESTART))));
		map.addIdentification(new SDataClass1AppSvc(manager, new MapAsdu(
				new AsduKey103(Interoperability103.ID_IDENTIFICATION,
						(short) 129, Interoperability103.COT_POWER,
						protocol.caa, protocol.ft,
						Interoperability103.IN_POWER_ON))));
		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();) {
			KXMLElement child = (KXMLElement) e.nextElement();
			SMapAsdu asdu = inputMapAsdu(child);
			// event asdus.
			if (asdu.class1) {
				if (asdu.key.cot == AsduKey.NOTUSED)
					asdu.key.cot = Interoperability103.COT_SPONTANEOUS;
				map.addEvent(new SDataClass1AppSvc(manager, asdu));
			}
			// measure asdus
			if (asdu.class2) {
				if (asdu.key.cot == AsduKey.NOTUSED)
					asdu.key.cot = Interoperability103.COT_CYCLIC;
				map.addMeasure(new SDataClass2AppSvc(manager, asdu));
			}
			// general interrogation asdus
			if (asdu.gi) {
				if (asdu.key.cot == AsduKey.NOTUSED)
					asdu.key.cot = Interoperability103.COT_GI;
				map.addGi(new SDataClass1AppSvc(manager, asdu));
			}
			// writer
			if (asdu.write)
				map.addWriter(asdu);
		}
		return map;
	}

	/**
	 * Instanciate a KXMLElement from an instance of Mapping.
	 * 
	 * @param node
	 *            An instance of Mapping.
	 * 
	 * @return a KXMLElement Mapping.
	 */
	// protected KXMLElement outputMapping(Mapping103 map)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(MAPPING);
	//		
	// Enumeration wkeys = map.enumerateWriterKeys();
	// Enumeration writers = map.enumerateWriters();
	// Enumeration readers = map.enumerateReaders();
	// Enumeration procedures = map.enumerateProcedures();
	//		
	// for(;wkeys.hasMoreElements();)
	// node.addChild(outputWriter((Writer)writers.nextElement(),(WriterKey103)wkeys.nextElement()));
	// for(;readers.hasMoreElements();)
	// node.addChild(outputReader((Reader)readers.nextElement()));
	// for(;procedures.hasMoreElements();)
	// node.addChild(outputProcedure((AppService)procedures.nextElement()));
	//		
	// return node;
	// }
	/**
	 * Instanciate a private asdu from xml configuration.
	 * 
	 * @param node
	 *            a KXMLElement PrvtAsdu.
	 * 
	 * @return an instance of PrvtAsdu.
	 */
	protected PrivateAsdu103 inputPrvtAsdu(KXMLElement node) {
		int i = 0;
		String[] ie = new String[node.countChildren()];
		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();) {
			KXMLElement child = (KXMLElement) e.nextElement();
			// WARN: KXMLElement REHASH!!! WE SHOULD ADD A POSITION FLAG
			ie[i++] = child.getProperty(TYPE);
		}
		return new PrivateAsdu103(node.getProperty(LABEL), (short) node
				.getProperty(ID, -1), (short) node.getProperty(VSQ, -1),
				(short) node.getProperty(COT, -1), node.getProperty(CAA, -1),
				(short) node.getProperty(FT, -1), (short) node.getProperty(IN,
						-1), ie);
	}

	/**
	 * Instanciate a KXMLElement from an instance of private asdu.
	 * 
	 * @param asdu
	 *            an instance of PrivateAsdu.
	 * 
	 * @return a KXMLElement PrvtAsdu.
	 */
	// protected KXMLElement outputPrvtAsdu(PrivateAsdu asdu)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PRVTASDU);
	// node.addProperty(ID,asdu.id);
	// node.addProperty(LABEL,asdu.label);
	// for(int i=0;i<asdu.ie.length;)
	// {
	// KXMLElement child = new KXMLElement();
	// child.setTagName(PRVTELEMENT);
	// child.addProperty(TYPE,asdu.ie[i++]);
	// node.addChild(child);
	// }
	// return node;
	// }
	/**
	 * <p>
	 * Instanciate a Private cause of trnasmission from xml configuration.
	 * 
	 * @param node
	 *            a KXMLElement PrvtCot
	 * 
	 * @return an instance of PrivateCot
	 */
	protected PrivateCot inputPrvtCot(KXMLElement node) {
		return new PrivateCot((short) node.getProperty(ID, -1), node
				.getProperty(LABEL));
	}

	/**
	 * <p>
	 * Instanciate a KXMLElement from an instance of Private cause of
	 * transmission..
	 * 
	 * @param in
	 *            an instance of PrvtCot.
	 * 
	 * @return a KXMLElement PrivateCot.
	 */
	// protected KXMLElement outputPrvtCot(PrivateCot cot)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PRVTCOT);
	// node.addProperty(ID,cot.id);
	// node.addProperty(LABEL,cot.label);
	// return node;
	// }
	/**
	 * <p>
	 * Instanciate a Private information number from xml configuration.
	 * 
	 * @param node
	 *            a KXMLElement PrvtIn
	 * 
	 * @return an instance of PrivateIn
	 */
	protected PrivateIn103 inputPrvtIn(KXMLElement node) {
		return new PrivateIn103((short) node.getProperty(ID, -1), node
				.getProperty(LABEL));
	}

	/**
	 * <p>
	 * Instanciate a KXMLElement from an instance of Private information number.
	 * 
	 * @param in
	 *            an instance of PrivateIn.
	 * 
	 * @return a KXMLElement PrvtIn
	 */
	// protected KXMLElement outputPrvtIn(PrivateIn103 in)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PRVTIN);
	// node.addProperty(ID,in.id);
	// node.addProperty(LABEL,in.label);
	// return node;
	// }
	/**
	 * <p>
	 * Instanciate a Private function type from xml configuration.
	 * 
	 * @param node
	 *            a KXMLElement PrvtFt
	 * 
	 * @return an instance of PrivateFt
	 */
	protected PrivateFt103 inputPrvtFt(KXMLElement node) {
		return new PrivateFt103((short) node.getProperty(ID, -1), node
				.getProperty(LABEL));
	}

	/**
	 * <p>
	 * Instanciate a KXMLElement from an instance of Private function type.
	 * 
	 * @param ft
	 *            an instance of PrivateFt103.
	 * 
	 * @return a KXMLElement PrvtFt
	 */
	// protected KXMLElement outputPrvtFt(PrivateFt103 ft)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PRVTFT);
	// node.addProperty(ID,ft.id);
	// node.addProperty(LABEL,ft.label);
	// return node;
	// }
	/**
	 * <p>
	 * Instanciate a private range from xml configuration.
	 * 
	 * @param node
	 *            PrivateRange KXMLElement.
	 * 
	 * @return an instance of PrivateRange.
	 */
	protected PrivateRange103 inputPrivateRange(KXMLElement node) {
		PrivateRange103 prange = new PrivateRange103();
		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();) {
			KXMLElement child = (KXMLElement) e.nextElement();
			String name = child.getTagName();
			if (name.equals(PRVTASDU))
				prange.addAsdu(inputPrvtAsdu(child));
			else if (name.equals(PRVTCOT))
				prange.addCot(inputPrvtCot(child));
			else if (name.equals(PRVTIN))
				prange.addIn(inputPrvtIn(child));
			else if (name.equals(PRVTFT))
				prange.addFt(inputPrvtFt(child));
			// else
			// throw new
			// IllegalArgumentException("PrivateRange has unknown sub element");
		}
		return prange;
	}

	/**
	 * <p>
	 * Instanciate a KXMLElement from an instance of private range.
	 * 
	 * @param node
	 *            an instance of PrivateRange.
	 * 
	 * @return PrivateRange KXMLElement.
	 */
	// protected KXMLElement outputPrivateRange(PrivateRange103 prange)
	// {
	// KXMLElement node = new KXMLElement();
	// node.setTagName(PRIVATERANGE);
	// for(Enumeration e = prange.enumerateAsdus();e.hasMoreElements();)
	// node.addChild(outputPrvtAsdu((PrivateAsdu)e.nextElement()));
	// for(Enumeration e = prange.enumerateCots();e.hasMoreElements();)
	// node.addChild(outputPrvtCot((PrivateCot)e.nextElement()));
	// for(Enumeration e = prange.enumerateFts();e.hasMoreElements();)
	// node.addChild(outputPrvtFt((PrivateFt103)e.nextElement()));
	// for(Enumeration e = prange.enumerateIns();e.hasMoreElements();)
	// node.addChild(outputPrvtIn((PrivateIn103)e.nextElement()));
	// return node;
	// }
	/**
	 * <p>
	 * Instanciate a Protocol for general application settings from a
	 * KXMLElement.
	 * 
	 * @param node
	 *            a KXMLElement Protocol
	 * 
	 * @return an instance of Protocol.
	 */
	protected Protocol inputProtocol(KXMLElement node) {
		return new Protocol(node.getProperty(IDENTIFICATION), node.getProperty(
				ADDRESS, 0), (short) node.getProperty(FT, 0), node.getProperty(
				CAA, 0), node.getProperty(POLLING, 0), node.getProperty(CLOCK,
				TRUE, FALSE, false));
	}

	/**
	 * <p>
	 * Instanciate KXMLElement from an instance of Protocol.
	 * 
	 * @param protocol
	 *            an instance of Protocol.
	 * 
	 * @return a KXMLElement Protocol.
	 */
	protected KXMLElement outputProtocol(Protocol protocol) {
		KXMLElement node = new KXMLElement();
		node.setTagName(PROTOCOL);
		node.addProperty(IDENTIFICATION, protocol.id);
		node.addProperty(ADDRESS, protocol.address);
		node.addProperty(FT, protocol.ft);
		node.addProperty(CAA, protocol.caa);
		node.addProperty(POLLING, protocol.polling);
		node.addProperty(CLOCK, protocol.clock ? TRUE : FALSE);
		return node;
	}

	/**
	 * <p>
	 * Instanciate a Communication for general comm settings from a KXMLElement.
	 * 
	 * @param node
	 *            a KXMLElement Communication.
	 * 
	 * @return an instance of Communication.
	 */
	protected Communication inputCommunication(KXMLElement node) {
		return new Communication(node.getProperty(PORT).toLowerCase(), node
				.getProperty(BAUDRATE, 9600));

	}

	/**
	 * <p>
	 * Instanciate a KXMLElement from an instance of Communication. from a
	 * KXMLElement.
	 * 
	 * @param comm
	 *            an instance of Communication.
	 * 
	 * @return a KXMLElement Communication.
	 */
	protected KXMLElement outputCommunication(Communication comm) {
		KXMLElement node = new KXMLElement();
		node.setTagName(COMMUNICATION);
		node.addProperty(PORT, comm.port.toUpperCase());
		node.addProperty(BAUDRATE, comm.baud);
		return node;
	}

	/**
	 * <p>
	 * Update a whole application configuration from a KXMLElement node.
	 * 
	 * @param node
	 *            a KXMLElement ROOT.
	 */
	public void inputFromXML(KXMLElement node) {
		if (!node.getTagName().equals(ROOT))
			return;

		author = node.getProperty(AUTHOR);
		version = node.getProperty(VERSION);
		date = node.getProperty(DATE);
		fileId = node.getProperty(FILEID);

		SLinkLayer lnk = null;
		PhysicalLayer phy = null;

		for (Enumeration e = node.enumerateChildren(); e.hasMoreElements();) {
			KXMLElement child = (KXMLElement) e.nextElement();
			String name = child.getTagName();
			if (name.equals(COMMUNICATION)) {
				comm = inputCommunication(child);
			} else if (name.equals(PROTOCOL)) {
				protocol = inputProtocol(child);
				app = new SApplicationLayer();
				lnk = new SLinkLayer();
				// phy = new
				// PhysicalLayer(comm.port,comm.baud,Communication.data,Communication.stop,Communication.parity);
				SContext103 context = new SContext103(protocol.address,
						protocol.id, protocol.ft, protocol.caa, protocol.clock);
				app.lower = lnk;
				app.reader = new SAsduReader103();
				app.writer = new SAsduWriter103();
				lnk.upper = app;
				lnk.lower = phy;
				phy.upper = lnk;
				app.manager = new SAppManager103(app, context, protocol.polling);
			} else if (name.equals(PRIVATERANGE)) {
				backupPrvt = child;
				app.manager.prange = inputPrivateRange(child);
				((SAsduWriter103) app.writer).prange = (PrivateRange103) app.manager.prange;
			} else if (name.equals(MAPPING)) {
				backupMap = child;
				app.manager.map = inputMapping(child,
						(SAppManager103) app.manager);
				app.manager.map.manager = app.manager;
				int i = 0;
				for (Enumeration e2 = ((SMapping103) app.manager.map)
						.enumerateMeasures(); e.hasMoreElements();) {
					Object tmp = e2.nextElement();
					i++;
				}
				lnk.manager.setNbMeasure(i);
			}
		}
	}

	/**
	 * <p>
	 * Instanciate a KXMLElement ROOT from an configuration.
	 */
	public KXMLElement outputToXML() {
		KXMLElement node = new KXMLElement();
		node.setTagName(ROOT);

		node.addProperty(AUTHOR, author);
		node.addProperty(VERSION, version);
		node.addProperty(DATE, date);
		node.addProperty(FILEID, fileId);

		node.addChild(outputCommunication(comm));
		node.addChild(outputProtocol(protocol));

		node.addChild(backupPrvt);
		node.addChild(backupMap);
		return node;
	}

	/**
	 * <p>
	 * Kills all threads in current session.
	 */
	public void kill() {
		if (app != null) {
			app.kill();
			try {
				app.join();
			} catch (Exception e) {
				CodeProperties.log("S103 killed.");
			}
			CodeProperties.log("S103 really killed.");
		}
	}

	/**
	 * <p>
	 * starts application
	 */
	public void start() {
		if (!comm.port.equals("unknown")) {
			app.start();
		}
	}

	/*
	 * /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\*\ < < >
	 * Constructor > < <
	 * \*\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
	 */

	/**
	 * Instanciate a new configurator for master 103 application.
	 */
	public SConfiguration103() {
		comm = new Communication();
		protocol = new Protocol();
		app = new SApplicationLayer();
		backupPrvt = new KXMLElement();
		backupMap = new KXMLElement();
	}

} // end SConfiguration103.java