/*
 * Class to store string properties.
 *
 * Class      com.itlity.util.Properties
 * File       Properties.java
 * Author     Alexis BIETTI <alexis.bietti@sysaware.com>
 * (c)        SysAware <http://www.sysaware.com>
 */

//-----------------------------------------------------------------------------
// Imports
//-----------------------------------------------------------------------------
package br.org.scadabr.protocol.util;

import java.util.Enumeration;
import java.util.Hashtable;

import br.org.scadabr.protocol.lib.math.Math32FP;

//-----------------------------------------------------------------------------
// Class
//-----------------------------------------------------------------------------

/**
 * This class stores string properties. The keys are the properties, as the
 * elements are the values of these properties.
 *
 * @author <a href="mailto:alexis.bietti@sysaware.com">Alexis BIETTI</a>
 */
public class Properties extends Hashtable {
  //---------------------------------------------------------------------------
  // Constructors
  //---------------------------------------------------------------------------

  /**
   * Construct a new Properties object with a default initial capacity.
   */
  public Properties() {
    super();
  }

  /**
   * Construct a new Properties object with the specified default capacity.
   */
  public Properties(int initialCapacity) {
    super(initialCapacity);
  }

  /**
   * Construct a newly allocated Properties object that is a copy of a given
   * Properties object.
   *
   * @param orig a Properties object to copy
   */
  public Properties(Properties orig) {
    for (Enumeration e = orig.keys(); e.hasMoreElements();) {
      String key   = (String) e.nextElement();
      String value = (String) orig.get(key);
      put(key, value);
    }
  }

  /**
   * Basic test
   *
   * @param args unused
   */
  public static void main(String[] args) {
    Properties p = new Properties();
    p.setProperty("a", "1");
    System.out.println(p.getIntProperty("a"));
  }

  //---------------------------------------------------------------------------
  // Public methods
  //---------------------------------------------------------------------------

  /**
   * Get a property's value.
   *
   * @param key a String
   *
   * @return a property's value.
   */
  public String getProperty(String key) {
    return (String) get(key);
  }

  /**
   * Get a property's value, or the default value if the property is not
   * defined.
   *
   * @param key a String
   * @param defaultValue a String
   */
  public String getProperty(String key, String defaultValue) {
    String val = getProperty(key);

    return ((val != null) ? val : defaultValue);
  }

  /**
   * Return a property cast as a boolean.
   */
  public boolean getBooleanProperty(String key) {
    int value = Integer.parseInt((String) get(key));

    return (value != 0);
  }

  /**
   * @param key
   * @param defaultValue
   *
   * @return boolean
   */
  public boolean getProperty(String key, boolean defaultValue) {
    String value = getProperty(key);

    if (value == null) {
      return defaultValue;
    }

    if (value.equals("1") || value.equals("true")) {
      return true;
    }

    if (value.equals("0") || value.equals("false")) {
      return false;
    }

    return defaultValue;
  }

  /**
   * Get a property's value converted to an integer.
   *
   * @param key a String
   *
   * @return a property's value converted to an integer.
   */
  public int getIntProperty(String key) {
    return Integer.parseInt((String) get(key));
  }

  /**
   * @param key
   * @param defaultValue
   *
   * @return int
   */
  public int getProperty(String key, int defaultValue) {
    try {
      return getIntProperty(key);
    } catch (Exception e) {
      return defaultValue;
    }
  }

  /**
   * Return a property cast as a fixed point decimal number.
   */
  public int getFPProperty(String key) {
    return Math32FP.toFP((String) get(key));
  }

  /**
   * Return a property cast as a <code> float </code> number.
   */
  public float getFloatProperty(String key) {
    return Float.parseFloat(key);
  }

  /**
   * @param key
   * @param defaultValue
   *
   * @return float
   */
  public float getProperty(String key, float defaultValue) {
    try {
      return getFloatProperty(key);
    } catch (Exception e) {
      return defaultValue;
    }
  }

  /**
   * Update a property value, or create it if necessary.
   *
   * @param key
   * @param value a String value
   *
   * @return the old value, or <tt>null</tt>
   */
  public String setProperty(String key, String value) {
    return (String) put(key, value);
  }

  /**
   * Update a property value, or create it if necessary.
   *
   * @param key
   * @param value an integer value
   *
   * @return the old value, or <tt>null</tt>
   */
  public String setProperty(String key, int value) {
    return (String) put(key, Integer.toString(value));
  }

  /**
   * Update a property value, or create it if necessary.
   *
   * @param key
   * @param value a fixed point decimal value
   *
   * @return the old value, or <tt>null</tt>
   */
  public String setFPProperty(String key, int value) {
    return setProperty(key, Math32FP.toString(value));
  }

  /**
   * Update a property value, or create it if necessary.
   *
   * @param key
   * @param value a boolean value
   *
   * @return the old value, or <tt>null</tt>
   */
  public String setProperty(String key, boolean value) {
    return setProperty(key, value ? "1" : "0");
  }

  /**
   * Removes the property designed by the given key.
   *
   * @param key the key to remove
   *
   * @deprecated use method remove(Object) instead
   */
  public String removeProperty(String key) {
    return (String) remove(key);
  }
}


// End Properties.java
