package com.i2msolucoes.alpha24j.layer.physical.serial;

import gnu.io.SerialPort;

import java.io.IOException;

import com.i2msolucoes.alpha24j.element.Elements.ControlCodes;
import com.i2msolucoes.alpha24j.exception.COMMException;
import com.i2msolucoes.alpha24j.layer.physical.PhysicalLayer;
import com.i2msolucoes.alpha24j.util.ByteUtils;
import com.serotonin.io.serial.SerialParameters;
import com.serotonin.io.serial.SerialUtils;
import com.serotonin.messaging.IncomingMessage;
import com.serotonin.messaging.IncomingRequestMessage;
import com.serotonin.messaging.IncomingResponseMessage;
import com.serotonin.messaging.MessageControl;
import com.serotonin.messaging.MessageParser;
import com.serotonin.messaging.OutgoingRequestMessage;
import com.serotonin.messaging.OutgoingResponseMessage;
import com.serotonin.messaging.StreamTransport;
import com.serotonin.messaging.WaitingRoomKey;
import com.serotonin.messaging.WaitingRoomKeyFactory;
import com.serotonin.util.queue.ByteQueue;

public class SerialPhysicalLayer implements PhysicalLayer {

	public final static String COMM_PORT = "COMM_PORT";
	public final static String BAUDRATE = "BAUDRATE";
	public final static String DATABITS = "DATABITS";
	public final static String STOPBITS = "STOPBITS";
	public final static String PARITY = "PARITY";
	public final static String RETRIES = "RETRIES";
	public final static String TIMEOUT = "TIMEOUT";

	private SerialParameters parameters;
	private int timeout;
	private int retries;
	private SerialPort serialPort;
	private MessageControl conn;

	public SerialPhysicalLayer(SerialParameters parameters, int timeout,
			int retries) {
		this.parameters = parameters;
		this.timeout = timeout;
		this.retries = retries;
	}

	@Override
	public void init() throws COMMException {
		try {
			this.serialPort = SerialUtils.openSerialPort(parameters);
			this.serialPort.setRTS(true);
			this.serialPort.setDTR(true);

			this.conn = new MessageControl();
			conn.setRetries(retries);
			conn.setTimeout(timeout);

			StreamTransport transport = null;

			transport = new StreamTransport(serialPort.getInputStream(),
					serialPort.getOutputStream());
			conn.start(transport, new SingleParser(), null,
					new SerialWaitingRoomKeyFactory());
			transport.start("Alpha2 Serial COMM");
		} catch (Exception e) {
			e.printStackTrace();
			throw new COMMException(e.getMessage());
		}

	}

	@Override
	public byte[] sendAndReceive(byte[] frame) throws IOException {
		SingleResponseMessage message = (SingleResponseMessage) conn
				.send(new SingleRequestMessage(frame));
		return message.getBytes();
	}

	@Override
	public void send(byte[] frame) throws IOException {
		conn.send(new SingleRequestMessage(frame, false));
	}

	@Override
	public void close() throws COMMException {
		try {
			SerialUtils.close(serialPort);
		} catch (Exception e) {
			e.printStackTrace();
			throw new COMMException(e.getMessage());
		}
	}

	class SingleParser implements MessageParser {
		@Override
		public IncomingMessage parseMessage(ByteQueue bytes) throws Exception {
			if (isExpectedMessage(bytes)) {
				ByteUtils.displayBytes(bytes.peekAll());
				return new SingleResponseMessage(bytes);
			} else {
				return null;
			}
		}

		private boolean isExpectedMessage(ByteQueue bytes) {
			if (bytes.size() < 5)
				return false;

			if (bytes.size() == 5
					&& bytes.peek(4) == ControlCodes.NACK.getCode())
				return true;

			if (bytes.size() == 6
					&& bytes.peek(4) == ControlCodes.NACK.getCode())
				return true;

			if ((bytes.peek(bytes.size() - 3) == ControlCodes.ETX.getCode())
					|| (bytes.peek(bytes.size() - 1) == ControlCodes.ACK
							.getCode()))
				return true;

			return false;
		}
	}

	class SingleResponseMessage extends SerialMessage implements
			OutgoingResponseMessage, IncomingResponseMessage {

		private byte[] bytes;

		public SingleResponseMessage(ByteQueue bytes) {
			this.bytes = (bytes.popAll());
		}

		@Override
		public byte[] getMessageData() {
			return bytes;
		}

		public byte[] getBytes() {
			return bytes;
		}
	}

	class SerialWaitingRoomKey implements WaitingRoomKey {

		@Override
		public int hashCode() {
			int result = 1;
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			return true;
		}
	}

	public class SingleRequestMessage extends SerialMessage implements
			OutgoingRequestMessage, IncomingRequestMessage {

		private byte[] bytes;
		private boolean expectsResponse = true;

		public SingleRequestMessage(byte[] bytes) {
			this.bytes = bytes;
		}

		public SingleRequestMessage(byte[] bytes, boolean expectsResponse) {
			this.bytes = bytes;
			this.expectsResponse = expectsResponse;
		}

		@Override
		public byte[] getMessageData() {
			return this.bytes;
		}

		@Override
		public boolean expectsResponse() {
			return this.expectsResponse;
		}

		public byte[] getBytes() {
			return bytes;
		}

	}

	class SerialMessage {

		public WaitingRoomKey getWaitingRoomKey() {
			return new SerialWaitingRoomKey();
		}
	}

	class SerialWaitingRoomKeyFactory implements WaitingRoomKeyFactory {
		@Override
		public WaitingRoomKey createWaitingRoomKey(
				OutgoingRequestMessage request) {
			return new SerialWaitingRoomKey();
		}

		@Override
		public WaitingRoomKey createWaitingRoomKey(
				IncomingResponseMessage response) {
			return new SerialWaitingRoomKey();
		}

	}

}
