package com.i2msolucoes.alpha24j.layer.user;

import java.util.ArrayList;
import java.util.List;

import com.i2msolucoes.alpha24j.DeviceLocator;
import com.i2msolucoes.alpha24j.DeviceValue;
import com.i2msolucoes.alpha24j.DeviceWrite;
import com.i2msolucoes.alpha24j.exception.COMMException;
import com.i2msolucoes.alpha24j.exception.ErrorMessageReceivedException;
import com.i2msolucoes.alpha24j.exception.InvalidFrameReceivedException;
import com.i2msolucoes.alpha24j.frame.request.Alpha2RequestFrame;
import com.i2msolucoes.alpha24j.frame.request.LineCheckCommand;
import com.i2msolucoes.alpha24j.frame.request.ReadCommand;
import com.i2msolucoes.alpha24j.frame.request.RunStopCommand;
import com.i2msolucoes.alpha24j.frame.request.WriteCommand;
import com.i2msolucoes.alpha24j.frame.request.RunStopCommand.RunStopCodes;
import com.i2msolucoes.alpha24j.frame.response.Alpha2ResponseFrame;
import com.i2msolucoes.alpha24j.frame.response.ErrorResponseFrame;
import com.i2msolucoes.alpha24j.frame.response.ReadResponseFrame;
import com.i2msolucoes.alpha24j.layer.application.ApplicationLayer;
import com.i2msolucoes.alpha24j.layer.physical.PhysicalLayer;
import com.i2msolucoes.alpha24j.layer.physical.serial.SerialPhysicalLayer;
import com.serotonin.io.serial.SerialParameters;

public class Alpha2Master {

	public static final int MAX_DEVICES_PER_READ_MESSAGE = 82;
	private byte station;
	private ApplicationLayer appLayer;
	private boolean initialized = false;

	public Alpha2Master(byte station, SerialParameters parameters, int timeout,
			int retries) {
		this.station = station;
		PhysicalLayer phyLayer = new SerialPhysicalLayer(parameters, timeout,
				retries);
		this.appLayer = new ApplicationLayer(phyLayer);
	}

	public void init() throws COMMException {
		if (!initialized) {
			this.appLayer.init();
			initialized = true;
		}

	}

	public void terminate() throws COMMException {
		if (initialized) {
			this.appLayer.terminate();
			initialized = false;
		}

	}

	public void runController() throws COMMException,
			ErrorMessageReceivedException, InvalidFrameReceivedException {

		Alpha2RequestFrame runCommand = new RunStopCommand(station,
				RunStopCodes.RUN);
		Alpha2ResponseFrame response = appLayer.sendFrame(runCommand);
		treatErrorResponse(response);
	}

	public void stopController() throws COMMException,
			ErrorMessageReceivedException, InvalidFrameReceivedException {

		Alpha2RequestFrame runCommand = new RunStopCommand(station,
				RunStopCodes.STOP);
		Alpha2ResponseFrame response = appLayer.sendFrame(runCommand);
		treatErrorResponse(response);
	}

	public void lineCheck() throws COMMException,
			ErrorMessageReceivedException, InvalidFrameReceivedException {
		LineCheckCommand lineCheck = new LineCheckCommand(station);
		Alpha2ResponseFrame response = appLayer.sendFrame(lineCheck);
		treatErrorResponse(response);
		try {
			Thread.sleep(600);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	public List<DeviceValue> read(List<DeviceLocator> devices)
			throws COMMException, InvalidFrameReceivedException,
			ErrorMessageReceivedException {
		Alpha2RequestFrame readCommand = new ReadCommand(station, devices);
		Alpha2ResponseFrame response = appLayer.sendFrame(readCommand);

		treatErrorResponse(response);

		if (response.isReadResponseMessage())
			return ((ReadResponseFrame) response).getValues();
		return new ArrayList<DeviceValue>();
	}

	public void write(List<DeviceWrite> devicesToWrite) throws COMMException,
			InvalidFrameReceivedException, ErrorMessageReceivedException {
		Alpha2RequestFrame writeCommand = new WriteCommand(station,
				devicesToWrite);
		Alpha2ResponseFrame response = appLayer.sendFrame(writeCommand);

		treatErrorResponse(response);
	}

	private void treatErrorResponse(Alpha2ResponseFrame response)
			throws ErrorMessageReceivedException, InvalidFrameReceivedException {

		if (response.isErrorMessage()) {
			throw new ErrorMessageReceivedException(
					((ErrorResponseFrame) response).getErrorCode());
		} else if (response.isInvalidFrame()) {
			throw new InvalidFrameReceivedException(response.getReceived());
		}
	}
}
